//AI.C 
//By Ian Tomkins
//Started on 18.1.99
//Contains all code for performing computer AI.

#include "header.h"
#include "frontend.h"
#include "fade.h"
#include "render.h"
#include "camera.h"
#include "objects.h"
#include "collide.h"
#include "pickups.h"

#if   PSX

	#include <libsn.h>
	#include "sdlib\sdtypes.h"
   	#include "sdlib\sdfunc.h"

#endif

#include "ctrller.h"
#include "gameloop.h"
#include "state.h"

#if   PSX

	#include "PSX\loadpsx.h"
	#include "PSX\mathpsx.h"
	#include "PSX\PSX_pcks.h"
	#include "PSX\PSX_expn.h"
	#include "psx\TEXTpsx.h"

#endif

#include "AI.H"
#include "poon.h"
int	global_rot_range = 256;
int	global_vector_length = 1000;
int	global_tot_rot = 500;
int	global_ave_rot = 200;
int scale_value = 65;
int total_rotation, average_rotation;
int				glb_ct;
SVECTOR			trail[1000];
int				ian_index = 0;
int			display_grid = TRUE;
int			prev_distance = 0;
int			count = 0;
int			old_raceclock = 0;
int			old_rotation = 0;
int			global_player_num = 0;

//SETUP
void AI_init_static_objects(void);
void AI_init_all_players(void);
void AI_init_collision_objects(void);
void AI_init_collision_map(u8 *data);
void AI_clear_collision_map(void);

//UTILITY
short read_short(u8 *ptr);


//AI DCM

void AI_decision_making_machine(AI_NAVIGATION *ptr);
void check_weapons_in_range(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
void use_utility_pickups(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr);
																			  
void set_combat_status(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
void set_weapons_range(AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr);
short set_target_line(AI_NAVIGATION *ptr, AI_TARGET *ai_trgt_ptr);

void target_pickup(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
void target_player(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
void player_avoidance(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr);
void object_avoidance(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
void weapon_avoidance(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
void target_pathway(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);

char scan_pickups(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
//char scan_player_targets(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
char scan_players(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
char scan_static_objects(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
char scan_active_weapons(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);
char scan_pathways(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr);

//END OF AI DCM

//AI FLIGHT CONTROL

void AI_set_pad_values(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr);
void calibrate_rotation(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr);

void AI_collision_avoidance(AI_NAVIGATION *ptr);
void AI_calculate_parameters(AI_NAVIGATION *ptr, OBJ_STRUCT *obj_ptr);
void AI_calculate_range(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr);
void AI_direction_determination(AI_NAVIGATION *ptr);
void AI_extreme_turn_determination(AI_NAVIGATION *ptr);
void AI_velocity_determination(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr);
s8 AI_line_collision(AI_NAVIGATION *ptr, short angle, char angle_flag);
AI_NAVIGATION *init_AI_player_struct(OBJ_STRUCT *obj_ptr);

void AI_determine_navigation_state(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr);
char perform_collision_check(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr);

void alter_heading(AI_NAVIGATION *ptr);

AI_NAVIGATION			AI_pilot[MAXNUMPLAYERS];

PAIR_2D					end_vector;
MATRIX					craft_matrix;
int						global_distance;

//END OF AI FLIGHT CONTROL

//AI COLLISION
s8 test_AI_line(PAIR_2D *start, PAIR_2D *finish, u8 grid_section);
s8  raycast(PAIR_2D *start, PAIR_2D *finish);


AI_COLLISION_OBJECT *AI_init_new_object(void);
AI_DCM_PICKUP_OBJECT *AI_init_new_pickup(void);
void AI_init_all_pickups(void);

PAIR_2D					AI_inter;
AI_COLLISION_MAP		AI_course_map;
AI_COLLISION_OBJECT		AI_objects[MAX_AI_OBJECTS];
AI_DCM_PICKUP_OBJECT	AI_pickups[MAXNUMPICKUPS];

//END OF AI COLLISION



//FOR TESTING NAVIGATION ONLY
//SVECTOR			AI_trail[1000];
int				AI_index_n = 0;
int				AI_total_flag = FALSE;
u8				AI_timer_val = 0;

int				coll_pt_flag = 0;

REPLAY_STRUCT	act_rply;

void AI_debug_function(AI_NAVIGATION *ptr);

SVECTOR 			PSX_c_pos;
SVECTOR				point;

//END OF TESTING
		

//FOR DRAWING MAP ONLY

//SVECTOR					AI_drawing_map[1000];
int						AI_index = 0;

//END OF FOR DRAWING MAP ONLY

//AI_skill_level_table[10][][MAXNUMLEVELS] = {

/*
	will include ...

	CONE_SCALER
	ptr->viewing_distance
	weapon, and pickup, firing ranges and steering ranges etc...
	braking and accelerating values

	min and max angles for cone.

	..thus will be per ship per level specifics.

	ptr->detection_length
	tracking time
*/

//};




void AI_init_data(u8 *data)
{
	AI_clear_collision_map();
	AI_init_collision_map(data);
	
	AI_init_all_pickups();

	init_replay_data();
   	AI_init_all_players();

	AI_init_collision_objects();
	AI_init_all_players();

	AI_init_static_objects();

//	OPSX_init_new_explosion(&OBJ_objects[1], missile_trail, OTHER_COLLISION);

}

void AI_clear_collision_map(void)
{
	int					i;
	AI_COLLISION_MAP	*map_ptr;


	map_ptr = &AI_course_map;

	for (i = 0; i < MAX_AI_COLLISION_MAP_ENTRIES; i++)
	{
		map_ptr->list[i][0].vx = 0;//NULLVAL;
		map_ptr->list[i][0].vz = 0;//NULLVAL;

		map_ptr->list[i][1].vx = 0;//NULLVAL;
		map_ptr->list[i][1].vz = 0;//NULLVAL;
	}

	for (i = 0; i < MAX_AI_GRID_SECTIONS; i++)
		map_ptr->list_ptr[i] = NULL;

	map_ptr->current = NULL;

	map_ptr->total = NULLVAL;

	return;	
}
			
void AI_init_collision_objects(void)
{
	int						i;


	for (i = 0; i < MAX_AI_OBJECTS; i++)
	{
		AI_objects[i].in_use = FALSE;
		AI_objects[i].index = i;
		AI_objects[i].owner = NULLVAL;

		AI_objects[i].vx = 0;
		AI_objects[i].vz = 0;
		AI_objects[i].radius = 0;
	}

	return;
}


int AI_init_collision_object(int x_pos, int z_pos, int rad, int type)
{
	AI_COLLISION_OBJECT		*ptr;


	ptr = (AI_COLLISION_OBJECT *)AI_init_new_object();

	if (ptr == NULL)
		return NULLVAL;

	ptr->vx = x_pos;
	ptr->vz = z_pos;

	ptr->radius = rad;

	ptr->type = type;
					
	return ptr->index;
}

int AI_init_pickup(int x_pos, int y_pos, int z_pos, int type)
{
	AI_DCM_PICKUP_OBJECT		*ptr;


	ptr = (AI_DCM_PICKUP_OBJECT *)AI_init_new_pickup();

	if (ptr == NULL)
		return NULLVAL;

	ptr->position.vx = x_pos;
	ptr->position.vy = y_pos;
	ptr->position.vz = z_pos;

	ptr->type = type;
					
	return ptr->index;
}

void AI_init_static_objects(void)
{
	int						i;
	AI_COLLISION_OBJECT		*ptr;

/*
 	for (i = 0; i < 9; i++)
	{
		ptr = (AI_COLLISION_OBJECT *)AI_init_new_object();

		if (ptr == NULL)
			continue;

		switch (i)
		{
			case 0:
								
				ptr->vx = 13786;
				ptr->vz = 10152;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 1:
								
				ptr->vx = 19253;
				ptr->vz = 795;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 2:
								
				ptr->vx = 17817;
				ptr->vz = -17623;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 3:
								
				ptr->vx = -8482;
				ptr->vz = -17489;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 4:
								
				ptr->vx = -11593;
				ptr->vz = -18392;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 5:
								
				ptr->vx = -14313;
				ptr->vz = -19042;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 6:
								
				ptr->vx = -16899;
				ptr->vz = -17928;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 7:
								
				ptr->vx = -20691;
				ptr->vz = -14957;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 8:
								
				ptr->vx = -19877;
				ptr->vz = -7860;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 9:
								
				ptr->vx = -17399;
				ptr->vz = -1154;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 10:
								
				ptr->vx = -12334;
				ptr->vz = 1471;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 11:
								
				ptr->vx = -8354;
				ptr->vz = 1222;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 12:
								
				ptr->vx = -5900;
				ptr->vz = 1178;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 13:
							
				ptr->vx = -6127;
				ptr->vz = 6660;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 14:
								
				ptr->vx = -6731;
				ptr->vz = 12517;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 15:
								
				ptr->vx = -974;
				ptr->vz = 17926;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 16:
								
				ptr->vx = 5941;
				ptr->vz = 18283;

				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

			case 17:
								
				ptr->vx = 15406;
				ptr->vz = -13851;

 				ptr->radius = 0;

				ptr->type = multiple_path;
								
				break;

		}								
	}
	
*/	
 	for (i = 0; i < 9; i++)
	{
		ptr = (AI_COLLISION_OBJECT *)AI_init_new_object();

		if (ptr == NULL)
			continue;

		switch (i)
		{
			case 0:
								
				ptr->vx = -20382;
				ptr->vz = 25482;

				ptr->radius = 800;

				ptr->type = static_object;
								
				break;

			case 1:
								
				ptr->vx = -28038;
				ptr->vz = -1167;

				ptr->radius = 400;
				
				ptr->type = static_object;

				break;

 			case 2:
								
				ptr->vx = -27956;
				ptr->vz = -2697;

				ptr->radius = 400;
				
				ptr->type = static_object;

				break;

			case 3:
								
				ptr->vx = -7737;
				ptr->vz = -16630;

				ptr->radius = 400;
				
				ptr->type = static_object;

				break;

			case 4:
								
				ptr->vx = 9180;
				ptr->vz = 3772;

				ptr->radius = 400;
				
				ptr->type = static_object;

				break;

			case 5:
								
				ptr->vx = 8366;
				ptr->vz = 5874;

				ptr->radius = 400;
				
				ptr->type = static_object;

				break;

			case 6:
								
				ptr->vx = 14204;
				ptr->vz = 10952;

				ptr->radius = 400;
							  
				ptr->type = static_object;

				break;

			case 7:
								
				ptr->vx = 26772;
				ptr->vz = 14336;

				ptr->radius = 400;
				
				ptr->type = static_object;

				break;

			case 8:
								
				ptr->vx = 26297;
				ptr->vz = 18857;

				ptr->radius = 400;
				
				ptr->type = static_object;

				break;
		}
	}

	return;
}

void AI_init_collision_map(u8 *data)
{
	unsigned short			header[4], total_entries, offset, total_sections;
	unsigned short			sections_in_use = 0;
	s32						i, j;
	AI_COLLISION_MAP		*map_ptr;

	short					result, section_number;
	unsigned char			byte[8];
	int						section_ctr = 0; 
	int						section_no = 0;


	map_ptr = &AI_course_map;
	
//	total_sections = read_short(data);
//	data += 2;
	
	byte[0] = *data;
	data++;
	byte[1] = *data;
	data++;		   	
	result = (byte[1] << 8) + byte[0];
	total_sections = result;


//	total_entries = read_short(data);
//	data += 2;
	
	byte[0] = *data;
	data++;
	byte[1] = *data;
	data++;		   	
	result = (byte[1] << 8) + byte[0];
	total_entries = result;

//	header += read_short(data);
//	data += 2;

//	for (i = 0; i < AI_GRID_WIDTH; i++)
	for (i = 0; i < 2; i++)			   //old
	{
		byte[i] = *data;
		data++;
	}
	
//	for (i = 0, j = 0; i < 4; i++, j += 2)
//		header[i] = (byte[j + 1] << 8) + byte[j];
//	for (i = 0, j = 0; i < 4; i++, j += 2)
		header[0] = (byte[1] << 8) + byte[0]; //old

//	for (j = 0; j < 4; j++)
	for (j = 0; j < 1; j++)    //old
	{
	  	for (i = 0; i < 16; i++)
		{
			if ((header[j] & (1 << i)))
			{
				sections_in_use ++;
				section_number = i + (j * 16);

//				printf("section number = %d\n", section_number);
				byte[0] = *data;
 				data++;                            
   				byte[1] = *data;               
 				data++;		   	                   
 				result = (byte[1] << 8) + byte[0]; 
 				offset = result;                   

//				offset = read_short(data);
//				data += 2;
  				printf("section %d, offset = %d\n", section_number, offset);
				map_ptr->list_ptr[i + (j * 16)] = &map_ptr->list[offset][0];		
			}
			else
				map_ptr->list_ptr[i + (j * 16)] = NULL;	
		}	
	}			
	    
	printf("sections = %d : entries = %d : sections in use = %d\n", total_sections, total_entries, sections_in_use);

	for (i = 0; i < total_entries; i++)
	{		
		for (j = 0; j < 2; j++)
		{
    		byte[0] = *data;
			data++;
			byte[1] = *data;
			data++;		   						 
			result = (byte[1] << 8) + byte[0];
			map_ptr->list[i][j].vx = result;
//			map_ptr->list[i][j].vx = read_short(data);
//			data += 2;

			byte[0] = *data;
			data++;
			byte[1] = *data;
			data++;		   	
			result = (byte[1] << 8) + byte[0];
			map_ptr->list[i][j].vz = result;

//			map_ptr->list[i][j].vz = read_short(data);
//			data += 2;

//			if (j == 1)
//				printf("vx1 = %d : vz1 = %d\n", map_ptr->list[i][j].vx, map_ptr->list[i][j].vz); 
//			else
//				printf("vx0 = %d : vz0 = %d : ", map_ptr->list[i][j].vx, map_ptr->list[i][j].vz); 
		}

		if ((map_ptr->list[i][0].vx == -1) && (map_ptr->list[i][0].vz == -1) && 
			(map_ptr->list[i][1].vx == -1) && (map_ptr->list[i][1].vz == -1))
		{	
			printf("section %d = %d\n", section_no, section_ctr);
		
			section_no++;
			section_ctr = 0;
		}
		else
			section_ctr++;
	}

	for (i = 0; i < 16; i++)
		printf("list_ptr->x = %d : z = %d\n", map_ptr->list_ptr[i]->vx,  map_ptr->list_ptr[i]->vz);
	
	map_ptr->total = total_entries;
	map_ptr->current = map_ptr->list[0];

 	return;
}

void AI_init_player(OBJ_STRUCT *obj_ptr)
{
	AI_NAVIGATION 		*ptr;


	ptr = (AI_NAVIGATION *)init_AI_player_struct(obj_ptr);
	
	return;
}
/* start of backup values
	#define				MIN_ANGLE				32
	#define				MAX_ANGLE				512
	#define				MAX_VIEWING_DISTANCE	8000
//	#define				MAX_VIEWING_DISTANCE	6000
	#define				RANGE_DENOMONATOR		MAX_VIEWING_DISTANCE
	#define				CONE_SCALER				28
//	#define				CONE_SCALER				21
	#define				ABSOLUTE_ANGLE			0
	#define				RELATIVE_ANGLE			1
	#define				UPPER_LIMIT				150
	#define				LOWER_LIMIT				150
//	#define				UPPER_LIMIT				500
//	#define				LOWER_LIMIT				300
 // End of backup vals
*/ /*
	#define				MIN_ANGLE				32
	#define				MAX_ANGLE				768
	#define				MAX_VIEWING_DISTANCE	8000
//	#define				MAX_VIEWING_DISTANCE	6000
	#define				RANGE_DENOMONATOR		MAX_VIEWING_DISTANCE
	#define				CONE_SCALER				15
//	#define				CONE_SCALER				21
	#define				ABSOLUTE_ANGLE			0
	#define				RELATIVE_ANGLE			1
	#define				UPPER_LIMIT				400
	#define				LOWER_LIMIT				250
//	#define				UPPER_LIMIT				500
//	#define				LOWER_LIMIT				300
 // End of backup vals
	 */

char perform_collision_check(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	vec32				first_arg, second_arg;
	int					x_val, y_val, z_val, coll_val;
	int					collision_length, temp_length;
	short				rotation_x, rotation_y;


	collision_length = 750;

	rotation_x = plyr_ptr->pobj->r.vx >> 4;
	rotation_y = plyr_ptr->pobj->r.vy >> 4;

	y_val = (collision_length * rsin(rotation_x)) >> 12;
	temp_length = (collision_length * rcos(rotation_x)) >> 12;

	x_val = (temp_length * rsin(rotation_y)) >> 12;
	z_val = (temp_length * rcos(rotation_y)) >> 12;

    first_arg.vx = plyr_ptr->pobj->p.vx >> 8;
    first_arg.vy = plyr_ptr->pobj->p.vy >> 8;
    first_arg.vz = plyr_ptr->pobj->p.vz >> 8;

    second_arg.vx = first_arg.vx + x_val;
    second_arg.vy = first_arg.vy + y_val;
    second_arg.vz = first_arg.vz + z_val;

    coll_val = COL_test2(&first_arg, &second_arg); //Determines if the torpedo has hit the

    if (coll_val > -1)
    {
		x_val = COL_inter.vx - first_arg.vx;
		y_val = COL_inter.vy - first_arg.vy;
		z_val = COL_inter.vz - first_arg.vz;

		ptr->collision_distance = SquareRoot0((x_val * x_val) + (y_val * y_val) + (z_val * z_val));

		return TRUE;
	}

	return FALSE;
}

void AI_determine_navigation_state(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	int					i, velocity, x_val, z_val;
	char				state, is_stuck;
	short				angle_gap, angle;
	

	for (i = 14; i > 0; i--)
		ptr->velocity_store[i] = ptr->velocity_store[i - 1];

	angle_gap = vangdiff2(ptr->craft_orientation.vy, ptr->direction);
		
	velocity = (ptr->velocity * rcos(angle_gap)) >> 12;
	ptr->velocity_store[0] = velocity;
		
	state = stuck;

	for (i = 0; i < 15; i++)
	{
		if (ptr->velocity_store[i] > 25)
		{
			ptr->navigation_state = racing;

			state = racing;

			break;
		}
	}
	
	if (state == stuck)
	{
		is_stuck = perform_collision_check(ptr, plyr_ptr);

		if (is_stuck)
		{
			if (ptr->collision_distance < 400)
				ptr->navigation_state = stuck;
			else
			{
				state = racing;

				ptr->navigation_state = racing;
			}
		}				
		else
		{
			for (i = 0; i < 10; i++)
			{
				if (ptr->acc_pad_store[i] != PAD_RD)
				{
					ptr->navigation_state = racing;

					state = racing;

					break;
				}
			}

			if (state == stuck)
				ptr->navigation_state = out_of_bounds;
		}
	}

	if ((state == stuck) || (state == out_of_bounds))
	{
		if (STATE.arena_mode != TRUE)
		{
			angle_gap = vangdiff2(ptr->craft_orientation.vy, ptr->cutoff_angle);
			
			if (angle_gap <= 0)
				ptr->stuck_rotation = 127;
			else
				ptr->stuck_rotation = -128;
		}
		else
		{
			x_val = ptr->craft_position.vx;
			z_val = ptr->craft_position.vz;

			angle = ratan2(x_val, z_val);

			angle_gap = vangdiff2(angle, ptr->craft_orientation.vy);

			if (angle_gap <= 0)
				ptr->stuck_rotation = 127;
			else
				ptr->stuck_rotation = -128;
		}

		ptr->stuck_timer = 0;
	}

/*	if (STATE.arena_mode == TRUE)
	{
		if ((ptr->vector_length < 750) && (ptr->direction_length < 750) && (ptr->left_length < 750) && (ptr->right_length < 750))
		{
			ptr->navigation_state = turn_around;
					
			ptr->turn_around_timer = 10;

			if (ptr->rotation <= 0)
				ptr->stuck_rotation = -129;
			else
				ptr->stuck_rotation = 128;
		}	
	}
*/
	return;
}

void AI_pilot_simulator(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	#define				MIN_ANGLE				32
	#define				MAX_ANGLE				512
	#define				MAX_VIEWING_DISTANCE	8000
	#define				CONE_SCALER				35
	#define				ABSOLUTE_ANGLE			0
	#define				RELATIVE_ANGLE			1
	#define				UPPER_LIMIT				150
	#define				LOWER_LIMIT				50
	#define				PLAYER_RANGE			1024
	#define				TURN_RANGE				64
	#define				MAX_AGGRESSION			10
	#define				LEFT					0
	#define				RIGHT					1
	#define				TOTAL_PREV_ROTATIONS	4


	u8			 		text[50];
	char				temp, is_stuck;
	int					i;
	short				angle_gap;
		
	display_grid = TRUE;
	ptr->timer++;
	ptr->racing_timer ++;
	ptr->old_rotation = ptr->rotation;

	if (ptr->plyrnum == global_player_num)
		i = 0;

//	if (plyr_ptr->pickups.current != NULLVAL)
//	{
//		plyr_ptr->pickups.current = NULLVAL;
//		plyr_ptr->pickups.pickup_list[0] = NULLVAL;
//		plyr_ptr->pickups.quantity[0] = 0;
//	}

   if ((ptr->timer & 1)) 
		return;

	AI_calculate_range(ptr, plyr_ptr);
	AI_calculate_parameters(ptr, plyr_ptr->pobj);	

	PSX_c_pos.vx = ptr->craft_position.vx;
	PSX_c_pos.vy = 0;
	PSX_c_pos.vz = ptr->craft_position.vz;
	
	if ((ptr->racing_timer < 50) && (ptr->navigation_state == out_of_bounds))
		ptr->navigation_state = racing;

	switch (ptr->navigation_state)
	{
		case racing:
		
			AI_determine_navigation_state(ptr, plyr_ptr);

			AI_collision_avoidance(ptr);
			AI_direction_determination(ptr);
			AI_decision_making_machine(ptr);
			calibrate_rotation(ptr, plyr_ptr);
			
/*			if (ptr->plyrnum == global_player_num)
			{
				sprintf(text, "racing");

				TEXT_x = -100;
				TEXT_y = 50;
				TEXT_prstr_BIG(text,255,0, 0);
			}
*/							
			break;

		case wrong_way:
			break;

		case stuck:

			ptr->rotation = ptr->stuck_rotation;

			ptr->acceleration_pad_val = PAD_RL;
			
			is_stuck = perform_collision_check(ptr, plyr_ptr);

			if ((is_stuck == FALSE) || (ptr->collision_distance > 800))
			{
				ptr->racing_timer = 0;

				ptr->navigation_state = racing;
			}

/*			if (ptr->plyrnum == global_player_num)
			{
				sprintf(text, "stuck_beyond_belief");

				TEXT_x = -100;
				TEXT_y = 50;
				TEXT_prstr_BIG(text,255,0, 0);
			}
*/
			break;

		case out_of_bounds:

			ptr->rotation = ptr->stuck_rotation;

			ptr->acceleration_pad_val = PAD_RL;
			
			ptr->stuck_timer ++;

			if (ptr->stuck_timer == 20)
			{
				ptr->racing_timer = 0;

				ptr->navigation_state = racing;
			}

			if ((ptr->front_collision_status) || (ptr->left_collision_status) || (ptr->right_collision_status))
			{
				ptr->racing_timer = 0;

				ptr->navigation_state = racing;
			}

  			is_stuck = perform_collision_check(ptr, plyr_ptr);

			if ((is_stuck == TRUE) && (ptr->collision_distance < 400))
			{
				ptr->navigation_state = stuck;

				if (STATE.arena_mode != TRUE)
				{
					angle_gap = vangdiff2(ptr->craft_orientation.vy, ptr->cutoff_angle);
					
					if (angle_gap <= 0)
						ptr->stuck_rotation = 120;
					else
						ptr->stuck_rotation = -120;
				}
				else
					ptr->stuck_rotation = -120;
			}

/*			if (ptr->plyrnum == global_player_num)
			{
				sprintf(text, "out_of_bounds");

				TEXT_x = -100;
				TEXT_y = 50;
				TEXT_prstr_BIG(text,255,0, 0);
			}
*/
			break;

			case turn_around:

				if (ptr->velocity_store[0] > 0)
					ptr->acceleration_pad_val = PAD_RL;	

				if (ptr->velocity_store[0] < 0)
					ptr->acceleration_pad_val = PAD_RD;	
	
//				if ((ptr->velocity_store[0] < 50) && (ptr->velocity_store[0] > -50))
//					ptr->acceleration_pad_val = 0;	
	
				if (ptr->turn_rotation > 1750)
				{
					ptr->navigation_state = racing;
					
					for (i = 0; i < TOTAL_PREV_ROTATIONS; i++)
						ptr->old_rot[i] = 0;
					
					ptr->total_rotation = 0;
					ptr->average_rotation = 0;
				}
				else
					ptr->rotation = ptr->stuck_rotation;
				
				ptr->turn_rotation += abs(ptr->rotation);

				break;

	}

	AI_set_pad_values(ptr, &ptr->dcm, plyr_ptr);
	
//	if (ptr->plyrnum = 1)
//		printf("ptr->front_collision = %d : dir = %d\n", ptr->vector_length, ptr->direction_length);

//FOR DEBUGGING ONLY
	if (ptr->plyrnum == 2)
//	{
		AI_debug_function(ptr);
//		temp = ptr->dcm.combat_status;

//		if (ptr->dcm.navigation_status == target_pickup)
/*		switch (temp)
		{
			case avoid_player:

				sprintf(text, "mode: player_avoidance");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,255,0, 0);
				
				break;

			case static_object_avoidance:

				sprintf(text, "mode: static_object_avoidance");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,0, 255, 0);
				
				break;

			case active_weapons_avoidance:

				sprintf(text, "mode: active_weapons_avoidance");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,0,0, 255);
				
				break;

			case pickup_targeting:

				sprintf(text, "mode: pickup_targeting");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,255,0, 255);
				
				break;

			case player_targeting:

				sprintf(text, "mode: player_targeting");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,255,255, 0);
				
				break;

			case pathway_targeting:

				sprintf(text, "mode: pathway targeting");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,0,255, 255);
				
				break;

			default:
				break;
		}
*///	}			
//FOR DEBUGGING ONLY

	return;
}

void AI_set_pad_values(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	int								i;


	if (ptr->rotation < -128)
		ptr->analouge_pad_val = -128;

	if (ptr->rotation > 127)
		ptr->analouge_pad_val = 127; 
	  
	if ((ptr->rotation <= 127) && (ptr->rotation >= -128))
		ptr->analouge_pad_val = ptr->rotation;

	if (dcm_ptr->weapons_lock == TRUE)
		ptr->pickup_button = PAD_FRB;
	else
		ptr->pickup_button = FALSE;

	if (dcm_ptr->cannon_lock == TRUE)
		ptr->cannon_button = PAD_RR;
	else
		ptr->cannon_button = FALSE;

	plyr_ptr->SIM_input_data = (ptr->pickup_button | ptr->cannon_button | ptr->shoulder_pad_val | ptr->acceleration_pad_val);

	if ((ptr->rotation > 127) || (ptr->rotation < -128))
		plyr_ptr->SIM_input_data = (plyr_ptr->SIM_input_data | 2);

	plyr_ptr->SIM_input_push = ptr->pickup_button;
	plyr_ptr->SIM_input_id   = 7;
	plyr_ptr->SIM_input_an1x = ptr->analouge_pad_val;
	plyr_ptr->SIM_input_an1y = 0;
	plyr_ptr->SIM_input_an2x = 0;
	plyr_ptr->SIM_input_an2y = 0;

	for (i = 4; i > 0; i--)
		ptr->acc_pad_store[i] = ptr->acc_pad_store[i - 1];

	ptr->acc_pad_store[0] = ptr->acceleration_pad_val;

	return;
}

void AI_calculate_range(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	PAIR_2D					this_pt, next_pt;
	short					angle, temp_angle;
	int						x_val, z_val;


	if (STATE.arena_mode == FALSE)
	{
		next_pt.vx = RACE_points[plyr_ptr->rp2].vx;
		next_pt.vz = RACE_points[plyr_ptr->rp2].vz;

		this_pt.vx = RACE_points[plyr_ptr->rp1].vx;
		this_pt.vz = RACE_points[plyr_ptr->rp1].vz;

		x_val = next_pt.vx - this_pt.vx;
		z_val = next_pt.vz - this_pt.vz;

		angle = ratan2(x_val, z_val);
		
		ptr->cutoff_angle = angle;

//		temp_angle = vangdiff2(ptr->cutoff_angle, ptr->craft_orientation.vy);

//		if (temp_angle < -1024)
//			ptr->wrong_way_adj_angle = temp_angle;// + 1024;

//		if (temp_angle > 1024)
//			ptr->wrong_way_adj_angle = temp_angle;// - 1024;

//		if ((temp_angle > -1024) && (temp_angle < 1024))
		ptr->wrong_way_adj_angle = 0;
	}
	else
		ptr->wrong_way_adj_angle = 0;

	return;
}

void AI_decision_making_machine(AI_NAVIGATION *ptr)
{
	OBJ_STRUCT						*obj_ptr;
	OBJ_PLAYERSTRUCT  				*plyr_ptr;
	AI_DECISION_MAKING_MACHINE		*dcm_ptr;
	int								i;	
	
	u8						 		text[50];

	
	plyr_ptr = &OBJ_players[ptr->plyrnum];
	obj_ptr = plyr_ptr->pobj;
	dcm_ptr = &ptr->dcm;
	
	dcm_ptr->weapons_lock = FALSE;
	dcm_ptr->cannon_lock = FALSE;

	dcm_ptr->aquired_target.status = FALSE;

	if (dcm_ptr->combat_status == normal)
	{
		OBJ_players[OBJ_players[ptr->plyrnum].hunting].hunted = NULLVAL;
		OBJ_players[ptr->plyrnum].hunting = NULLVAL;
	}

	for (i = 0; i < STATE.wankers; i++)
	{
		if (i == OBJ_players[ptr->plyrnum].hunting)
			continue;

		if (OBJ_players[i].hunted == ptr->plyrnum)
			OBJ_players[i].hunted = NULLVAL;
	}

	set_weapons_range(dcm_ptr, &OBJ_players[ptr->plyrnum]);

//	if (dcm_ptr->combat_status == normal)
	  	dcm_ptr->players.status = scan_players(ptr, dcm_ptr);
//	else
//	{
		if (dcm_ptr->combat_status == fight)
		{
			if (dcm_ptr->aquired_target.distance > ptr->detection_length)
			{
				dcm_ptr->combat_status = normal;
				dcm_ptr->aquired_target.status = FALSE;
			}
		}
		else
		{
			if (dcm_ptr->players.distance > ptr->detection_length)
				dcm_ptr->combat_status = normal;
		}
//	}

	if ((plyr_ptr->pickups.current != NULLVAL) || (plyr_ptr->flamer_in_use == TRUE) || (plyr_ptr->plasma_in_use == TRUE))
		dcm_ptr->pickups.status = FALSE;
	else
		dcm_ptr->pickups.status = scan_pickups(ptr, dcm_ptr);

// 	dcm_ptr->static_objects.status = scan_static_objects(ptr, dcm_ptr);
// 	dcm_ptr->active_weapons.status = scan_active_weapons(ptr, dcm_ptr);
//	dcm_ptr->pathways.status = scan_pathways(ptr, dcm_ptr);
  											
	if (dcm_ptr->combat_status == normal)
		set_combat_status(ptr, dcm_ptr);

	switch (dcm_ptr->combat_status)
	{
		case fight:

			target_player(ptr, dcm_ptr);
//			target_pickup(ptr, dcm_ptr);

			if (ptr->plyrnum == camera_craft_num)
			{
				sprintf(text, "mode: fight");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,0, 255, 0);
			}
						
			break;
		
		case flee:
		
			player_avoidance(ptr, dcm_ptr, plyr_ptr);

			if (ptr->plyrnum == camera_craft_num)
			{
				sprintf(text, "mode: flee");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,0, 255, 0);
			}
						
			break;
			
		case normal:
		
			if (dcm_ptr->pickups.status == TRUE)
				target_pickup(ptr, dcm_ptr);

			if (ptr->navigation_state == racing)
				AI_velocity_determination(ptr, plyr_ptr);
						 	
			if (ptr->plyrnum == camera_craft_num)
			{
				sprintf(text, "mode: normal");

				TEXT_x = -100;
				TEXT_y = 100;
				TEXT_prstr(text,0, 255, 0);
			}
			
			use_utility_pickups(ptr, dcm_ptr, plyr_ptr);
						
			break;
			
		default:
			break;
	}

	return;
}

void use_utility_pickups(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	if ((plyr_ptr->pickups.current == pck_shield) ||
		(plyr_ptr->pickups.current == pck_cannon) ||
		(plyr_ptr->pickups.current == pck_missile) ||
		(plyr_ptr->pickups.current == pck_gas))
		
		dcm_ptr->weapons_lock = TRUE;

	if ((plyr_ptr->pickups.current == pck_turbo) && (ptr->vector_length > 4000))
		dcm_ptr->weapons_lock = TRUE;

	return;
}

void set_weapons_range(AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	if (STATE.arena_mode == TRUE)
	{
		dcm_ptr->weapons_range = 2048;
	}
	else
	{
		switch (plyr_ptr->pickups.current)
		{
			case pck_rocket:
			case pck_missile:
			case pck_torpedo:
			case pck_ion:
			case pck_plasma:
			case pck_flame:

				dcm_ptr->weapons_range = 512;

				break;
							
			case pck_cluster:
			case pck_mine:
			case pck_gas:

				dcm_ptr->weapons_range = 1536;

				break;

			default:

				dcm_ptr->weapons_range = (plyr_ptr->ammo / 10);

				if (dcm_ptr->weapons_range < 32)
					dcm_ptr->weapons_range = 32;

				break;
		}
	}

	return;
}

void AI_calculate_parameters(AI_NAVIGATION *ptr, OBJ_STRUCT *obj_ptr)
{
	int			  		x_val, z_val;
	char				collision_status;


	ptr->craft_orientation.vx = 0;
	ptr->craft_orientation.vy = obj_ptr->r.vy >> 4;
	ptr->craft_orientation.vz = 0;

    RotMatrix(&ptr->craft_orientation, &craft_matrix);

	ptr->craft_position.vx = obj_ptr->p.vx >> 8;
	ptr->craft_position.vz = obj_ptr->p.vz >> 8;

	ptr->direction = ratan2((obj_ptr->v.vx >> 8), (obj_ptr->v.vz >> 8));

	x_val = (obj_ptr->v.vx >> 8) * (obj_ptr->v.vx >> 8);
	z_val = (obj_ptr->v.vz >> 8) * (obj_ptr->v.vz >> 8);

	ptr->velocity = SquareRoot0(x_val + z_val);
	
	collision_status = AI_line_collision(ptr, ptr->direction, ABSOLUTE_ANGLE);

	if (collision_status == TRUE)
	{
		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		ptr->direction_length = SquareRoot0(x_val + z_val);
	}
	else
		ptr->direction_length = ptr->viewing_distance;

	return;
}

s8 AI_line_collision(AI_NAVIGATION *ptr, short angle, char angle_flag)
{
    PAIR_2D				vector;
	int					temp_x, temp_z;
	char				collision_status;
	SVECTOR				point1, point2;

	glb_ct ++ ;

	if (angle_flag)
	{
		angle += ptr->craft_orientation.vy;

		if (angle > 4096)
			angle -= 4096;

		if (angle < -4096)
			angle += 4096;

		vector.vx = (ptr->viewing_distance * rsin(angle)) >> 12;
		vector.vz = (ptr->viewing_distance * rcos(angle)) >> 12;
	}
	else
	{
		vector.vx = (ptr->viewing_distance * rsin(angle)) >> 12;
		vector.vz = (ptr->viewing_distance * rcos(angle)) >> 12;
	}

	temp_x = vector.vx + ptr->craft_position.vx;
	temp_z = vector.vz + ptr->craft_position.vz;

	if (temp_x < MAX_NEGATIVE_POSITION)
		vector.vx = MAX_NEGATIVE_POSITION;
	else
	{
		if (temp_x > MAX_POSITIVE_POSITION)
			vector.vx = MAX_POSITIVE_POSITION;
		else
			vector.vx += ptr->craft_position.vx;
	}
				
	if (temp_z < MAX_NEGATIVE_POSITION)
		vector.vz = MAX_NEGATIVE_POSITION;
	else
	{
		if (temp_z > MAX_POSITIVE_POSITION)
			vector.vz = MAX_POSITIVE_POSITION;
		else
			vector.vz += ptr->craft_position.vz;
	}
	
	end_vector.vx = vector.vx;
	end_vector.vz = vector.vz;
   
   	if (glb_ct == 3)
	{
		point1.vx = ptr->craft_position.vx;	
		point1.vy = 0;
		point1.vz = ptr->craft_position.vz;	

		point2.vx = end_vector.vx;	
		point2.vy = 0;
		point2.vz = end_vector.vz;	

//		display_vector(point1, point2, 0, 0, 0);

		point2.vx = (5000 * rsin(angle)) >> 12;
		point2.vz = (5000 * rcos(angle)) >> 12;
		point2.vy = 0;

		point2.vx += ptr->craft_position.vx;
		point2.vz += ptr->craft_position.vz;

//		display_vector(point1, point2, 255, 255, 255);
	}

//    collision_status = AI_check_point(&ptr->craft_position, &end_vector); 
    collision_status = raycast(&ptr->craft_position, &end_vector); 

	return collision_status;
}

void AI_collision_avoidance(AI_NAVIGATION *ptr)
{
	int					x_val, z_val, total_length;

	
	ptr->front_collision_status = AI_line_collision(ptr, 0, RELATIVE_ANGLE);
					
	if (ptr->front_collision_status == 1)
	{
		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		ptr->vector_length = SquareRoot0(x_val + z_val);
		
		if (ptr->vector_length > ptr->viewing_distance)
			FntPrint("VECTOR LENGTH FUCKED UP!!!!\n");
		
		if (!ptr->object_collision_flag)
		{
			total_length = (ptr->direction_length + ptr->vector_length) / 2;
			ptr->total_length = (ptr->direction_length / 2) + ptr->vector_length / 4;
		}
		else
		{
			total_length = (ptr->object_collision_length + ptr->vector_length) / 2;
			ptr->total_length = (ptr->object_collision_length / 2) + ptr->vector_length / 4;
		}

//		ptr->angle = (ptr->viewing_distance - total_length) / CONE_SCALER;

//		if (ptr->angle < MIN_ANGLE)
//			ptr->angle = MIN_ANGLE;
	}
	else
	{
		ptr->vector_length = ptr->viewing_distance;

//		ptr->angle = MIN_ANGLE;

		ptr->total_length = (ptr->direction_length / 2) + ptr->vector_length / 4;
	}

if (ptr->front_collision_status == 1)
{
//	ptr->angle = 128;
	
	ptr->left_collision_status = AI_line_collision(ptr, -ptr->angle, RELATIVE_ANGLE);

	if (ptr->left_collision_status == 1)
	{
		ptr->left_point.vx = AI_inter.vx;
		ptr->left_point.vz = AI_inter.vz;

		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		ptr->left_length = SquareRoot0(x_val + z_val);
	}
	else
	{
		ptr->left_point.vx = end_vector.vx;
		ptr->left_point.vz = end_vector.vz;

		x_val = (end_vector.vx - ptr->craft_position.vx) * (end_vector.vx - ptr->craft_position.vx);
		z_val = (end_vector.vz - ptr->craft_position.vz) * (end_vector.vz - ptr->craft_position.vz);

		ptr->left_length = ptr->viewing_distance;
	}

	ptr->right_collision_status = AI_line_collision(ptr, ptr->angle, RELATIVE_ANGLE);

	if (ptr->right_collision_status == 1)
	{
		ptr->right_point.vx = AI_inter.vx;
		ptr->right_point.vz = AI_inter.vz;

		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		ptr->right_length = SquareRoot0(x_val + z_val);
	}
	else
	{
		ptr->right_point.vx = end_vector.vx;
		ptr->right_point.vz = end_vector.vz;

		x_val = (end_vector.vx - ptr->craft_position.vx) * (end_vector.vx - ptr->craft_position.vx);
		z_val = (end_vector.vz - ptr->craft_position.vz) * (end_vector.vz - ptr->craft_position.vz);

		ptr->right_length = ptr->viewing_distance;
	}
}
else
{
	ptr->right_collision_status = FALSE;
	ptr->left_collision_status = FALSE;

	ptr->left_length = ptr->viewing_distance;
	ptr->right_length = ptr->viewing_distance;
}

//	printf("front = %d : left = %d : right = %d\n", ptr->vector_length, ptr->left_length, ptr->right_length);
	
	return;
}

void AI_direction_determination(AI_NAVIGATION *ptr)
{
	int					x_val, z_val;
	short				angle;


	if ((ptr->left_collision_status == TRUE) || (ptr->right_collision_status == TRUE))
	{
		x_val = ptr->right_point.vx - ptr->left_point.vx;
		z_val = ptr->right_point.vz - ptr->left_point.vz;
		
		angle = ratan2(x_val, z_val);

		ptr->impact_angle = vangdiff2(angle, ptr->craft_orientation.vy);

		if (ptr->impact_angle < 0)
			printf("NEGATIVE ROTATION\n");

		ptr->impact_angle -= 1024;
		alter_heading(ptr);
		ptr->impact_angle = 1024 - abs(ptr->impact_angle);

		ptr->range = ptr->angle - abs(ptr->rotation);
		ptr->heading = ptr->craft_orientation.vy + ptr->rotation;
	}
	else
	{
		ptr->heading = ptr->craft_orientation.vy;
		ptr->rotation = 0;
		ptr->range = ptr->angle;
		ptr->viewing_distance = MAX_VIEWING_DISTANCE;
		ptr->impact_angle = 0;
	}

	return;
}

void alter_heading(AI_NAVIGATION *ptr)
{
	int					impact_angle, rotation, distance;
	int					numerator, denomonator;
	OBJ_PLAYERSTRUCT	*plyr_ptr;
	int					i, rot_range, answer, answer2;
 

	plyr_ptr = &OBJ_players[ptr->plyrnum];

	if (ptr->plyrnum == global_player_num)
		i = 0;

	if (ptr->impact_angle < 0)
		impact_angle = -1024 - ptr->impact_angle;

	if (ptr->impact_angle > 0)
		impact_angle = 1024 - ptr->impact_angle;

	if (ptr->impact_angle == 0)
		return;

	if ((ptr->left_collision_status) && (ptr->right_collision_status))
	{
//		impact_angle *= 100;

		if (ptr->vector_length > 0)
			answer = ptr->vector_length / 100;
		else
			answer = 1;

		if (answer > 0)
			rotation = impact_angle / answer;
		else
			rotation = impact_angle;

		if (ptr->velocity)
			answer2 = ptr->velocity / 30;
		else
			answer2 = 0;

		rotation *= answer2;
					 
		ptr->rotation = -rotation;
//		printf("answer = %d\n", answer);

	//	impact_angle /= 100;
	}
	else
	{
		distance = ptr->viewing_distance - ptr->vector_length;
		numerator = -impact_angle * ptr->velocity;
		denomonator = 1024 * 255; //1100 was 1024

	//	rot_range = (global_rot_range * distance) / ptr->viewing_distance;
	//	rot_range = 128;
	//	printf("rot_range = %d\n", rot_range);
		rotation = 136 * numerator;
	//	rotation = ptr->angle * numerator;
		rotation = rotation / denomonator;

	   	rotation = (rotation * distance) / 2000;
		
		if (ptr->vector_length < 2000)
		   	rotation = (rotation * distance) / 4000;

		if (ptr->vector_length < 1000)
		   	rotation = (rotation * distance) / 2000;

		ptr->rotation = rotation;

//		if (STATE.arena_mode == FALSE)
//			ptr->rotation += ptr->wrong_way_adj_angle;
	}
		
	return;
}

void calibrate_rotation(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	int				ang_vel, range, rot_val, i, grip_value, collision_gap, angle_inc;
	int				total_rot, average_rot;
	short			ang_diff, temp_angle;

	
	if (ptr->plyrnum == global_player_num)
		i = 0;

   	ang_vel = plyr_ptr->wvect.vy;
	rot_val = ang_vel / scale_value;


	if (ptr->rotation < 0)
	{
		if (ptr->left_collision_status == FALSE)
		{
			range = ptr->range - abs(ptr->rotation);

			if (range > 0)
			{
				collision_gap = ptr->viewing_distance - ptr->right_length;
	 			angle_inc = (range * collision_gap) / ptr->viewing_distance;
	 
	 			ptr->rotation -= angle_inc;

//				printf("angle_inc = %d\n", angle_inc);
			}
 		}
 	}
 
 	if (ptr->rotation > 0)
 	{
 		if (ptr->right_collision_status == FALSE)
 		{
			range = ptr->range - abs(ptr->rotation);

			if (range > 0)
			{
	 			collision_gap = ptr->viewing_distance - ptr->left_length;
	 			angle_inc = (range * collision_gap) / ptr->viewing_distance;
	 
	 			ptr->rotation += angle_inc;

//				printf("angle_inc = %d\n", angle_inc);
			}
		}
	}

	if (rot_val >= 0)
	{
		if (ptr->rotation >= 0)
			ptr->rotation += abs(rot_val);
		else
			ptr->rotation += abs(rot_val);
	}
	else
	{
		if (ptr->rotation >= 0)
			ptr->rotation -= abs(rot_val);
		else
			ptr->rotation -= abs(rot_val);
	}
							   
	if (ptr->acceleration_pad_val == PAD_RD)
	{
		grip_value = plyr_ptr->latgrip - plyr_ptr->latslip;

		if (grip_value <= 0)
			ptr->acceleration_pad_val = 0;
		else
			ptr->acceleration_pad_val = PAD_RD;
	}
									
	if ((ptr->rotation > 127) || (ptr->rotation < -128))
	{
		if (ptr->velocity_store[0] > 0)
			ptr->acceleration_pad_val = PAD_RL;
		else
			ptr->acceleration_pad_val = PAD_RD;
	}

//	ptr->rotation = (ptr->old_rotation + ptr->rotation) / 2;

//	if (((ptr->old_rotation < 0) && (ptr->rotation > 0)) || ((ptr->old_rotation > 0) && (ptr->rotation < 0)))
//	{

	average_rot = ptr->rotation + ptr->old_rotation;
	total_rot = abs(ptr->rotation) + abs(ptr->old_rotation);

//	if ((total_rot > global_tot_rot) && (abs(average_rot) < global_ave_rot))
//	{
//		ptr->rotation = average_rot;

//		ptr->total_rotation += abs(ptr->rotation);
//		ptr->average_rotation += ptr->rotation;
//	}
//	printf("tot = %d : ave = %d : vec = %d\n", total_rot, average_rot, ptr->vector_length);
/*
	else
	{
		ptr->total_rotation = 0;
		ptr->average_rotation = 0;
	}

		if (ptr->navigation_state != turn_around)
		{
			if (STATE.arena_mode == TRUE)
			{

				if ((ptr->total_rotation > 500) && (abs(ptr->average_rotation) < 250))
				{
					if (ptr->vector_length < 1000)
					{
				//	printf("time to about turn!!! : total = %d : average = %d\n", ptr->total_rotation, ptr->average_rotation);
						ptr->navigation_state = turn_around;
						ptr->turn_rotation = 0;
							
						if (ptr->rotation <= 0)
							ptr->stuck_rotation = -129;
						else
							ptr->stuck_rotation = 128;
					}
				}	
				else
				{
					if ((ptr->total_rotation > 1000) && (abs(ptr->average_rotation) < 500))
						ptr->rotation = ptr->average_rotation;				
				}
			}
		}
//	}
	else
/*			
	if (ptr->navigation_state != turn_around)
	{
		if (STATE.arena_mode == TRUE)
		{
			if (ptr->vector_length < 2000)
			{
				ptr->total_rotation += abs(ptr->rotation);
				ptr->average_rotation += ptr->rotation;

				if ((ptr->total_rotation > 2000) && (abs(ptr->average_rotation) < 750))
				{
				//	printf("time to about turn!!! : total = %d : average = %d\n", ptr->total_rotation, ptr->average_rotation);
					ptr->navigation_state = turn_around;
					ptr->turn_rotation = 0;
							
					if (ptr->rotation <= 0)
						ptr->stuck_rotation = -129;
					else
						ptr->stuck_rotation = 128;
				}	
			}
			else
			{
				ptr->total_rotation = 0;
				ptr->average_rotation = 0;
			}
		}
	}
*/
	if (STATE.arena_mode == TRUE)
	{
		for (i = (TOTAL_PREV_ROTATIONS - 1); i > 0; i--)
			ptr->old_rot[i] = ptr->old_rot[i - 1];
		
		ptr->old_rot[0] = ptr->rotation;
		
		ptr->total_rotation = 
		ptr->average_rotation = 0;

		for (i = 0; i < TOTAL_PREV_ROTATIONS; i++)
		{
			ptr->total_rotation += abs(ptr->old_rot[i]);
			ptr->average_rotation += ptr->old_rot[i];	
		}

		if ((ptr->total_rotation > global_tot_rot) && (abs(ptr->average_rotation) < global_ave_rot)) 
		{
			if (ptr->vector_length < global_vector_length)
			{
				ptr->navigation_state = turn_around;
				ptr->turn_rotation = 0;
						
				if (ptr->rotation <= 0)
					ptr->stuck_rotation = -129;
				else
					ptr->stuck_rotation = 128;
			}
			else
				ptr->rotation = ptr->average_rotation;
		//	printf("tot = %d : ave = %d : vec = %d\n", ptr->total_rotation, ptr->average_rotation, ptr->vector_length);
		}
	}
/*	
	if (ptr->rotation > 0)
	{
		if (ptr->old_rotation < 0)
			ptr->rotation = (ptr->rotation + ptr->old_rotation) / 2;
	}
	
	if (ptr->rotation < 0)
	{
		if (ptr->old_rotation > 0)
			ptr->rotation = (ptr->rotation + ptr->old_rotation) / 2;
	}

	ptr->old_rotation = ptr->rotation;
*/		

	if (STATE.arena_mode == FALSE)
	{
		temp_angle = vangdiff2(ptr->cutoff_angle, ptr->craft_orientation.vy);

		if (temp_angle < -1024)
		{
			ang_diff = abs(temp_angle) - 1024;
			ptr->rotation -= ang_diff;
		}

		if (temp_angle > 1024)
		{
			ang_diff = temp_angle - 1024;
			ptr->rotation += ang_diff;
		}
	}

	return;
}

void AI_extreme_turn_determination(AI_NAVIGATION *ptr)
{
	if (ptr->rotation > 128)
		ptr->shoulder_pad_val = PAD_FRT;

	if (ptr->rotation < -128)
		ptr->shoulder_pad_val = PAD_FLT;
	
	if ((ptr->rotation >= -128) && (ptr->rotation <= 128))
		ptr->shoulder_pad_val = 0;
												
	return;
}

void AI_velocity_determination(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	int					vector_gap, direction_gap;
	int					grip_value;
	int					time_to_impact, final_val;

		
	vector_gap = 0;//(ptr->vector_length - ptr->old_vector_length);// / 2;
	direction_gap = ptr->direction_length - ptr->old_direction_length;
			
//	total_length = (ptr->direction_length);// + ptr->vector_length) / 2;
//	if ((ptr->player_collision_flag) && (ptr->direction_length > ptr->player_collision_length))
//		ptr->total_length = (ptr->direction_length / 2) + ptr->vector_length / 4;

	if ((ptr->velocity_store[0] < 5) || (ptr->total_length > 4000))
		ptr->acceleration_pad_val = PAD_RD;
	else
	{
		time_to_impact = (ptr->total_length * 100) / ptr->velocity_store[0];
		
		time_to_impact *= 100;
		
		if (ptr->impact_angle != 0)
		{
			final_val = time_to_impact / (ptr->impact_angle * 2);

			if (final_val < LOWER_LIMIT)
				ptr->acceleration_pad_val = PAD_RL;

			if (final_val > UPPER_LIMIT)
			{
				grip_value = plyr_ptr->latgrip - plyr_ptr->latslip;

				if (grip_value < 0)
					ptr->acceleration_pad_val = 0;
				else
					ptr->acceleration_pad_val = PAD_RD;
			}

			if ((final_val < UPPER_LIMIT) && (final_val > LOWER_LIMIT))
				ptr->acceleration_pad_val = 0;

			//printf("final_val = %d\n", final_val);
		}		
		else
			ptr->acceleration_pad_val = PAD_RD;
	}
	
	return;
}

void AI_debug_function(AI_NAVIGATION *ptr)
{
	SVECTOR				left_point, right_point, target;
	OBJ_STRUCT			*obj_ptr;
	int					i;


 	for (i = 0; i < MAX_AI_OBJECTS; i++)
	{
		if (!AI_objects[i].in_use)
			continue;
		
//		if (AI_objects[i].type != static_object)
//			continue;
//		if (AI_objects[i].type != pickup)
//			continue;
		if (AI_objects[i].type != multiple_path)
			continue;
		
		left_point.vx = AI_objects[i].vx;
		left_point.vy = 0;
		left_point.vz = AI_objects[i].vz - 500;//AI_objects[i].radius;

		right_point.vx = AI_objects[i].vx;
		right_point.vy = 0;
		right_point.vz = AI_objects[i].vz + 500;//AI_objects[i].radius;

//		display_vector(left_point, right_point, 0, 255, 0);

		left_point.vx = AI_objects[i].vx - 500;//AI_objects[i].radius;
		left_point.vy = 0;
		left_point.vz = AI_objects[i].vz;

		right_point.vx = AI_objects[i].vx + 500;//AI_objects[i].radius;
		right_point.vy = 0;
		right_point.vz = AI_objects[i].vz;

//		display_vector(left_point, right_point, 0, 255, 0);
	}

	target.vx = (1000 * rsin(ptr->left_range)) >> 12;
	target.vz = (1000 * rcos(ptr->left_range)) >> 12;
	target.vy = 0;

	target.vx += ptr->craft_position.vx;
	target.vz += ptr->craft_position.vz;

//	display_vector(PSX_c_pos, target, 0, 255, 255);

	target.vx = (1000 * rsin(ptr->right_range)) >> 12;
	target.vz = (1000 * rcos(ptr->right_range)) >> 12;
	target.vy = 0;

	target.vx += ptr->craft_position.vx;
	target.vz += ptr->craft_position.vz;

//	display_vector(PSX_c_pos, target, 0, 255, 255);

//	return;
//	if ((ptr->left_collision_status == 1) || (ptr->right_collision_status == 1))
//	{
		left_point.vx = ptr->left_point.vx;
		left_point.vy = 0;
		left_point.vz = ptr->left_point.vz;

		right_point.vx = ptr->right_point.vx;
		right_point.vy = 0;
		right_point.vz = ptr->right_point.vz;

		target.vx = ptr->target.vx;
		target.vy = 0;
		target.vz = ptr->target.vz;

		target.vx = (ptr->vector_length * rsin((ptr->craft_orientation.vy))) >> 12;
		target.vz = (ptr->vector_length * rcos((ptr->craft_orientation.vy))) >> 12;
		target.vy = 0;

		target.vx += ptr->craft_position.vx;
		target.vz += ptr->craft_position.vz;
						 
		display_vector(PSX_c_pos, target, 255, 0, 255);

		target.vx = (ptr->vector_length * rsin(ptr->heading)) >> 12;
		target.vz = (ptr->vector_length * rcos(ptr->heading)) >> 12;
		target.vy = 0;

		target.vx += ptr->craft_position.vx;
		target.vz += ptr->craft_position.vz;
	
		display_vector(PSX_c_pos, target, 255, 255, 255);

		target.vx = (ptr->vector_length * rsin((ptr->rotation + ptr->craft_orientation.vy))) >> 12;
		target.vz = (ptr->vector_length * rcos((ptr->rotation + ptr->craft_orientation.vy))) >> 12;
		target.vy = 0;

		target.vx += ptr->craft_position.vx;
		target.vz += ptr->craft_position.vz;
	
		display_vector(PSX_c_pos, target, 0, 255, 0);


//	if (ptr->front_collision_status)
//	{
//		if (ptr->vector_length > 4000)
//		{
			display_vector(PSX_c_pos, left_point, 255, 0, 0);
	 		display_vector(PSX_c_pos, right_point, 0, 0, 255);
			display_vector(left_point, right_point, 255, 255, 0);
//		}
//	}


//	}

	obj_ptr = &OBJ_objects[ptr->plyrnum];
/*
	if (ptr->index == 1)
	{
		for (i = 99; i > 0; i--)
		{
			AI_trail[i].vx = AI_trail[i - 1].vx;
			AI_trail[i].vy = AI_trail[i - 1].vy;
			AI_trail[i].vz = AI_trail[i - 1].vz;
		}

		AI_trail[0].vx = obj_ptr->p.vx >> 8;
		AI_trail[0].vy = 0;
		AI_trail[0].vz = obj_ptr->p.vz >> 8;

		if (AI_index_n < 100)
			AI_index_n++;

		for (i = 1; i < 100; i++)
		{
			if (i >= AI_index_n)
				break;

			display_vector(AI_trail[i - 1], AI_trail[i], 128, 255, 64);
		}
	}
*/
//	if (ptr->analouge_pad_val != 0)
		display_direction(ptr->analouge_pad_val);

	if (act_rply.playback == FALSE)
		action_replay_capture(ptr->analouge_pad_val);
	else
	{
		action_replay_playback();
	
		FntPrint("REPLAY MODE : index = %d\n", act_rply.count);

		act_rply.count ++;
	
		if (act_rply.count == 250)
			act_rply.playback = FALSE;
	}

	return;	
}

s8  raycast(PAIR_2D *start, PAIR_2D *finish)
{
   #define  BITSWORLDSIZE  16
   #define  BITSGRID       2
   #define  GRIDUNITLEN    ( 1<< (BITSWORLDSIZE-BITSGRID) )
   #define  SHIFT          (BITSWORLDSIZE-BITSGRID)

	s16 maxlen=30;
	s8 quad;
	s32   dp;

	s32   bx,bz;
	s32   gx,gz;
	s32   l0x,l0z;
	s32   l1x,l1z;
	s32   dx,dz;
	s32   dgx,dgz;
	s32   cx,cz;

	u8	  grid_section, result, z_pos, x_pos;

	l0x   =  (start->vx + 32768);
	l0z   =  (start->vz + 32768);
	l1x   =  (finish->vx + 32768);
	l1z   =  (finish->vz + 32768);

	if(l1x>l0x)
	{
		quad= -1;
		cx =  (l1x>>SHIFT)-(l0x>>SHIFT);
		dx =   (l1z-l0z)<<SHIFT;
		bx =  (l0x&~(GRIDUNITLEN-1))+GRIDUNITLEN;
		dgx= 1;
	}
	else
	{
		quad= 1;
		cx =  (l0x>>SHIFT)-(l1x>>SHIFT);
		dx =  -(l1z-l0z)<<SHIFT;
		bx =  (l0x&~(GRIDUNITLEN-1));
		dgx=-1;
	}

	if(l1z>l0z)
	{
		cz =  (l1z>>SHIFT)-(l0z>>SHIFT);
		dz =  -(l1x-l0x)<<SHIFT;
		bz =  (l0z&~(GRIDUNITLEN-1))+GRIDUNITLEN;
		dgz= 1;
	}
	else
	{//isign
		quad=-quad;
		cz =  (l0z>>SHIFT)-(l1z>>SHIFT);
		dz =  (l1x-l0x)<<SHIFT;
		bz =  (l0z&~(GRIDUNITLEN-1));
		dgz=-1;
	}

	gx =  l0x>>SHIFT;
	gz =  l0z>>SHIFT;

	dp =  (bx-l0x)*(l1z-l0z)-(bz-l0z)*(l1x-l0x);

	if (old_raceclock == STATE.raceclock)
		count ++;
	else
	{
		old_raceclock = STATE.raceclock;

		count = 0;
	}
	
//	global_distance = ptr->viewing_distance * 2;

	while( (cx>=0) && (cz>=0) && ( (maxlen--) >=0) )
	{
//		z_pos = (2 + gz);// * 4;
//		x_pos = (2 + gx) * 4;

		z_pos = ( gz);// * 4;
		x_pos = ( gx) * 4;

		grid_section = z_pos + x_pos;
//		printf("gx = %d : gz = %d : ", gx, gz);
	//	printf("grid_section = %d\n", grid_section);


//		if (count == 2)
//			printf("clock = %d\n", STATE.raceclock);

		result = test_AI_line(start, finish, grid_section);
		
//		if (result == TRUE)
//			final_result = TRUE;

		if (result == TRUE)
			return TRUE;

		if(sgn(dp) == quad)
		{
			cx--;
			dp+=dx;
			gx+=dgx;
		}
		else
		{
			cz--;
			dp+=dz;
			gz+=dgz;
		}
	}

//	if (final_result == TRUE)
//	if (result == TRUE)
//		return TRUE;
//	else
		return FALSE;
}

s8 test_AI_line(PAIR_2D *start, PAIR_2D *finish, u8 grid_section)
{
	PAIR_2D					first_arg, second_arg;
	s8						collision, intersection_flag;
	s32						distance, old_distance, x_sqr, z_sqr;
	unsigned int			cntr = 0;
	SVECTOR					arg1, arg2;
	int						section_ctr = 0;


	if (AI_course_map.list_ptr[grid_section] != NULL)				
		AI_course_map.current = AI_course_map.list_ptr[grid_section];		
	else
		return FALSE;
	
	first_arg.vx = 0;
	first_arg.vz = 0;

	second_arg.vx = 0;
	second_arg.vz = 0;

	intersection_flag = FALSE;
	old_distance = MAX_VIEWING_DISTANCE;
	
//	while ((first_arg.vx != NULLVAL) && (first_arg.vz != NULLVAL) &&
//			(second_arg.vx != NULLVAL) && (second_arg.vz != NULLVAL))
//	while ((first_arg.vx != NULLVAL) || (first_arg.vz != NULLVAL) ||
//			(second_arg.vx != NULLVAL) || (second_arg.vz != NULLVAL))
	while (1)
	{
		first_arg.vx = AI_course_map.current->vx;
		first_arg.vz = AI_course_map.current->vz;
		
		AI_course_map.current++;
				
		second_arg.vx = AI_course_map.current->vx;
		second_arg.vz = AI_course_map.current->vz;
					
		AI_course_map.current++;


//		if ((first_arg.vx == -14574) && (first_arg.vz == 27259) &&
//			(second_arg.vx == -16132) && (second_arg.vz == 27505) && (count == 2))
//				printf("hit our line\n");

		if ((first_arg.vx == NULLVAL) && (first_arg.vz == NULLVAL) &&
			(second_arg.vx == NULLVAL) && (second_arg.vz == NULLVAL))
		{
//			if (count == 1)
//				printf("cnt1 = %d : section = %d : ", cntr, grid_section);

//			if (count == 2)
//				printf("cnt2 = %d : section = %d : %d\n", cntr, grid_section, STATE.raceclock);
//			printf("grid_section = %d : lines checked = %d\n", grid_section, section_ctr);

			return intersection_flag;
		}
		else
			section_ctr++;

		arg1.vx = first_arg.vx;
		arg1.vy = 0;//first_arg.vy;
		arg1.vz = first_arg.vz;

		arg2.vx = second_arg.vx;
		arg2.vy = 0;//second_arg.vy;
		arg2.vz = second_arg.vz;

		collision = COL_2dline(start, finish, &first_arg, &second_arg);
		
		if (count == 1)
			cntr ++;

		if (count == 2)
			cntr ++;

//		if (count == 1)
//			display_vector(arg1, arg2, 0, 0, 255);

		if (collision == 1)
		{
			arg1.vx = first_arg.vx;
			arg1.vy = 0;//first_arg.vy;
			arg1.vz = first_arg.vz;

			arg2.vx = second_arg.vx;
			arg2.vy = 0;//second_arg.vy;
			arg2.vz = second_arg.vz;

//			printf("i = %d : index = %d : coll_pt_flag = %d\n", i, act_rply.count, coll_pt_flag);
//			printf("list:collision - x = %d : y = %d : z = %d\n", arg1.vx, arg1.vy, arg1.vz);
//			printf("list:finish - x = %d : y = %d : z = %d\n", arg2.vx, arg2.vy, arg2.vz);

//			if (count == 2)
//				display_vector(arg1, arg2, 0, 255, 0);

//			if (count == 1)
//				display_vector(arg1, arg2, 0, 255, 255);

			arg1.vx = start->vx;
			arg1.vy = 0;//start->vy;
			arg1.vz = start->vz;

			arg2.vx = finish->vx;
			arg2.vy = 0;//finish->vy;
			arg2.vz = finish->vz;

//			display_vector(arg1, arg2, 0, 255, 255);
	
//			printf("line:collision - x = %d : y = %d : z = %d\n", arg1.vx, arg1.vy, arg1.vz);
//			printf("line:finish - x = %d : y = %d : z = %d\n\n", arg2.vx, arg2.vy, arg2.vz);

			x_sqr = (COL_2dinter.vx - start->vx) * (COL_2dinter.vx - start->vx);
			z_sqr = (COL_2dinter.vz - start->vz) * (COL_2dinter.vz - start->vz);

			distance = SquareRoot0(x_sqr + z_sqr);
			

			if (distance < old_distance)	
			{
//				if (count == 1)
//					printf("x0 = %d : z0 = %d : x1 = %d : z1 = %d :: %d\n", first_arg.vx, first_arg.vz, second_arg.vx, second_arg.vz, STATE.raceclock);

				old_distance = distance;
				
				intersection_flag = TRUE;

				AI_inter.vx = COL_2dinter.vx;
				AI_inter.vz = COL_2dinter.vz;
			}
		}
	}
}

s8 AI_check_point(PAIR_2D *start, PAIR_2D *finish)
{
	s32					i; 
	s8					collision;
	PAIR_2D				first_arg, second_arg;
	SVECTOR				arg1, arg2;
	int					x_sqr, z_sqr, distance, old_distance;
	int					section_inc;
	SVECTOR				start_box, finish_box, point1, point2;
	short				counter = 0;
	char				intersection_flag = FALSE;
	int					check_lines = 0;


	old_distance = MAX_VIEWING_DISTANCE;
	AI_course_map.current = AI_course_map.list[0];		

	section_inc = 65536 / AI_GRID_WIDTH;

	start_box.vx = -32768 - section_inc;
	start_box.vz = -32768 - section_inc;

	start_box.vy = 0;
	finish_box.vy = 0;

	for (i = 0; i < MAX_AI_GRID_SECTIONS; i++)
	{
//		printf("start_box_x = %d : :finish_box_x = %d : start_z = %d : fin_z = %d\n", start_box.vx,  finish_box.vx,  start_box.vz,  finish_box.vz);

//		if (start_box.vz == 24576)
		if (start_box.vz == 16384)
		{
			start_box.vz = -32768;
			start_box.vx += section_inc;
		}
		else
			start_box.vz += section_inc;

//		if (start_box.vz == 24576)
		if (start_box.vz == 16384)
			finish_box.vz = 32767;
		else
			finish_box.vz = start_box.vz + section_inc;

//		if (start_box.vx == 24576)
		if (start_box.vx == 16384)
			finish_box.vx = 32767;
		else
			finish_box.vx = start_box.vx + section_inc;



/*
		if (start_box_x == 16384)
		{
			start_box_x = -32768;
			start_box_z += section_inc;
		}
		else
			start_box_x += section_inc;
*/
//		if (display_grid == TRUE)
//		{
//			printf("start_x = %d : finish_x = %d\n", start->vx, finish->vx);
//			printf("start_box_x = %d : finish_box_x = %d\n", start_box.vx, finish_box.vx);
//			printf("start_z = %d : finish_z = %d\n", start->vz, finish->vz);
//			printf("start_box_z = %d : finish_box_z = %d\n\n", start_box.vz, finish_box.vz);
//		}

 		if (finish->vx >= start->vx)
		{	
			if (((start_box.vx < start->vx) && (finish_box.vx < start->vx)))
				continue;

			if (((start_box.vx > finish->vx) && (finish_box.vx > finish->vx)))
				continue;
		}
		else
		{	
			if (((start_box.vx > start->vx) && (finish_box.vx > start->vx)))
				continue;

			if (((start_box.vx < finish->vx) && (finish_box.vx < finish->vx)))
				continue;
		}

		if (finish->vz >= start->vz)
		{	
			if (((start_box.vz < start->vz) && (finish_box.vz < start->vz)))
				continue;

			if (((start_box.vz > finish->vz) && (finish_box.vz > finish->vz)))
				continue;
		}
		else												
		{	
			if (((start_box.vz > start->vz) && (finish_box.vz > start->vz)))
				continue;

			if (((start_box.vz < finish->vz) && (finish_box.vz < finish->vz)))
				continue;
		}
			
		counter ++;

		if (display_grid == TRUE)
		{
			point1.vx = start_box.vx;
			point1.vz = start_box.vz;

			point2.vx = start_box.vx;
			point2.vz = finish_box.vz;
		
			point1.vy = 0;
			point2.vy = 0;

//			display_vector(point1, point2, 0, 255, 0);

			point1.vx = start_box.vx;
			point1.vz = start_box.vz;

			point2.vx = finish_box.vx;
			point2.vz = start_box.vz;
		
//			display_vector(point1, point2, 0, 255, 0);

			display_grid = FALSE;
		}

		first_arg.vx = 0;
		first_arg.vz = 0;

		second_arg.vx = 0;
		second_arg.vz = 0;

//		printf("grid_section_real = %d\n", i);
//		continue;

		if (AI_course_map.list_ptr[i] != NULL)				
			AI_course_map.current = AI_course_map.list_ptr[i];		
		else
			continue;
				
//		printf("i = %d\n", i);

		while ((first_arg.vx != NULLVAL) && (first_arg.vz != NULLVAL) &&
				(second_arg.vx != NULLVAL) && (second_arg.vz != NULLVAL))
		{
			check_lines ++;

			first_arg.vx = AI_course_map.current->vx;
			first_arg.vz = AI_course_map.current->vz;
			
			AI_course_map.current++;
					
			second_arg.vx = AI_course_map.current->vx;
			second_arg.vz = AI_course_map.current->vz;

			AI_course_map.current++;

			collision = COL_2dline(start, finish, &first_arg, &second_arg);
			
			arg1.vx = first_arg.vx;
			arg1.vy = 0;//first_arg.vy;
			arg1.vz = first_arg.vz;

			arg2.vx = second_arg.vx;
			arg2.vy = 0;//second_arg.vy;
			arg2.vz = second_arg.vz;

//			display_vector(arg1, arg2, 0, 255, 255);

			if ((collision == 1))// || (collision == 2))
			{
				arg1.vx = first_arg.vx;
				arg1.vy = 0;//first_arg.vy;
				arg1.vz = first_arg.vz;

				arg2.vx = second_arg.vx;
				arg2.vy = 0;//second_arg.vy;
				arg2.vz = second_arg.vz;

	//			printf("i = %d : index = %d : coll_pt_flag = %d\n", i, act_rply.count, coll_pt_flag);
	//			printf("list:collision - x = %d : y = %d : z = %d\n", arg1.vx, arg1.vy, arg1.vz);
	//			printf("list:finish - x = %d : y = %d : z = %d\n", arg2.vx, arg2.vy, arg2.vz);

				display_vector(arg1, arg2, 0, 255, 0);

				arg1.vx = start->vx;
				arg1.vy = 0;//start->vy;
				arg1.vz = start->vz;

				arg2.vx = finish->vx;
				arg2.vy = 0;//finish->vy;
				arg2.vz = finish->vz;

//				display_vector(arg1, arg2, 0, 255, 255);
		
	//			printf("line:collision - x = %d : y = %d : z = %d\n", arg1.vx, arg1.vy, arg1.vz);
	//			printf("line:finish - x = %d : y = %d : z = %d\n\n", arg2.vx, arg2.vy, arg2.vz);

//				x_sqr = (first_arg.vx - start->vx) * (first_arg.vx - start->vx);
//				z_sqr = (first_arg.vz - start->vz) * (first_arg.vz - start->vz);
				x_sqr = (COL_2dinter.vx - start->vx) * (COL_2dinter.vx - start->vx);
				z_sqr = (COL_2dinter.vz - start->vz) * (COL_2dinter.vz - start->vz);

				distance = SquareRoot0(x_sqr + z_sqr);
				
				if (distance < old_distance)	
				{
					old_distance = distance;
					
					intersection_flag = TRUE;

					AI_inter.vx = COL_2dinter.vx;
					AI_inter.vz = COL_2dinter.vz;
				}
			}
		}
 	}

/*
	for (i = 0; i < AI_course_map.total - 1; i++)
	{
		first_arg.vx = AI_course_map.current->vx;
		first_arg.vz = AI_course_map.current->vz;
		
		AI_course_map.current ++;

		if ((AI_course_map.current->vx == 0xFE) && (AI_course_map.current->vz == 0xFE))
			continue;

		second_arg.vx = AI_course_map.current->vx;
		second_arg.vz = AI_course_map.current->vz;
						
		collision = COL_2dline(start, finish, &first_arg, &second_arg);
		
		if ((collision == 1))// || (collision == 2))
		{
			arg1.vx = first_arg.vx;
			arg1.vy = 0;//first_arg.vy;
			arg1.vz = first_arg.vz;

			arg2.vx = second_arg.vx;
			arg2.vy = 0;//second_arg.vy;
			arg2.vz = second_arg.vz;

//			printf("i = %d : index = %d : coll_pt_flag = %d\n", i, act_rply.count, coll_pt_flag);
//			printf("list:collision - x = %d : y = %d : z = %d\n", arg1.vx, arg1.vy, arg1.vz);
//			printf("list:finish - x = %d : y = %d : z = %d\n", arg2.vx, arg2.vy, arg2.vz);

//				display_vector(arg1, arg2, 0, 255, 0);

			arg1.vx = start->vx;
			arg1.vy = 0;//start->vy;
			arg1.vz = start->vz;

			arg2.vx = finish->vx;
			arg2.vy = 0;//finish->vy;
			arg2.vz = finish->vz;
	
//			printf("line:collision - x = %d : y = %d : z = %d\n", arg1.vx, arg1.vy, arg1.vz);
//			printf("line:finish - x = %d : y = %d : z = %d\n\n", arg2.vx, arg2.vy, arg2.vz);

			x_sqr = (first_arg.vx - start->vx) * (first_arg.vx - start->vx);
			z_sqr = (first_arg.vz - start->vz) * (first_arg.vz - start->vz);

			distance = SquareRoot0(x_sqr + z_sqr);
			
			if ((distance < old_distance) || (old_distance == 0))	
			{
				old_distance = distance;

				AI_inter.vx = COL_2dinter.vx;
				AI_inter.vz = COL_2dinter.vz;
			}
		}
	}
*/
/*	
	if (display_grid == TRUE)
	{
		printf("check_lines = %d : ", check_lines);
		display_grid = FALSE;	
		printf("Timer = %d : counter = %d\n", STATE.raceclock, counter);
	}
*/
	if (intersection_flag == TRUE)
		return 1;
	else
		return 0;
}
/*	  
void AI_place_point(void)
{
//	AI_drawing_map[AI_index].vx = CAM_camp[CAM_c].vx;
//	AI_drawing_map[AI_index].vy = -100;//CAM_camp[CAM_c].vy;
//	AI_drawing_map[AI_index].vz = CAM_camp[CAM_c].vz;

	AI_index++;

	printf("{%d, 0, %d},\n", CAM_camp[CAM_c].vx, CAM_camp[CAM_c].vz);

	return;
}

void AI_remove_point(void)
{
	AI_index--;

	printf("\nRemove last line\n\n");
	
	return;
}

void AI_draw_map_on_screen(void)
{
	FntPrint("index = %d\n", AI_index);
	
	RENDER_crosshair();
	
//	for (i = 1; i < AI_index; i++)
//		display_vector(AI_drawing_map[i - 1], AI_drawing_map[i], 255, 255, 255);
										
	return;
}
*/
void action_replay_capture(s8 pad_value)
{
	int				i;

   	return;
	for (i = 0; i < MAXNUMPLAYERS; i++)
//	{
//		act_rply.pad_values[i][act_rply.replay_index] = pad_value;	

//	   	memcpy(&act_rply.players[i][act_rply.replay_index], &OBJ_players[i], sizeof(OBJ_PLAYERSTRUCT));
//	   	memcpy(&act_rply.objects[i][act_rply.replay_index], &OBJ_objects[i], sizeof(OBJ_STRUCT));
//		memcpy(&act_rply.pilot[i][act_rply.replay_index], &AI_pilot[i], sizeof(AI_NAVIGATION));
//	}
	   
   	act_rply.replay_index++;

	if (act_rply.replay_index == 250)
	{
		act_rply.full = TRUE;
		act_rply.replay_index = 0;
	}

	return;
}

s16 action_replay_playback(void)
{
	int i;

//	FntPrint("REPLAY MODE\n");
	
//	return_val = act_rply.pad_values[act_rply.replay_index];
//	for (i = 0; i <  STATE.wankers; i++)
//	{
//		memcpy(&OBJ_players[i] , &act_rply.players[i][act_rply.replay_index], sizeof(OBJ_PLAYERSTRUCT));
//		memcpy(&OBJ_objects[i], &act_rply.objects[i][act_rply.replay_index], sizeof(OBJ_STRUCT));
//		memcpy(&AI_pilot[i], &act_rply.pilot[i][act_rply.replay_index], sizeof(AI_NAVIGATION));
//	}

	act_rply.replay_index ++;	

	if (act_rply.replay_index == 250)
		act_rply.replay_index = 0;

	if (act_rply.replay_index == act_rply.finish_point)
	{
		act_rply.replay_index ++;
		act_rply.playback = FALSE;
	}

	return 1;//return_val;
}

void select_replay_mode(u32 pad_data)
{
	if (act_rply.full == TRUE)
	{
		if (pad_data & PAD_RU)
		{
//			if (act_rply.playback == FALSE)
//			{
				act_rply.playback = TRUE;

//		   		memcpy(&OBJ_players[1], &act_rply.players[act_rply.replay_index + 1], sizeof(OBJ_PLAYERSTRUCT));
//	   			memcpy(&OBJ_objects[1], &act_rply.objects[act_rply.replay_index + 1], sizeof(OBJ_STRUCT));
				 
				act_rply.replay_index -= act_rply.count;
				act_rply.count = 0;
				
				if (act_rply.replay_index < 0)
					act_rply.replay_index += 250;

				act_rply.finish_point = act_rply.replay_index - 1;
	
				if (act_rply.finish_point == -1)
					act_rply.finish_point = 254;
//			}
//			else
//			{
//		   		memcpy(&OBJ_players[1], &act_rply.players[act_rply.finish_point], sizeof(OBJ_PLAYERSTRUCT));
//	   			memcpy(&OBJ_objects[1], &act_rply.objects[act_rply.finish_point], sizeof(OBJ_STRUCT));

//				act_rply.count = 0;
//			}
		}
	}

	return;
}

void init_replay_data(void)
{
	REPLAY_STRUCT		*ptr;


	ptr = &act_rply;

	ptr->replay_index = 0;
	ptr->playback = FALSE;
	ptr->full = FALSE;
	ptr->finish_point = NULLVAL;

/*	for (i = 0; i < 250; i++)

		ptr->pad_values[0][i] = 0;
		ptr->pad_values[0][i] = 0;
		ptr->pad_values[0][i] = 0;
		ptr->pad_values[0][i] = 0;
		ptr->pad_values[0][i] = 0;
		ptr->pad_values[0][i] = 0;
		ptr->pad_values[0][i] = 0;
		ptr->pad_values[0][i] = 0;
*/
	return;
}

AI_NAVIGATION *init_AI_player_struct(OBJ_STRUCT *obj_ptr)
{
	int					i;
	AI_NAVIGATION		*ptr;


	ptr = &AI_pilot[obj_ptr->index];

	ptr->in_use = TRUE;
//	ptr->plyrnum = OBJ_players[obj_ptr->index].plyrnum;
	
	ptr->timer = ptr->plyrnum / 2;

	ptr->craft_position.vx = 0;
	ptr->craft_position.vz = 0;
		
	ptr->craft_orientation.vx = 0;
	ptr->craft_orientation.vy = 0;
	ptr->craft_orientation.vz = 0;

		
	ptr->left_point.vx = 0;
	ptr->left_point.vz = 0;
		
	ptr->right_point.vx = 0;
	ptr->right_point.vz = 0;
		
	ptr->target.vx = 0;
	ptr->target.vz = 0;
		
	ptr->direction_length = ptr->viewing_distance;
	ptr->old_direction_length = ptr->viewing_distance;

//	ptr->angle = 128 - (16 * ptr->plyrnum);//MIN_ANGLE;
	ptr->angle = 100;
	ptr->vector_length = ptr->viewing_distance;
	ptr->old_vector_length = ptr->viewing_distance;
	
	ptr->total_length = ptr->viewing_distance;

 	ptr->front_collision_status = FALSE;
 	ptr->left_collision_status = FALSE;
 	ptr->right_collision_status = FALSE;

 	ptr->rotation = 0;

 	ptr->analouge_pad_val = 0;          
 	                                    
 	ptr->shoulder_pad_val = 0;          
 	ptr->acceleration_pad_val = 0;      

// 	ptr->sliding_time = 0;              

	if (STATE.arena_mode == TRUE)
		ptr->dcm.pickup_range = 1024;
	else
		ptr->dcm.pickup_range = 256;

	ptr->dcm.aggression = ptr->plyrnum * 2;

	ptr->dcm.cannon_range = 64;
//	ptr->dcm.weapon_range = 512;

	ptr->dcm.grudge_mode = FALSE;
	ptr->dcm.grudge_id = NULLVAL;

	ptr->detection_length = 8000;
	ptr->dcm.cannon_shooting_range = ptr->detection_length;

	ptr->viewing_distance = MAX_VIEWING_DISTANCE;
	ptr->navigation_state = racing;
//	ptr->stuck_flag = FALSE;
	ptr->racing_timer = 0;

	ptr->dcm.combat_status = normal;
	
	for (i = 0; i < TOTAL_PREV_ROTATIONS; i++)
		ptr->old_rot[i] = 0;

	return &AI_pilot[obj_ptr->index];
}

void AI_init_all_players(void)
{
	int					i;


	for (i = 0; i <  STATE.wankers; i++)
	{
		AI_pilot[i].in_use = FALSE;
		AI_pilot[i].index = i;
		AI_pilot[i].plyrnum = i;
	  	AI_pilot[i].viewing_distance = MAX_VIEWING_DISTANCE;
		AI_pilot[i].navigation_state = racing;
		AI_pilot[i].racing_timer = 0;
		AI_pilot[i].dcm.combat_status = normal;
	}

//	AI_pilot[0].timer = 0;
	AI_pilot[1].timer = 0;
	AI_pilot[2].timer = 0;
	AI_pilot[3].timer = 0;
	AI_pilot[4].timer = 1;
	AI_pilot[5].timer = 1;
//	AI_pilot[6].timer = 1;
//	AI_pilot[7].timer = 1;

	return;
}

AI_COLLISION_OBJECT *AI_init_new_object(void)
{
	int						i;


	for (i = 0; i < MAX_AI_OBJECTS; i++)
	{
		if (AI_objects[i].in_use)
			continue;

		AI_objects[i].in_use = TRUE;

		return &AI_objects[i];
	}

	return NULL;
}

AI_DCM_PICKUP_OBJECT *AI_init_new_pickup(void)
{
	int						i;


	for (i = 0; i < MAXNUMPICKUPS; i++)
	{
		if (AI_pickups[i].in_use)
			continue;

		AI_pickups[i].in_use = TRUE;

		return &AI_pickups[i];
	}

	return NULL;
}

void AI_init_all_pickups(void)
{
	int					i;


	for (i = 0; i < MAXNUMPICKUPS; i++)
	{
		AI_pickups[i].in_use = FALSE;
		AI_pickups[i].index = i;
		AI_pickups[i].type = NULLVAL;
	}

	return;
}		

short read_short(u8 *ptr)
{
	short				result;
	unsigned char		byte[2];

	
	byte[0] = *ptr;
	ptr++;
	byte[1] = *ptr;
			   	
	result = (byte[1] << 8) + byte[0];

	return result;
}	

char scan_pickups(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	int							i, x_val, z_val, distance, closest_distance;
	short						angle, final_angle, rotation, object_id;
	AI_DCM_PICKUP_OBJECT 		*ai_ptr;
	AI_DCM_OBJECT				*dcm_obj_ptr;


 	
	if (ptr->plyrnum == global_player_num)
		i = 0;

 	
 	dcm_obj_ptr = &dcm_ptr->pickups;

	closest_distance = ptr->detection_length;
	final_angle = dcm_ptr->pickup_range;

	object_id = NULLVAL;
		
	for (i = 0; i < MAXNUMPICKUPS; i++)
	{
		if (AI_pickups[i].in_use == FALSE)
			continue;

		if (AI_pickups[i].type == NULLVAL)
			continue;

		ai_ptr = &AI_pickups[i];

		x_val = (ai_ptr->position.vx - ptr->craft_position.vx);
		z_val = (ai_ptr->position.vz - ptr->craft_position.vz);

		angle = ratan2(x_val, z_val);
		rotation = vangdiff2(ptr->craft_orientation.vy, angle);

		distance = SquareRoot0((x_val * x_val) + (z_val * z_val));
				
		if (abs(rotation) < final_angle)
		{
			if (distance < closest_distance)
			{
				closest_distance = distance;
				final_angle = rotation;
				object_id = i;
			}
		}		
	}
		
	if (object_id != NULLVAL)
	{
		dcm_obj_ptr->distance = closest_distance;
		dcm_obj_ptr->id = object_id;
		dcm_obj_ptr->angle = -final_angle;
	
		return TRUE;
	}
	else
	{
		dcm_obj_ptr->distance = NULLVAL;
		dcm_obj_ptr->id = NULLVAL;
		dcm_obj_ptr->angle = NULLVAL;
		
		return FALSE;
	}
}

char scan_players(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	int							i, x_val, z_val, distance, closest_distance;
	short						angle, final_angle, rotation;
	int							trgt_pos_x, trgt_pos_z, player_id;
	OBJ_PLAYERSTRUCT			*trgt_ptr;
	AI_DCM_OBJECT				*dcm_obj_ptr;


	if (ptr->plyrnum == global_player_num)
		i = 0;

	dcm_obj_ptr = &dcm_ptr->players;

	closest_distance = ptr->detection_length;

	player_id = NULLVAL;

	for (i = 0; i < STATE.wankers; i++)
	{
		if (i == ptr->plyrnum)
			continue;

		if ((OBJ_players[i].hunting != NULLVAL) || (OBJ_players[i].hunted != NULLVAL))
			continue;
//		if (i == 0)
//			continue;

		trgt_ptr = &OBJ_players[i];

		trgt_pos_x = trgt_ptr->pobj->p.vx >> 8;
		trgt_pos_z = trgt_ptr->pobj->p.vz >> 8;

		x_val = (trgt_pos_x - ptr->craft_position.vx);
		z_val = (trgt_pos_z - ptr->craft_position.vz);

		angle = ratan2(x_val, z_val);
//		rotation = vangdiff2(ptr->heading, angle);
		rotation = vangdiff2(ptr->craft_orientation.vy, angle);

		distance = SquareRoot0((x_val * x_val) + (z_val * z_val));

		if (distance > closest_distance)
			continue;
		else
		{
			if ((abs(rotation) < dcm_ptr->cannon_range) && (distance < dcm_ptr->cannon_shooting_range))
				dcm_ptr->cannon_lock = TRUE;

			if ((OBJ_players[ptr->plyrnum].pickups.current != pck_gas) || (OBJ_players[ptr->plyrnum].pickups.current != pck_mine) || (OBJ_players[ptr->plyrnum].pickups.current != pck_cluster))
			{
				if (abs(rotation) > dcm_ptr->weapons_range)
					continue;				   
				else
				{
					closest_distance = distance;
					final_angle = rotation;
					player_id = i;
				}
			}
			else
			{
				if (abs(rotation) < dcm_ptr->weapons_range)
					continue;
				else
				{
					closest_distance = distance;
					final_angle = rotation;
					player_id = i;
				}
			}
		}		
	}
		
	if (player_id != NULLVAL)
	{
		dcm_obj_ptr->distance = closest_distance;
		dcm_obj_ptr->id = player_id;
		dcm_obj_ptr->angle = final_angle;
	
		if (STATE.arena_mode == TRUE)
		{
			OBJ_players[ptr->plyrnum].hunting = player_id;

			OBJ_players[player_id].hunted = ptr->plyrnum;
		}

		return TRUE;
	}
	else
	{
		dcm_obj_ptr->distance = NULLVAL;
		dcm_obj_ptr->id = NULLVAL;
		dcm_obj_ptr->angle = NULLVAL;

		if (STATE.arena_mode == TRUE)
			OBJ_players[ptr->plyrnum].hunting = NULLVAL;
	
		return FALSE;
	}
}
/*
char scan_player_targets(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	int							i;
	int							x_val, z_val, distance, closest_distance;
	short						angle, closest_angle, rotation, closest_rotation;
	int							plyr_pos_x, plyr_pos_z, player_id;
	OBJ_PLAYERSTRUCT			*plyr_ptr, *my_plyr_ptr;
	AI_DCM_OBJECT				*dcm_obj_ptr, *target_ptr;

   	
   	if (ptr->plyrnum == 2)
   		return FALSE;
	
	dcm_obj_ptr = &dcm_ptr->players;
	closest_distance = ptr->viewing_distance;//ptr->detection_length;
	closest_rotation = PLAYER_RANGE;
	player_id = NULLVAL;

	my_plyr_ptr = &OBJ_players[ptr->plyrnum];

	for (i = 0; i <  STATE.wankers; i++)
	{
		if (i == ptr->plyrnum)
			continue;

		plyr_ptr = &OBJ_players[i];

		plyr_pos_x = plyr_ptr->pobj->p.vx >> 8;
		plyr_pos_z = plyr_ptr->pobj->p.vz >> 8;

		x_val = (plyr_pos_x - ptr->craft_position.vx) * (plyr_pos_x - ptr->craft_position.vx);
		z_val = (plyr_pos_z - ptr->craft_position.vz) * (plyr_pos_z - ptr->craft_position.vz);

		distance = SquareRoot0(x_val + z_val);

		x_val = (plyr_pos_x - ptr->craft_position.vx);
		z_val = (plyr_pos_z - ptr->craft_position.vz);

		angle = ratan2(x_val, z_val);
		
		switch (my_plyr_ptr->pickups.current)
		{
			case pck_rocket:
			case pck_torpedo:
 			case NULLVAL:
			case pck_ion:
			case pck_plasma:
			case pck_flame:
				
				rotation = vangdiff2(ptr->heading, angle);

				if (abs(rotation) < PLAYER_RANGE)
				{
					if (distance < closest_distance)
					{
						rotation = vangdiff2(dcm_obj_ptr->angle, ptr->craft_orientation.vy);
			
						if (abs(rotation < FIRING_RANGE))
						{
							target_ptr = &dcm_ptr->aquired_target;

							target_ptr->status = TRUE;
							target_ptr->distance = distance;
							target_ptr->angle = angle;
							target_ptr->id = i;
						}
						
						closest_distance = distance;
						closest_angle = angle;
						player_id = i;
					}
				}		
			
				break;

			case pck_cluster:
			case pck_gas:

				rotation = vangdiff2(ptr->heading, angle);

				if (abs(rotation) < PLAYER_RANGE)
				{
					if (distance < closest_distance)
					{
						rotation = vangdiff2(dcm_obj_ptr->angle, -ptr->craft_orientation.vy);
			
						if (abs(rotation < FIRING_RANGE))
						{
							target_ptr = &dcm_ptr->aquired_target;

							target_ptr->status = TRUE;
							target_ptr->distance = distance;
							target_ptr->angle = angle;
							target_ptr->id = i;
						}
						
						closest_distance = distance;
						closest_angle = angle;
						player_id = i;
					}
				}		
			
				break;

			case pck_mine:
			case pck_missile:
				
				if (my_plyr_ptr->stats.position <  STATE.wankers)
				{
					player_id = 0;
					closest_distance = 0;
					closest_angle = 0;
				}
				else
				{
					if (ptr->dcm.pickups.status)
					{
						player_id = 0;
						closest_distance = 0;
						closest_angle = 0;
					}
				}

				break;

			case pck_shield:
			case pck_flare:
			case pck_cannon:

				player_id = 0;
				closest_distance = 0;
				closest_angle = 0;

				break;

			case pck_turbo:
				
				if (!ptr->front_collision_status)
				{
					player_id = 0;
					closest_distance = 0;
					closest_angle = 0;
				}

				break;
	
			default:
				break;
		}
	}
		
	if (player_id != NULLVAL)
	{
		dcm_obj_ptr->distance = closest_distance;
		dcm_obj_ptr->id = player_id;
		dcm_obj_ptr->angle = closest_angle;
	
//		printf("target distance = %d\n", closest_distance);

		return TRUE;
	}
	else
	{
	
		dcm_obj_ptr->distance = NULLVAL;
		dcm_obj_ptr->id = NULLVAL;
		dcm_obj_ptr->angle = NULLVAL;

		return FALSE;
	}
}
*/
char scan_static_objects(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	AI_DCM_OBJECT				*dcm_obj_ptr;
	AI_COLLISION_OBJECT			*obj_ptr;
	int							i;
	long						closest_rotation, closest_distance;
	long						distance, rotation, angle, object_id;
	long						h_val, x_val, z_val, o_val, adjacent, range;


	if (STATE.currlevel != 0)
    	return FALSE;

	dcm_obj_ptr = &dcm_ptr->static_objects;

	closest_rotation = 96;
	closest_distance = ptr->viewing_distance;
	object_id = NULLVAL;

	for (i = 0; i < MAX_AI_OBJECTS; i++)
	{
		obj_ptr = &AI_objects[i];

		if (obj_ptr->in_use == FALSE)
			continue;

		if (obj_ptr->type != static_object)
			continue;

		x_val = (obj_ptr->vx - ptr->craft_position.vx) * (obj_ptr->vx - ptr->craft_position.vx);
		z_val = (obj_ptr->vz - ptr->craft_position.vz) * (obj_ptr->vz - ptr->craft_position.vz);

		distance = SquareRoot0(x_val + z_val);

		h_val = distance * distance;
		o_val = obj_ptr->radius * obj_ptr->radius;

		adjacent = SquareRoot0((h_val - o_val));

		range = ratan2(obj_ptr->radius, adjacent);

		x_val = (obj_ptr->vx - ptr->craft_position.vx);
		z_val = (obj_ptr->vz - ptr->craft_position.vz);

		angle = ratan2(x_val, z_val);

//		rotation = vangdiff2(ptr->heading, angle);
		rotation = vangdiff2(ptr->craft_orientation.vy, angle);
			
		if (abs(rotation) < (abs(closest_rotation) + abs(range)))
		{
			if (distance < closest_distance)
			{
				closest_distance = distance;
				closest_rotation = rotation;
				object_id = i;
			}
		}
	}
		
	if (object_id > NULLVAL)
	{
		dcm_obj_ptr->distance = closest_distance;
		dcm_obj_ptr->id = object_id;
		dcm_obj_ptr->angle = closest_rotation;

		ptr->object_collision_length = closest_distance;

		if (closest_distance < ptr->direction_length)
			ptr->object_collision_flag = TRUE;

		return TRUE;
	}
	else
	{
		ptr->object_collision_flag = FALSE;

		dcm_obj_ptr->distance = NULLVAL;
		dcm_obj_ptr->id = NULLVAL;
		dcm_obj_ptr->angle = NULLVAL;

		return FALSE;
	}
}

char scan_active_weapons(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	return FALSE;
}

char scan_pathways(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	int							i;
	int							x_val, z_val, distance, closest_distance;
	short						angle, closest_angle, rotation, closest_rotation, object_id;
	AI_COLLISION_OBJECT			*ai_ptr;
	AI_DCM_OBJECT				*dcm_obj_ptr;

 	return FALSE;

 	dcm_obj_ptr = &dcm_ptr->pathways;
	closest_distance = ptr->viewing_distance;
	closest_rotation = 256;
	object_id = NULLVAL;
		
	for (i = 0; i < MAX_AI_OBJECTS; i++)
	{
		if (AI_objects[i].in_use == FALSE)
			continue;

		if (AI_objects[i].type != multiple_path)
			continue;

		ai_ptr = &AI_objects[i];

		x_val = (ai_ptr->vx - ptr->craft_position.vx) * (ai_ptr->vx - ptr->craft_position.vx);
		z_val = (ai_ptr->vz - ptr->craft_position.vz) * (ai_ptr->vz - ptr->craft_position.vz);

		distance = SquareRoot0(x_val + z_val);

		x_val = (ai_ptr->vx - ptr->craft_position.vx);
		z_val = (ai_ptr->vz - ptr->craft_position.vz);

		angle = ratan2(x_val, z_val);
//		rotation = vangdiff2(ptr->heading, angle);
		rotation = vangdiff2(ptr->craft_orientation.vy, angle);

		if (abs(rotation) < 256)
		{
			if (distance < closest_distance)
			{
				closest_distance = distance;
				closest_angle = angle;
				object_id = i;
			}
		}		
	}
		
	if (object_id != NULLVAL)
	{
		dcm_obj_ptr->distance = closest_distance;
		dcm_obj_ptr->id = object_id;
		dcm_obj_ptr->angle = closest_angle;
	
		return TRUE;
	}
	else
	{
	
		dcm_obj_ptr->distance = NULLVAL;
		dcm_obj_ptr->id = NULLVAL;
		dcm_obj_ptr->angle = NULLVAL;

		return FALSE;
	}
}

void set_combat_status(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	AI_DCM_OBJECT				*dcm_obj_ptr;
	AI_TARGET					*dcm_target_ptr;
	OBJ_PLAYERSTRUCT			*my_ptr, *trgt_ptr;
	int							fight_score, flee_score;	
											

	if (STATE.arena_mode == TRUE)
	{
		if (dcm_ptr->players.status == FALSE)
		{
			dcm_ptr->combat_status = normal;
			dcm_ptr->grudge_mode = FALSE;

			return;
		}

		dcm_obj_ptr = &dcm_ptr->players;

		my_ptr = &OBJ_players[ptr->plyrnum];

		if (dcm_obj_ptr->id == dcm_ptr->grudge_id)
		{
			dcm_ptr->grudge_mode = TRUE;
			
			return;
		}
		else
			dcm_ptr->grudge_mode = FALSE;
				
		trgt_ptr = &OBJ_players[dcm_obj_ptr->id];

		fight_score = my_ptr->shield_strength / 10;
		fight_score += my_ptr->armour_level / 10;
		fight_score += my_ptr->ammo / 100;

		switch (my_ptr->pickups.current)
		{
			case pck_rocket:
			case pck_missile:
			case pck_torpedo:
			case pck_ion:
			case pck_plasma:
			case pck_flame:

				fight_score += 50;
				break;

			case pck_shield:

				fight_score += (500 - my_ptr->shield_strength) / 10;
				break;

			default:
				break;
		}
		
		flee_score = trgt_ptr->shield_strength / 10;
		flee_score += trgt_ptr->armour_level / 10;
		flee_score += trgt_ptr->ammo / 100;		

		if (AI_pilot[dcm_obj_ptr->id].dcm.combat_status == flee)
			fight_score += 50;
						
		if (AI_pilot[dcm_obj_ptr->id].dcm.combat_status == fight)
			flee_score += 50;
		
		fight_score += ((fight_score * dcm_ptr->aggression) / MAX_AGGRESSION);
		flee_score += ((flee_score * (MAX_AGGRESSION - dcm_ptr->aggression)) / MAX_AGGRESSION);

		if (fight_score > flee_score)
		{
			dcm_ptr->combat_status = fight;
			dcm_target_ptr = &dcm_ptr->aquired_target;

			dcm_target_ptr->id = dcm_ptr->players.id;
			dcm_target_ptr->distance = dcm_ptr->players.distance;
			dcm_target_ptr->rotation = dcm_ptr->players.angle;
		
			dcm_target_ptr->status =  TRUE;

			dcm_target_ptr->trail_idx = 0;
			dcm_target_ptr->trail_full = FALSE;
		}
		else
		{
			dcm_ptr->combat_status = flee;

			dcm_ptr->escape_direction = NULLVAL;
		}

//		if (dcm_obj_ptr->id == 0)
//			dcm_ptr->combat_status = flee;			
	}
	else
	{
		dcm_ptr->grudge_mode = FALSE;
		dcm_ptr->combat_status = normal;

		check_weapons_in_range(ptr, dcm_ptr);
	}
			
	if (STATE.arena_mode == TRUE)
	{
		if ((dcm_ptr->combat_status == normal) || (dcm_ptr->combat_status == flee))
		{
			if (OBJ_players[ptr->plyrnum].hunting != NULLVAL)
			{
				OBJ_players[OBJ_players[ptr->plyrnum].hunting].hunted = NULLVAL;
				OBJ_players[ptr->plyrnum].hunting = NULLVAL;
			}
		}
	}

	return;
}

void check_weapons_in_range(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	OBJ_PLAYERSTRUCT				*my_ptr;
	AI_DCM_OBJECT					*dcm_obj_ptr;


	my_ptr = &OBJ_players[ptr->plyrnum];

	if (my_ptr->pickups.current != NULLVAL)
	{
		switch (my_ptr->pickups.current)
		{
			case pck_rocket:
			case pck_missile:
			case pck_torpedo:
			case pck_ion:

				if (dcm_ptr->players.status == TRUE)
				{
					dcm_obj_ptr = &dcm_ptr->players;

					if (abs(dcm_obj_ptr->angle) < dcm_ptr->weapons_range)
						dcm_ptr->weapons_lock = TRUE;

				}

				break;

			case pck_flame:
			case pck_plasma:

				if (dcm_ptr->players.status == TRUE)
				{
					dcm_obj_ptr = &dcm_ptr->players;

					if ((abs(dcm_obj_ptr->angle) < dcm_ptr->weapons_range) && (dcm_obj_ptr->distance < 3000))
						dcm_ptr->weapons_lock = TRUE;

				}

				break;

			case pck_gas:
			case pck_cluster:

				if (dcm_ptr->players.status == TRUE)
				{
					dcm_obj_ptr = &dcm_ptr->players;

					if ((abs(dcm_obj_ptr->angle) > dcm_ptr->weapons_range) && (dcm_obj_ptr->distance < 3000))
					{
						if (dcm_obj_ptr->distance > 4000)
							dcm_ptr->weapons_lock = TRUE;
					}
				}

				break;

			case pck_mine:
			case pck_flare:
			case pck_cannon:
				
				dcm_ptr->weapons_lock = TRUE;

				break;
				
			case pck_turbo:

				if (ptr->total_length > 4000)
					dcm_ptr->weapons_lock = TRUE;

				break;
				
			case pck_shield:

				if (my_ptr->shield_strength < 500) // should be MAX_SHIELD_STRENGTH
					dcm_ptr->weapons_lock = TRUE;

				break;
				
			default:
				break;
		}
	}

	return;
}

void target_pickup(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	AI_DCM_OBJECT				*dcm_obj_ptr;
	short						rotation;
	char						collision_status;
	int							x_val, z_val, distance, scaler, i, range;
	u8							text[80];	

/*	if (ptr->plyrnum == camera_craft_num)
	{
		sprintf(text, "pickup targetting");

		TEXT_x = -100;
		TEXT_y = 80;
		TEXT_prstr(text,0, 255, 255);
	}
*/	
	if (ptr->plyrnum == global_player_num)
		i = 0;
					
	dcm_obj_ptr = &dcm_ptr->pickups;

//	rotation = ptr->heading + dcm_obj_ptr->angle;
	rotation = ptr->craft_orientation.vy + dcm_obj_ptr->angle;

	collision_status = AI_line_collision(ptr, rotation, ABSOLUTE_ANGLE);

	if (collision_status == TRUE)
	{
		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		distance = SquareRoot0(x_val + z_val);

		if (distance < dcm_obj_ptr->distance)
			return;
	}

//	scaler = dcm_obj_ptr->distance / 1500;

//	if (scaler > 0)
//		dcm_obj_ptr->angle /= scaler;
/*	range = 64;
	if (STATE.arena_mode == FALSE)
	{
		if (abs(dcm_obj_ptr->angle) <= (1 * range))
			ptr->rotation = dcm_obj_ptr->angle;
		else
		{
			if (dcm_obj_ptr->angle < -(1 * range))
				ptr->rotation -= (1 * range);

			if (dcm_obj_ptr->angle > (1 * range))
				ptr->rotation += (1 * range);
		}
	}
	else
*/	
	if (dcm_obj_ptr->angle < -128)
		ptr->rotation = -128;
	
	if (dcm_obj_ptr->angle > 127)
		ptr->rotation =  127;
	
	if ((dcm_obj_ptr->angle <= 127) && (dcm_obj_ptr->angle >= -128))
		ptr->rotation = dcm_obj_ptr->angle;
	
	return;
}

void target_player(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	AI_TARGET				*ai_trgt_ptr;
	OBJ_STRUCT				*obj_ptr, *pobj;
	OBJ_PLAYERSTRUCT		*plyr_ptr;
	PAIR_2D					target_position;
	int						target_velocity;
	int						offset_vector, x_val, z_val, distance;
	short					offset_angle, angle, rotation;
	char					collision_status;
	short					his_dir, my_dir, his_angle, my_angle;
	int						his_velocity, my_velocity, resultant_velocity;
	int						range, result, scaler, i;
			

	u8						 		text[50];


	if (ptr->plyrnum == global_player_num)
		i = 0;

	ai_trgt_ptr = &ptr->dcm.aquired_target;
	obj_ptr = &OBJ_objects[ai_trgt_ptr->id];
	plyr_ptr = &OBJ_players[ptr->plyrnum];
	pobj = plyr_ptr->pobj;

	target_position.vx = (obj_ptr->p.vx >> 8);
	target_position.vz = (obj_ptr->p.vz >> 8);

	x_val = target_position.vx - ptr->craft_position.vx;
	z_val = target_position.vz - ptr->craft_position.vz;
	
	angle = ratan2(x_val, z_val);

	offset_vector = -1000;
	offset_angle = angle;//(obj_ptr->r.vy >> 4);
	
	if (offset_angle >= 4096)
		offset_angle -= 4096;

	if (offset_angle <= -4096)
		offset_angle += 4096;

	x_val = (offset_vector * rsin(offset_angle)) >> 12;
	z_val = (offset_vector * rcos(offset_angle)) >> 12;
	
	target_position.vx += x_val;
	target_position.vz += z_val;

	x_val = target_position.vx - ptr->craft_position.vx;
	z_val = target_position.vz - ptr->craft_position.vz;
	
	angle = ratan2(x_val, z_val);

//	ai_trgt_ptr->rotation = vangdiff2(angle, ptr->heading);
	ai_trgt_ptr->rotation = vangdiff2(angle, ptr->craft_orientation.vy);
	ai_trgt_ptr->distance = SquareRoot0((x_val * x_val) + (z_val * z_val));

	if ((abs(ai_trgt_ptr->rotation) < dcm_ptr->cannon_range) && (ai_trgt_ptr->distance < dcm_ptr->cannon_shooting_range))
		dcm_ptr->cannon_lock = TRUE;
  
	collision_status = AI_line_collision(ptr, angle, ABSOLUTE_ANGLE);

	if (collision_status == TRUE)
	{
		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		distance = SquareRoot0(x_val + z_val);

		if (distance < ai_trgt_ptr->distance)
			ai_trgt_ptr->visible = FALSE;
		else
		{
			ai_trgt_ptr->tracking_idx = NULLVAL;
			ai_trgt_ptr->trail_full = FALSE;
			ai_trgt_ptr->trail_idx = 0;
			ai_trgt_ptr->visible = TRUE;
		}
	}
	else
		ai_trgt_ptr->visible = TRUE;

	if (ai_trgt_ptr->visible)
	{
/*		if (ptr->plyrnum == camera_craft_num)
		{
			sprintf(text, "visible");

			TEXT_x = 100;
			TEXT_y = 100;
			TEXT_prstr(text,0, 255, 255);
		}
*/
		ai_trgt_ptr->trail[ai_trgt_ptr->trail_idx].vx = target_position.vx;
		ai_trgt_ptr->trail[ai_trgt_ptr->trail_idx].vz = target_position.vz;
			
		ai_trgt_ptr->trail_idx ++;
		
		if (ai_trgt_ptr->trail_idx == 15)
		{
			ai_trgt_ptr->trail_full = TRUE;
			
			ai_trgt_ptr->trail_idx = 0;
		}

		x_val = (obj_ptr->v.vx >> 8) * (obj_ptr->v.vx >> 8);
		z_val = (obj_ptr->v.vz >> 8) * (obj_ptr->v.vz >> 8);
		
		target_velocity = SquareRoot0((x_val + z_val));

		x_val = obj_ptr->v.vx >> 8;
		z_val = obj_ptr->v.vz >> 8;

		his_dir = ratan2(x_val, z_val);

		x_val = pobj->v.vx >> 8;
		z_val = pobj->v.vz >> 8;

		my_dir = ratan2(x_val, z_val);

		his_angle = vangdiff2(his_dir, angle);
		my_angle = vangdiff2(my_dir, angle);
		
		his_velocity = (target_velocity * rcos(his_angle)) >> 12;
		my_velocity = (ptr->velocity * rcos(my_angle)) >> 12;

		resultant_velocity = my_velocity - his_velocity;

		if (resultant_velocity < 0)
			ptr->acceleration_pad_val = PAD_RD;
		else
		{
			if (resultant_velocity != 0)
				result = (ai_trgt_ptr->distance * 100) / resultant_velocity;
			else
				result = ai_trgt_ptr->distance;

			if (result < 3000)
				ptr->acceleration_pad_val = PAD_RL;

			if (result > 3000)
				ptr->acceleration_pad_val = PAD_RD;
		}
	//	if (ptr->rotation == 0)
	//		range = PLAYER_RANGE;
	//	else
			range = ptr->range;
/*
		scaler = ai_trgt_ptr->distance / 1500;

		if (scaler > 0)
			rotation = ai_trgt_ptr->rotation / scaler;
		else
*/			rotation = ai_trgt_ptr->rotation;

//		if (abs(rotation) <= (5 * range))
			ptr->rotation = rotation;
//		else
//		{
//			if (rotation < -(5 * range))
//				ptr->rotation -= (5 * range);

//			if (rotation > (5 * range))
//				ptr->rotation += (5 * range);
//		}

		dcm_ptr->players.id = ai_trgt_ptr->id;
		dcm_ptr->players.distance = ai_trgt_ptr->distance;
		dcm_ptr->players.angle = ai_trgt_ptr->rotation;

		check_weapons_in_range(ptr, dcm_ptr);
	}
	else
	{						 
		dcm_ptr->combat_status = normal;

		OBJ_players[OBJ_players[ptr->plyrnum].hunting].hunted = NULLVAL;
		OBJ_players[ptr->plyrnum].hunting = NULLVAL;

		return;
/*	
		if (ptr->plyrnum == camera_craft_num)
	 	{
			sprintf(text, "blind");

			TEXT_x = 100;
			TEXT_y = 100;
			TEXT_prstr(text,255, 255, 0);
		}
*/
		angle = set_target_line(ptr, ai_trgt_ptr);

//		rotation = vangdiff2(angle, ptr->heading);
		rotation = vangdiff2(angle, ptr->craft_orientation.vy);
	
		range = ptr->range;
		
		scaler = ai_trgt_ptr->distance / 1500;

		if (scaler > 0)
			rotation = rotation / scaler;

		if (abs(rotation) <= (5 * range))
			ptr->rotation += rotation;
		else
		{
			if (rotation < -(5 * range))
				ptr->rotation -= (5 * range);

			if (rotation > (5 * range))
				ptr->rotation += (5 * range);
		}

		if (abs(rotation) < 512)
			ptr->acceleration_pad_val = PAD_RD;

		if (abs(rotation) > 1024)
			ptr->acceleration_pad_val = PAD_RL;

		if ((abs(rotation) >= 512) && (abs(rotation) <= 1024))
			ptr->acceleration_pad_val = 0;
			
//		AI_velocity_determination(ptr, plyr_ptr);
	}

	return;
}

short set_target_line(AI_NAVIGATION *ptr, AI_TARGET *ai_trgt_ptr)
{
	int					x_val, z_val, distance, distance_to_point;
	short				angle, rotation;
	char				collision_status;


	if (ai_trgt_ptr->tracking_idx == NULLVAL)
	{
		if (ai_trgt_ptr->trail_full)
			ai_trgt_ptr->tracking_idx = ai_trgt_ptr->trail_idx + 1;
		else
			ai_trgt_ptr->tracking_idx = 0;
	}
		
	if (ai_trgt_ptr->tracking_idx == 14)
	{
		x_val = ai_trgt_ptr->trail[0].vx - ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx].vx;
		z_val = ai_trgt_ptr->trail[0].vz - ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx].vz;
	}
	else
	{
		x_val = ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx + 1].vx - ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx].vx;
		z_val = ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx + 1].vz - ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx].vz;
	}

	ai_trgt_ptr->cutoff_angle = ratan2(x_val, z_val);

	x_val = ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx].vx - ptr->craft_position.vx;
	z_val = ai_trgt_ptr->trail[ai_trgt_ptr->tracking_idx].vz - ptr->craft_position.vz;

	distance_to_point = SquareRoot0((x_val * x_val) + (z_val * z_val));
	angle = ratan2(x_val, z_val);

	collision_status = AI_line_collision(ptr, angle, ABSOLUTE_ANGLE);

	if (collision_status == TRUE)
	{
		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		distance = SquareRoot0(x_val + z_val);

		if (distance < distance_to_point)
			ai_trgt_ptr->tracking_idx++;
	}

	rotation = vangdiff2(ai_trgt_ptr->cutoff_angle, angle);

	if (abs(rotation) < 1024)
		ai_trgt_ptr->tracking_idx++;
	 
	if (ai_trgt_ptr->trail_full)
	{
		if (ai_trgt_ptr->tracking_idx == ai_trgt_ptr->trail_idx)
		{
			ai_trgt_ptr->tracking_idx = NULLVAL;
			ptr->dcm.combat_status = normal;		
		}
	}
	else
	{
		if (ai_trgt_ptr->tracking_idx == 15)
		{
			ai_trgt_ptr->tracking_idx = NULLVAL;
			ptr->dcm.combat_status = normal;		
		}
	}

	if (ai_trgt_ptr->tracking_idx == 15)
		ai_trgt_ptr->tracking_idx = 0;

	return angle;
}


/*
void target_player(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	AI_DCM_OBJECT			*dcm_obj_ptr;
	OBJ_STRUCT				*obj_ptr, *pobj;
	OBJ_PLAYERSTRUCT		*plyr_ptr;
	PAIR_2D					target_position;
	short					angle, rotation, range;
	int						offset_vector, offset_angle;
	int						x_val, z_val, target_velocity;
	int						result, i;
	int						distance, scaler;
	int						my_velocity, his_velocity, resultant_velocity;
	short					his_angle, my_angle, my_dir, his_dir, ang_vel, rot_val;
	char					collision_status;
	SVECTOR					point1, point2, left_point, right_point;	

   	if (ptr->plyrnum == 2)
   		return FALSE;

//GsClearVcount();
//	dcm_obj_ptr = &dcm_ptr->player_targets;
	obj_ptr = &OBJ_objects[dcm_obj_ptr->id];
	plyr_ptr = &OBJ_players[ptr->plyrnum];
	pobj = plyr_ptr->pobj;

	switch (plyr_ptr->pickups.current)
	{
//		case pck_rocket:
//		case NULLVAL:
//		case pck_mine:
		default:
			target_position.vx = (obj_ptr->p.vx >> 8);// + (obj_ptr->v.vx >> 8);
			target_position.vz = (obj_ptr->p.vz >> 8);// + (obj_ptr->v.vz >> 8);

			offset_vector = -1000;
			offset_angle = (obj_ptr->r.vy >> 4);// + 2048;
			
			if (offset_angle >= 4096)
				offset_angle -= 4096;

			if (offset_angle <= -4096)
				offset_angle += 4096;

			x_val = (offset_vector * rsin(offset_angle)) >> 12;
			z_val = (offset_vector * rcos(offset_angle)) >> 12;

//			target_position.vx += x_val;
//			target_position.vz += z_val;

			for (i = 99; i > 0; i--)
			{
				trail[i].vx = trail[i - 1].vx;
				trail[i].vy = trail[i - 1].vy;
				trail[i].vz = trail[i - 1].vz;
			}

			trail[0].vx = target_position.vx;
			trail[0].vy = 0;//obj_ptr->p.vy;
			trail[0].vz = target_position.vz;

			if (ian_index < 100)
				ian_index++;

			for (i = 1; i < 100; i++)
			{
				if (i >= ian_index)
					break;

				display_vector(trail[i - 1], trail[i], 255, 255, 255);
			}

			x_val = target_position.vx - ptr->craft_position.vx;
			z_val = target_position.vz - ptr->craft_position.vz;
			
			dcm_obj_ptr->distance = SquareRoot0((x_val * x_val) + (z_val * z_val));

			angle = ratan2(x_val, z_val);
			rotation = vangdiff2(angle, ptr->heading);

			collision_status = AI_line_collision(ptr, (rotation + ptr->heading), ABSOLUTE_ANGLE);

			if (collision_status == TRUE)
			{
				x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
				z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

				distance = SquareRoot0(x_val + z_val);

				if (distance < dcm_obj_ptr->distance)
				{
					if (rotation < 0)
						ptr->rotation -= ptr->range;
					
					if (rotation > 0)
						ptr->rotation += ptr->range;

					return;
				}	
			}

			x_val = (obj_ptr->v.vx >> 8) * (obj_ptr->v.vx >> 8);
			z_val = (obj_ptr->v.vz >> 8) * (obj_ptr->v.vz >> 8);
			
			target_velocity = SquareRoot0((x_val + z_val));

			x_val = target_position.vx - ptr->craft_position.vx;
			z_val = target_position.vz - ptr->craft_position.vz;
			
			angle = ratan2(x_val, z_val);

//			new_pos.vx = (-1000 * rsin(angle)) >> 12;
//			new_pos.vz = (-1000 * rcos(angle)) >> 12;

//			target_position.vx += new_pos.vx;
//			target_position.vz += new_pos.vz;

//			x_val = target_position.vx - ptr->craft_position.vx;
//			z_val = target_position.vz - ptr->craft_position.vz;

//			angle = ratan2(x_val, z_val);
			distance = SquareRoot0((x_val * x_val) + (z_val * z_val));
			
			if (distance > ptr->viewing_distance)
				return;

			x_val = obj_ptr->v.vx >> 8;
			z_val = obj_ptr->v.vz >> 8;

			his_dir = ratan2(x_val, z_val);

			x_val = pobj->v.vx >> 8;
			z_val = pobj->v.vz >> 8;

			my_dir = ratan2(x_val, z_val);

//			his_angle = vangdiff2(angle, his_dir);
//  		my_angle = vangdiff2(angle, my_dir);
			his_angle = vangdiff2(his_dir, angle);
  			my_angle = vangdiff2(my_dir, angle);
			
  		   //	my_angle = his_angle;
			
			his_velocity = (target_velocity * rcos(his_angle)) >> 12;
			my_velocity = (ptr->velocity * rcos(my_angle)) >> 12;

			resultant_velocity = my_velocity - his_velocity;

//			printf("his = %d : my = %d : res = %d\n", his_velocity, my_velocity, resultant_velocity);

			if (resultant_velocity < 0)
				ptr->acceleration_pad_val = PAD_RD;
			else
			{
				if (resultant_velocity < 0)
					ptr->acceleration_pad_val = 0;
				else
				{
					if (resultant_velocity != 0)
						result = (distance * 100) / resultant_velocity;
					else
						result = distance;

					printf("resl = %d\n", result);
					
//					if (result > 10000)
//						printf("why???\n");

					if (result < 3000)
						ptr->acceleration_pad_val = PAD_RL;

//					if (result < 500)
//						ptr->acceleration_pad_val = 0;
				
					if (result > 3000)
						ptr->acceleration_pad_val = PAD_RD;

	//				if ((result > 2750) && (result < 3000))		
	//					ptr->acceleration_pad_val = 0;
				}
			}
								
			if (resultant_velocity)
				scaler = dcm_obj_ptr->distance / (resultant_velocity * 16);
			else
				scaler = dcm_obj_ptr->distance / 2000;

//			if (scaler > 0)
//				rotation /= scaler;

//			new_craft_pos.vx = ptr->craft_position.vx + (plyr_ptr->pobj->v.vx >> 8);
//			new_craft_pos.vz = ptr->craft_position.vz + (plyr_ptr->pobj->v.vz >> 8);

//			x_val = target_position.vx - new_craft_pos.vx;
//			z_val = target_position.vz - new_craft_pos.vz;


			if (ptr->rotation == 0)
				range = PLAYER_RANGE;
			else
				range = ptr->range;

   			if (abs(rotation) <= (10 * range))
//				ptr->rotation = rotation;
				ptr->rotation += rotation;
			else
			{
				if (rotation < -(10 * range))
					ptr->rotation -= (10 * range);

				if (rotation > (10 * range))
					ptr->rotation += (10 * range);
			}

		ptr->rotation += (ptr->rotation / 3);
		
		ang_vel = plyr_ptr->wvect.vy;

		rot_val = ang_vel / 65;

		//printf("rot_val = %d : old_rot = %d", rot_val, ptr->rotation);

		if (rot_val >= 0)
		{
			if (ptr->rotation >= 0)
				ptr->rotation += abs(rot_val);
			else
				ptr->rotation += abs(rot_val);
		}
		else
		{
			if (ptr->rotation >= 0)
				ptr->rotation -= abs(rot_val);
			else
				ptr->rotation -= abs(rot_val);
		}

	//	printf(" : new_rot = %d\n", ptr->rotation);
								   
		ptr->rotation = (ptr->old_rotation + ptr->rotation) / 2;

		ptr->old_rotation = ptr->rotation;

		left_point.vx = ptr->craft_position.vx;
		left_point.vy = 0;
		left_point.vz = ptr->craft_position.vz - 500;//AI_objects[i].radius;

		right_point.vx = ptr->craft_position.vx;
		right_point.vy = 0;
		right_point.vz = ptr->craft_position.vz + 500;//AI_objects[i].radius;

		display_vector(left_point, right_point, 0, 255, 0);

		left_point.vx = ptr->craft_position.vx - 500;//AI_objects[i].radius;
		left_point.vy = 0;
		left_point.vz = ptr->craft_position.vz;

		right_point.vx = ptr->craft_position.vx + 500;//AI_objects[i].radius;
		right_point.vy = 0;
		right_point.vz = ptr->craft_position.vz;

		display_vector(left_point, right_point, 0, 255, 0);

			point1.vx = target_position.vx;
			point1.vz = target_position.vz;
			point1.vy = 0;

			point2.vx = ptr->craft_position.vx;
			point2.vz = ptr->craft_position.vz;
			point2.vy = 0;

//			display_vector(point1, point2, 0, 0, 0);
			
			break;
	}
//printf("cnt = %d\n", GsGetVcount());
	return;
}
*/
void object_avoidance(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	AI_DCM_OBJECT			*dcm_obj_ptr;
	int						angle_inc;
	SVECTOR					point1, point2;	
	

	dcm_obj_ptr = &dcm_ptr->static_objects;

	point1.vx = ptr->craft_position.vx;
	point1.vy = 0;
	point1.vz = ptr->craft_position.vz;


	point2.vx = AI_objects[dcm_obj_ptr->id].vx;
	point2.vy = 0;
	point2.vz = AI_objects[dcm_obj_ptr->id].vz;

//	display_vector(point1, point2, 255, 255, 255);

	angle_inc = TURN_RANGE * (ptr->viewing_distance - ptr->object_collision_length);
	angle_inc = angle_inc / ptr->viewing_distance;
												   
	if (dcm_obj_ptr->angle > 0)
		ptr->rotation += (angle_inc);
	else
		ptr->rotation -= (angle_inc);
					   
	point2.vx = (dcm_obj_ptr->distance * rsin(angle_inc + ptr->craft_orientation.vy + ptr->rotation)) >> 12;
	point2.vz = (dcm_obj_ptr->distance * rcos(angle_inc + ptr->craft_orientation.vy + ptr->rotation)) >> 12;
	point2.vy = 0;

	point2.vx += ptr->craft_position.vx;
	point2.vz += ptr->craft_position.vz;

//	display_vector(point1, point2, 0, 255, 0);

	return;
}

void player_avoidance(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr, OBJ_PLAYERSTRUCT *plyr_ptr)
{
	AI_DCM_OBJECT			*dcm_obj_ptr;
	OBJ_STRUCT				*trgt_ptr;
	int						x_val, z_val, distance, coll_distance, range;
	short					angle, rotation, new_rotation;
	char					collision_status,i;


	if (ptr->plyrnum == global_player_num)
		i = 0;

	dcm_obj_ptr = &dcm_ptr->players;	
	trgt_ptr = &OBJ_objects[dcm_obj_ptr->id];

	x_val = (trgt_ptr->p.vx >> 8) - ptr->craft_position.vx;
	z_val = (trgt_ptr->p.vz >> 8) - ptr->craft_position.vz;

	angle = ratan2(x_val, z_val);
//	rotation = vangdiff2(angle, ptr->heading);
  	rotation = vangdiff2(angle, ptr->craft_orientation.vy);

	distance = SquareRoot0((x_val * x_val) + (z_val * z_val));

	if (distance > ptr->detection_length)
	{
		dcm_ptr->combat_status = normal;

		OBJ_players[OBJ_players[ptr->plyrnum].hunting].hunted = NULLVAL;
		OBJ_players[ptr->plyrnum].hunting = NULLVAL;
	}
			
	collision_status = AI_line_collision(ptr, angle, ABSOLUTE_ANGLE);
		
	if (collision_status == TRUE)
	{
		x_val = (AI_inter.vx - ptr->craft_position.vx) * (AI_inter.vx - ptr->craft_position.vx);
		z_val = (AI_inter.vz - ptr->craft_position.vz) * (AI_inter.vz - ptr->craft_position.vz);

		coll_distance = SquareRoot0(x_val + z_val);

		if (coll_distance < distance)
		{
			dcm_ptr->combat_status = normal;
	
			OBJ_players[OBJ_players[ptr->plyrnum].hunting].hunted = NULLVAL;
			OBJ_players[ptr->plyrnum].hunting = NULLVAL;
		}
	}
				  
	if (abs(rotation) > 1024) //running away allready!
	{
		dcm_ptr->escape_direction = NULLVAL;

		if (dcm_ptr->pickups.status == TRUE)
			target_pickup(ptr, dcm_ptr);

		AI_velocity_determination(ptr, plyr_ptr);
	}
	else
	{
		if (dcm_ptr->escape_direction == NULLVAL)
		{
			if (rotation < 0)
				dcm_ptr->escape_direction = LEFT;
			else
				dcm_ptr->escape_direction = RIGHT;
		}

		if (ptr->front_collision_status == TRUE)
		{
			if (dcm_ptr->escape_direction == LEFT)
				ptr->rotation = -ptr->range;
			else
				ptr->rotation = ptr->range;
		}
		else
		{
			if (dcm_ptr->escape_direction == LEFT)
				ptr->rotation = -129;
			else
				ptr->rotation = 128;
		}
	}

	return;
}

void weapon_avoidance(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	return;
}

void target_pathway(AI_NAVIGATION *ptr, AI_DECISION_MAKING_MACHINE *dcm_ptr)
{
	return;
}

