//AI.H	
//By Ian Tomkins
//Started on 18.1.99
//Contains all headers etc. for performing computer AI.

#ifndef AI_H
#define AI_H
						    
//AI COLLISION 

#define					MAX_AI_OBJECTS			50
#define					MAX_AI_COLLISION_LISTS	6
#define					MAX_NEGATIVE_POSITION	-32768
#define					MAX_POSITIVE_POSITION	32767

enum {

	static_object, 
	pickup,
	weapon,
	multiple_path

};

typedef struct {

	vec16			position;
	
	int				index;

	signed char 	type;
	signed char 	in_use;

}AI_DCM_PICKUP_OBJECT;

typedef struct
{
	char				in_use;
	int					index;
	char				owner;
	int					type;

	short				vx;
	short				vz;

	short				radius;

}AI_COLLISION_OBJECT;

typedef struct
{
	PAIR_2D				list[MAX_AI_COLLISION_MAP_ENTRIES][2];
	PAIR_2D				*list_ptr[MAX_AI_GRID_SECTIONS];

	PAIR_2D				*current;

	int					total;

}AI_COLLISION_MAP;

extern	AI_COLLISION_OBJECT			AI_objects[MAX_AI_OBJECTS];
extern	AI_COLLISION_MAP			AI_course_map;
extern	AI_DCM_PICKUP_OBJECT		AI_pickups[MAXNUMPICKUPS];
extern	PAIR_2D						AI_inter;

void AI_init_collision_map(u8 *data);
s8 AI_check_point(PAIR_2D *start, PAIR_2D *finish);

int AI_init_collision_object(int x_pos, int z_pos, int rad, int type);
int AI_init_pickup(int x_pos, int y_pos, int z_pos, int type);

//END OF AI COLLISION 

//AI SETUP STUFF


void AI_init_data(u8 *data);
void AI_init_player(OBJ_STRUCT *obj_ptr);

//END OF AI SETUP STUFF

typedef struct {

	char		 		status;
	int					distance;
	int					old_distance;
	short				angle;
	int					id;

} AI_DCM_OBJECT;

//AI DCM 

typedef struct {

	char				status;
	int					distance;
	short				rotation;
	short				id;
	char				visible;

	PAIR_2D				trail[15];
	int					trail_idx;
	int					tracking_idx;
	char				trail_full;

	short				cutoff_angle;
	
}AI_TARGET;

typedef struct {

	AI_DCM_OBJECT		pickups;
	AI_DCM_OBJECT		players;
	AI_DCM_OBJECT		pathways;
	AI_DCM_OBJECT		static_objects;
	AI_DCM_OBJECT		active_weapons;

	AI_TARGET			aquired_target;
	
	char				combat_status;
	char				grudge_mode;
	char				grudge_id;

	char				weapons_lock;
	char				cannon_lock;

	short				pickup_range;
	short				cannon_range;
	short				weapons_range;

	int					cannon_shooting_range;
	int					escape_direction;

	int					aggression;

}AI_DECISION_MAKING_MACHINE;

enum {

	pickup_targeting,
	player_targeting,
	avoid_player,
	static_object_avoidance,
	active_weapons_avoidance,
	pathway_targeting

};

enum {

	fight,
	flee,
	normal
	
};

//END OF AI DCM

//AI FLIGHT CONTROL

typedef struct
{
	char				index;
	char				in_use;
	int					plyrnum;

	int					timer;

	int					viewing_distance;
	char				navigation_state;
//	char				prev_navigation_state;

	short				left_range;
	short				right_range;

	PAIR_2D				craft_position;
	SVECTOR				craft_orientation;

	char				front_collision_status;
	char				left_collision_status;
	char				right_collision_status;

	short				angle;
	int					vector_length;
	short				old_vector_length;

	PAIR_2D				left_point;
	PAIR_2D				right_point;
	PAIR_2D				target;

	short				left_length;
	short				right_length;

//	char				left_object;
//	char				right_object;

//	short				obj_rot_left;
//	short				obj_rot_right;

	short				direction;
	short				direction_length;
	short				old_direction_length;
	short				total_length;

	short				object_collision_length;
	short				object_collision_flag;

//	short				player_collision_length;
//	short				player_collision_flag;

	short				detection_length;
	short				heading;
		
	short				rotation;
	short				old_rotation;
	short				old_rot[15];
//	short				average_rotation;

	short				range;
	short				impact_angle;
//	short				sliding_time;

//	short				total_rot;
//	short				ave_rot;
//	short				occurances;

//	char				no_rotations;

	int					velocity;

	char				analouge_pad_val;

	unsigned int		shoulder_pad_val;
	unsigned int		acceleration_pad_val;
	unsigned int		cannon_button;
	unsigned int		pickup_button;

	unsigned int		acc_pad_store[15];
	int					velocity_store[15];

	short				cutoff_angle;
	
	short				stuck_timer;
	short				racing_timer;
	short				stuck_rotation;
//	char				stuck_flag;
	int					turn_rotation;

	int					total_rotation;
	int					average_rotation;

	AI_DECISION_MAKING_MACHINE		dcm;


//stuck stuff and wrong_way etc...

	int					collision_distance;
	short				wrong_way_adj_angle;

}AI_NAVIGATION;

enum {

	racing,
	wrong_way,
	stuck,
	out_of_bounds,
	turn_around

};

extern	AI_NAVIGATION				AI_pilot[MAXNUMPLAYERS];

void AI_pilot_simulator(AI_NAVIGATION *ptr, OBJ_PLAYERSTRUCT *plyr_ptr);

//END OF AI FLIGHT CONTROL

//FOR MAP DRAWING

//void AI_place_point(void);
//void AI_remove_point(void);
//void AI_draw_map_on_screen(void);

//END OF FOR DRAWING

//FOR TESTING 

typedef struct
{

//s8				pad_values[MAXNUMPLAYERS][250];
//OBJ_STRUCT		objects[MAXNUMPLAYERS][250];
//OBJ_PLAYERSTRUCT		players[MAXNUMPLAYERS][250];
//AI_NAVIGATION		pilot[MAXNUMPLAYERS][250];

s32				replay_index;
s32				count;
s32				finish_point;
s8				playback;
s8				full;

}REPLAY_STRUCT;

extern REPLAY_STRUCT	act_rply;

s16 action_replay_playback(void);
void action_replay_capture(s8 pad_value);
void select_replay_mode(u32 pad_data);
void init_replay_data(void);

//AI TEST MAP

//Lava 1 track
//extern PAIR_2D	AI_test_map_a[63];
//extern PAIR_2D	AI_test_map_b[89];
//extern PAIR_2D	AI_test_map_c[25];

//joels water 1

extern PAIR_2D	AI_test_map_a[175];
extern PAIR_2D	AI_test_map_b[190];
extern PAIR_2D	AI_test_map_c[8];
extern PAIR_2D	AI_test_map_d[13];
extern PAIR_2D 	AI_test_map_e[7];
extern PAIR_2D	AI_test_map_f[14];

// kams lava arena

//extern PAIR_2D	AI_test_map_a[145];
//extern PAIR_2D	AI_test_map_b[25];
//extern PAIR_2D	AI_test_map_c[28];
//extern PAIR_2D	AI_test_map_d[68];
//extern PAIR_2D	AI_test_map_e[26];
//extern PAIR_2D	AI_test_map_f[12];
extern PAIR_2D	AI_test_map_g[24];
extern PAIR_2D	AI_test_map_h[12];
extern PAIR_2D	AI_test_map_i[19];
extern PAIR_2D	AI_test_map_j[16];
extern PAIR_2D	AI_test_map_k[10];
extern PAIR_2D	AI_test_map_l[12];
extern PAIR_2D	AI_test_map_m[8];
extern PAIR_2D	AI_test_map_n[6];

extern int			global_player_num;
#endif


