#include <libmath.h>		
#include <memory.h>
#include "3d.h"
#include "psx_only.h"

extern int FntStream;

int Tmp;

int VectMag(VECTOR &V)
{
	VECTOR TmpV;
	TmpV.vx = V.vx;
	TmpV.vy = V.vy;
	TmpV.vz = V.vz;

	//N.B 16384 is the highest val (?)
	//that all 3 components of a vector can be to stop
	//overflow when getting mag.

	Tmp = 1;
	while( TmpV.vx>16384  || TmpV.vy>16384  || TmpV.vz>16384
		|| TmpV.vx<-16384 || TmpV.vy<-16384 || TmpV.vz<-16384 )
	{
		TmpV.vx >>= 1;
		TmpV.vy >>= 1;
		TmpV.vz >>= 1;

		Tmp *= 2;
	}

	return SD_SquareRootNF( (TmpV.vx*TmpV.vx) + (TmpV.vy*TmpV.vy) + (TmpV.vz*TmpV.vz) )
		   * Tmp;
}

int VectMag(SVECTOR &V)
{
	return SD_SquareRootNF( ( (int)V.vx * (int)V.vx )
						  + ( (int)V.vy * (int)V.vy )
						  + ( (int)V.vz * (int)V.vz ) );
}



int VectMagF(FIXED *V)
{
	return SD_SquareRootNF( ( (V[0]>>16) * (V[0]>>16) )
						  + ( (V[1]>>16) * (V[1]>>16) )
						  + ( (V[2]>>16) * (V[2]>>16) ) );
}


void VectUnit(VECTOR &Src)
{
	Tmp = SD_SquareRootNF(Src.vx*Src.vx + Src.vy*Src.vy + Src.vz*Src.vz);

	if(Tmp!=0)
	{
		Src.vx /= Tmp;
		Src.vy /= Tmp;
		Src.vz /= Tmp;
	}
	else
	{
		Src.vx = Src.vy = Src.vz = 0;
	}
}

void VectScale(VECTOR &Src, int Scale)
{
	Src.vx *= Scale;
	Src.vy *= Scale;
	Src.vz *= Scale;
}

void VectScale(SVECTOR &Src, int Scale)
{
	Src.vx *= Scale;
	Src.vy *= Scale;
	Src.vz *= Scale;
}

int VectDot(VECTOR &A, VECTOR &B)
{
	return (A.vx*B.vx) + (A.vy*B.vy) + (A.vz*B.vz);
}

int VectDotS(VECTOR &A, SVECTOR &B)
{
	return (A.vx*B.vx) + (A.vy*B.vy) + (A.vz*B.vz);
}

int VectDotSS(SVECTOR &A, SVECTOR &B)
{
	return (int)A.vx*(int)B.vx + (int)A.vy*(int)B.vy + (int)A.vz*(int)B.vz;
}

void VectCross(VECTOR &A, VECTOR &B, VECTOR &Dest)
{
//	VECTOR TmpVect;
//
//	TmpVect.vx = A.vy*B.vz - A.vz*B.vy;
//	TmpVect.vy = A.vz*B.vx - A.vx*B.vz;
//	TmpVect.vz = A.vx*B.vy - A.vy*B.vx;
//
//	Dest.vx = TmpVect.vx;
//	Dest.vy = TmpVect.vy;
//	Dest.vz = TmpVect.vz;
//

	Dest.vx = A.vy*B.vz - A.vz*B.vy;
	Dest.vy = A.vz*B.vx - A.vx*B.vz;
	Dest.vz = A.vx*B.vy - A.vy*B.vx;
}

void VectL2W(VECTOR &Src, GsCOORDINATE2 &Coord, VECTOR &Dest)
{
	SVECTOR sSrc;
	MATRIX  L2W;
	long flag;

	sSrc.vx = Src.vx;
	sSrc.vy = Src.vy;
	sSrc.vz = Src.vz;

	GsGetLw(&Coord, &L2W);

	SetRotMatrix(&L2W);
	SetTransMatrix(&L2W);
	RotTrans(&sSrc, &Dest, &flag);
}

void VectL2W(VECTOR *Src, int NumSrc, GsCOORDINATE2 &Coord, VECTOR *Dest)
{
	SVECTOR sSrc;
	MATRIX  L2W;
	long flag;

	GsGetLw(&Coord, &L2W);

	SetRotMatrix(&L2W);
	SetTransMatrix(&L2W);

	for(int i=0; i<NumSrc; i++)
	{
		sSrc.vx = Src[i].vx;
		sSrc.vy = Src[i].vy;
		sSrc.vz = Src[i].vz;

		RotTrans(&sSrc, &Dest[i], &flag);
	}
}


//function from JonC
void VectToMat(VECTOR *v, MATRIX *m)
{
	int	x, y, z, w, len, s;
	int	xs, ys, zs;
	int	wx, wy, wz;
	int	xx,	xy, xz;
	int	yy,	yz, zz;
	int	fsin, fcos;
	VECTOR sv;

	x = v->vx;
	y = v->vy;
	z = v->vz;

	sv.vx = x;
	sv.vy = y;
	sv.vz = z;
	Square0(&sv, &sv);
	len = SquareRoot0(sv.vx + sv.vy + sv.vz);

	if(len!=0)
	{
		x	=	( x<<14 )/len;   //x14
		y	=	( y<<14 )/len;   //y14
		z	=	( z<<14 )/len;   //z14

		w	=	len>>6;

		fsin	=	rsin( (w + 2048)>>1 )<<16;  //28
		fcos	=	rcos( (w + 2048)>>1 )<<2;   //14

		if( fcos ==0 )
		{
  			xs =	0; 	ys = 0;	zs = 0; // all 14
			w	=	0x4000;						//carefull!
		}
		else
		{
			w	=	fsin/fcos; //w14
			s	=	( ( fcos*fcos )>>2); //s24
			xs = ( (x>>6)*s  )>>14; 	ys = (  (y>>6)*s  )>>14;	zs = ( (z>>6)*s  )>>14;// all 14
		}

	   	wx = ( (w>>1)*xs )>>20; 	wy = ( (w>>1)*ys )>>20;	wz = ( (w>>1)*zs )>>20;
	   	xx = ( (x>>1)*xs )>>20; 	xy = ( (x>>1)*ys )>>20;	xz = ( (x>>1)*zs )>>20;
	   	yy = ( (y>>1)*ys )>>20; 	yz = ( (y>>1)*zs )>>20;	zz = ( (z>>1)*zs )>>20;
		// all 12

		m->m[0][0] = 4096 - (yy + zz);
		m->m[0][1] = xy - wz;
		m->m[0][2] = xz + wy;
	   	m->m[1][0] = xy + wz;
		m->m[1][1] = 4096 - (xx + zz);
		m->m[1][2] = yz - wx;
   		m->m[2][0] = xz - wy;
		m->m[2][1] = yz + wx;
		m->m[2][2] = 4096 - (xx + yy);
	}
	else
	{
		// zero len,  return unit mat
		m->m[0][0] = 4096;
		m->m[0][1] = 0;
		m->m[0][2] = 0;
	   	m->m[1][0] = 0;
		m->m[1][1] = 4096;
		m->m[1][2] = 0;
	   	m->m[2][0] = 0;
		m->m[2][1] = 0;
		m->m[2][2] = 4096;
	}
	// all done 12muls, 3divs(mulrecip?) , gtesqr, gtesqrt, rsin, rcos
}

void VectToMatF(FIXED *v, MATRIX *m)
{
	int	x, y, z, w, len, s;
	int	xs, ys, zs;
	int	wx, wy, wz;
	int	xx,	xy, xz;
	int	yy,	yz, zz;
	int	fsin, fcos;
	VECTOR sv;

//xxx?
	x = v[0]/1000;
	y = v[1]/1000;
	z = v[2]/1000;

	sv.vx = x;
	sv.vy = y;
	sv.vz = z;

	Square0(&sv, &sv);
	len = SquareRoot0(sv.vx + sv.vy + sv.vz);

	if(len!=0)
	{
		x	=	( x<<14 )/len;   //x14
		y	=	( y<<14 )/len;   //y14
		z	=	( z<<14 )/len;   //z14

		w	=	len>>6;

		fsin	=	rsin( (w + 2048)>>1 )<<16;  //28
		fcos	=	rcos( (w + 2048)>>1 )<<2;   //14

		if( fcos ==0 )
		{
  			xs =	0; 	ys = 0;	zs = 0; // all 14
			w	=	0x4000;						//carefull!
		}
		else
		{
			w	=	fsin/fcos; //w14
			s	=	( ( fcos*fcos )>>2); //s24
			xs = ( (x>>6)*s  )>>14; 	ys = (  (y>>6)*s  )>>14;	zs = ( (z>>6)*s  )>>14;// all 14
		}

	   	wx = ( (w>>1)*xs )>>20; 	wy = ( (w>>1)*ys )>>20;	wz = ( (w>>1)*zs )>>20;
	   	xx = ( (x>>1)*xs )>>20; 	xy = ( (x>>1)*ys )>>20;	xz = ( (x>>1)*zs )>>20;
	   	yy = ( (y>>1)*ys )>>20; 	yz = ( (y>>1)*zs )>>20;	zz = ( (z>>1)*zs )>>20;
		// all 12

		m->m[0][0] = 4096 - (yy + zz);
		m->m[0][1] = xy - wz;
		m->m[0][2] = xz + wy;
	   	m->m[1][0] = xy + wz;
		m->m[1][1] = 4096 - (xx + zz);
		m->m[1][2] = yz - wx;
   		m->m[2][0] = xz - wy;
		m->m[2][1] = yz + wx;
		m->m[2][2] = 4096 - (xx + yy);
	}
	else
	{
		// zero len,  return unit mat
		m->m[0][0] = 4096;
		m->m[0][1] = 0;
		m->m[0][2] = 0;
	   	m->m[1][0] = 0;
		m->m[1][1] = 4096;
		m->m[1][2] = 0;
	   	m->m[2][0] = 0;
		m->m[2][1] = 0;
		m->m[2][2] = 4096;
	}
	// all done 12muls, 3divs(mulrecip?) , gtesqr, gtesqrt, rsin, rcos
}

int VectCross2D(int x0, int y0, int x1, int y1)
{
	return( x0*y1 - x1*y0 );
}

void VectResize(VECTOR &Vect, int Scale)
{
	int Mag = VectMag(Vect);

	Vect.vx *= Scale;
	Vect.vy *= Scale;
	Vect.vz *= Scale;

	Vect.vx /= Mag;
	Vect.vy /= Mag;
	Vect.vz /= Mag;
}

//N.B line GA->GB is parallel to Z axis
//line AB's y value is ignored
int LineXZInterZ(VECTOR A, VECTOR B, VECTOR GA, VECTOR GB )
{
	//A and B on same side of Z line?
	if(A.vz<GA.vz)
		if(B.vz<GA.vz)
			return 0;
	if(A.vz>GA.vz)
		if(B.vz>GA.vz)
			return 0;


	int dx = B.vx - A.vx;
	int dz = B.vz - A.vz;

	//AB parallel to GAGB?
	if(dz==0)
		return 0;

	//at what X does the path of AB cross the Z line?
	int X = ( ((GA.vz-A.vz)*dx) / dz ) + A.vx;

	//is X in bounds of both lines
	if(X>=A.vx)
	{
		if(X<=B.vx)
		{
			if(X>=GA.vx)
			{
				if(X<=GB.vx)
					return 1;
			}
			else
			if(X>=GB.vx)
			{
				if(X<=GA.vx)
					return 1;
			}
		}
	}
	else
	if(X<=A.vx)
	{
		if(X>=B.vx)
		{
			if(X>=GA.vx)
			{
				if(X<=GB.vx)
					return 1;
			}
			else
			if(X>=GB.vx)
			{
				if(X<=GA.vx)
					return 1;
			}
		}
	}

	return 0;
}

//N.B line GA->GB is parallel to X axis
//line AB's y value is ignored
int LineXZInterX(VECTOR A, VECTOR B, VECTOR GA, VECTOR GB )
{
	//A and B on same side of X line?
	if(A.vx<GA.vx)
		if(B.vx<GA.vx)
			return 0;
	if(A.vx>GA.vx)
		if(B.vx>GA.vx)
			return 0;


	int dx = B.vx - A.vx;
	int dz = B.vz - A.vz;

	//AB parallel to GAGB?
	if(dx==0)
		return 0;

	//at what Z does the path of AB cross the X line?
	int Z = ( ((GA.vx-A.vx)*dz) / dx ) + A.vz;

	//is Z in bounds of both lines
	if(Z>=A.vz)
	{
		if(Z<=B.vz)
		{
			if(Z>=GA.vz)
			{
				if(Z<=GB.vz)
					return 1;
			}
			else
			if(Z>=GB.vz)
			{
				if(Z<=GA.vz)
					return 1;
			}
		}
	}
	else
	if(Z<=A.vz)
	{
		if(Z>=B.vz)
		{
			if(Z>=GA.vz)
			{
				if(Z<=GB.vz)
					return 1;
			}
			else
			if(Z>=GB.vz)
			{
				if(Z<=GA.vz)
					return 1;
			}
		}
	}

	return 0;
}

// very quick 3d dist function 10% error
Sint32	vqlen(Sint32 ix,Sint32 iy,Sint32 iz)
{
Sint32	t;

  ix= (ix<0 ? -ix : ix);	// absolute values
  iy= (iy<0 ? -iy : iy);
  iz= (iz<0 ? -iz : iz);

  if(ix<iy)			// needs only two comparisons
   {
   ix=ix^iy;
   iy=ix^iy;
   ix=ix^iy;
  	}

  if(ix<iz)
   {
   ix=ix^iz;
   iz=ix^iz;
   ix=ix^iz;
  	}
  // now ix is the largest

  t = iy + iz;

  return (ix - (ix>>4) + (t>>2) + (t>>3));
}


// very quick 2d dist function 10% error
Sint32	vqlen2( Sint32 ix, Sint32 iy )
{
Sint32 t;

   ix= (ix<0 ? -ix : ix);   /* absolute values */
   iy= (iy<0 ? -iy : iy);

   if(ix<iy)       /* swap ix and iy if (ix < iy) */
      {            /* See Wyvill (G1, 436)        */
      ix=ix^iy;
      iy=ix^iy;
      ix=ix^iy;
      }

   t = iy + (iy>>1);

   return (ix - (ix>>5) - (ix>>7)  + (t>>2) + (t>>6));
}