//PSX specific code ONLY. This file holds all code for functions that
//perform tasks for the PlayStation only. There will be an equivalent file
//with the same functions for thew Nintendo version aswell.


#include "..\header.h"
#include "..\objects.h"
#include "..\pickups.h"
#include "..\weapons.h"
#include "..\render.h"
#include "..\camera.h"
#include "..\state.h"
#include "..\collide.h"

#include <sys/types.h>
#include <libetc.h>
#include <libgte.h>
#include <libgpu.h>
#include <libgs.h>
#include <libsn.h>
#include <gtemac.h>
#include <inline_c.h>
#include <ctrller.h>
#include <libmath.h>
#include <rand.h>

#include "PSX_pcks.h"
#include "PSX_expn.h"
#include "MATHPSX.h"
#include "..\sdlib\sdfunc.h"
#include "soundPSX.h"//jof
#include "gpupsx.h"


// craft on fire data table

enum {

	release_val,
	delay_val,
	rgb_val,
	inc_val,
	tp_val,
	cl_val

};

int PSX_craft_flames[7][6] = {

//Parameters are:
//
//		release rate, delay, rgb, increment, tpage, clut.

{1, 2, 96, 1, getTPage(0, 1, 704, 256), getClut(704, 462)}, // armour level < 200
{1, 2, 72, 1, getTPage(0, 1, 704, 256), getClut(704, 462)}, // armour level < 175
{1, 3, 48, 2, getTPage(0, 1, 704, 256), getClut(704, 462)}, // armour level < 150
{1, 3, 164, 2, getTPage(0, 3, 704, 256), getClut(704, 462)}, // armour level < 125
{1, 3, 128, 2, getTPage(0, 3, 704, 256), getClut(704, 462)}, // armour level < 100
{1, 4, 96, 3, getTPage(0, 3, 704, 256), getClut(704, 462)}, // armour level < 75
{1, 4, 64, 3, getTPage(0, 3, 704, 256), getClut(704, 462)} // armour level < 50

};

//NEW EXPLOSION CODE
void calculate_camera_matrix(int display_loop);

void update_flame_ctrl(FLAME_CTRL *ptr);
void init_flame(FLAME_CTRL *ptr);
void update_flame(FLAME_ANIMATION *ptr);
void display_flame_animations(FLAME_ANIMATION *ptr[3], int no_flames, int display_loop, GsOT *ot_ptr);

FLAME_CTRL *create_flame_ctrl(int type);
void delete_flame_ctrl(FLAME_CTRL *ptr);

FLAME_ANIMATION *create_flame_animation(void);
void delete_flame_animation(FLAME_ANIMATION *ptr);

void set_flame_attributes(FLAME_ANIMATION *flame, FLAME_CTRL *ptr, OBJ_STRUCT *obj_ptr, int collision_typr);
void set_flame_ctrl_attributes(FLAME_CTRL *ptr, OBJ_STRUCT *obj_ptr);

//global variables
FLAME_CTRL				PSX_flame_ctrls[MAX_EXPLOSIONS];
FLAME_ANIMATION			PSX_flames[MAX_EXP_FLAMES];

FLAME_CTRL *flame_ctrl_start;
FLAME_CTRL *flame_ctrl_current;

FLAME_ANIMATION *flame_animation_start;
FLAME_ANIMATION *flame_animation_current;

int flame_ctrl_count;

MATRIX camera_matrix;
SVECTOR	camera_position;

//END OF NEW EXPLOSION CODE

vec16 start_finish_pt[16][2] = {

	{{21485, 700, 26477},{20959, 700, 24432}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},
	{{0, 0, 0},{0, 0, 0}},

};

extern s32 power_up_table[10][5];

   int exp_timer_on;
int exp_timer_off;

s32	muzzle_flash_offset_table[10][7] = {

	{0, -37, 90, -1, -1, -1, 1},
	{-24, -40, 25, 24, -40, 25, 2},
	{0, 40, 29, -1, -1, -1, 1},
	{0, -10, 95, -1, -1, -1, 1},
	{0, 35, 90, -1, -1, -1, 1},
	{-38, 16, 110, 38, 16, 110, 2},
	{-7, 40, 80, -1, -1, -1, 1},
	{-35, -11, 110, 35, -11, 110, 1},
	{0, -37, 68, -1, -1, -1, 1},
	{0, -27, 80, -1, -1, -1, 1}

};

//Externs - PSX specific
extern GsOT     Wotwin0[2];
extern GsOT     Wotwin1[2];
extern int      outbuf_idx;
extern GsOT		*ptrwot;
extern u32 rcossin_tbl[4096];


MATRIX					IDENTITY_MATRIX;



//EXPLOSION_CTRL_STRUCT	*init_new_explosion_ctrl_struct(void);
//EXPLOSION_FLAME	*init_new_explosion_flame(void);
//RIPPLE_SPLASH *init_ripple_struct(void);
//RIPPLE_CTRL_STRUCT *init_ripple_ctrl_struct(void);
//void init_new_ripple(RIPPLE_CTRL_STRUCT *ctrl_ptr);
//void release_explosion_flame(EXPLOSION_CTRL_STRUCT *exp_ptr);
//void process_explosion_flame(EXPLOSION_FLAME *exp_ptr);


//Declarations of Global Structures
//RIPPLE_SPLASH				    PSX_ripples[MAX_RIPPLE_SPLASHES];
//RIPPLE_CTRL_STRUCT				PSX_ripple_ctrls[MAX_RIPPLE_CTRL_STRUCTS];

//EXPLOSION_CTRL_STRUCT			PSX_expns[MAX_EXPLOSIONS];
//EXPLOSION_FLAME					PSX_expn_flames[MAX_EXP_FLAMES];

s16 						    EXP_fire_uv_table[MAX_EXP_TIMS][7] =
{      //x0, y0, x1, y1, w, h, size_scaler

    {0, 0, 4, 3, 4, 4, 4},                 //0
    {4, 0, 13, 6, 8, 7, 4},                //1
    {14, 0, 28, 10, 15, 11, 4},            //2
    {29, 0, 43, 13, 15, 14, 4},            //3
    {44, 0, 58, 14, 15, 14, 4},            //4
    {60, 0, 77, 17, 19, 18, 4},            //5
    {78, 0, 96, 19, 19, 20, 4},            //6
    {98, 0, 124, 21, 27, 22, 4},           //7
    {125, 0, 151, 21, 27, 22, 4},          //8
    {152, 0, 174, 26, 23, 27, 4},          //9
    {176, 0, 198, 26, 23, 27, 4},          //10
    {200, 0, 222, 26, 23, 27, 5},          //11
    {226, 0, 255, 27, 31, 28, 5},          //12
    {0, 28, 30, 57, 31, 30, 6},            //13
    {32, 28, 62, 57, 31, 30, 6},           //14
    {64, 28, 95, 59, 31, 32, 6},           //15
    {98, 28, 129, 57, 31, 30, 6},          //16
    {131, 28, 161, 58, 31, 31, 6},         //17
    {166, 28, 196, 57, 31, 30, 4},         //18
    {198, 28, 229, 58, 31, 31, 2},         //19
    {0, 60, 30, 89, 31, 30, 2},            //20
    {31, 60, 61, 88, 31, 29, 2},           //21
    {62, 61, 92, 88, 31, 29, 2},           //22
    {93, 61, 123, 89, 31, 30, 1},          //23
    {124, 61, 150, 89, 27, 30, 1},         //24
    {153, 60, 173, 78, 23, 19, 1}          //25
};

void test_poly(OBJ_STRUCT *obj_ptr);

void test_poly(OBJ_STRUCT *obj_ptr)
{
	POLY_FT4		face;
	SVECTOR			coord[4], cloud_origin, cloud_pos;
	SVECTOR			camera_orientation, ship_rotation;
	int				index, length, display_loop;
	GsOT			*ot_ptr;
	long			i, z_value, p, flag;
	MATRIX			shp_mtx, trans_mtx;


	SetPolyFT4(&face);
    SetSemiTrans(&face, 1);

    face.tpage = GetTPage(0, 1, 704, 256);
	face.clut = GetClut(704, 460);

	face.r0 = 96;
	face.g0 = 96;
	face.b0 = 96;

	length = 10;
    index = 15;

    face.u0 = EXP_fire_uv_table[index][0];
    face.v0 = EXP_fire_uv_table[index][1];

    face.u1 = EXP_fire_uv_table[index][2];
    face.v1 = EXP_fire_uv_table[index][1];

    face.u2 = EXP_fire_uv_table[index][0];
    face.v2 = EXP_fire_uv_table[index][3];

    face.u3 = EXP_fire_uv_table[index][2];
    face.v3 = EXP_fire_uv_table[index][3];

	display_loop = 0;

	camera_orientation.vx = CAM_came[display_loop].vx >> 4;
	camera_orientation.vy = CAM_came[display_loop].vy >> 4;
	camera_orientation.vz = CAM_came[display_loop].vz >> 4;

//    RotMatrix(&camera_orientation, &CAM_camm[display_loop]);

	cloud_origin.vx = obj_ptr->p.vx >> 8;
	cloud_origin.vy = obj_ptr->p.vy >> 8;
	cloud_origin.vz = obj_ptr->p.vz >> 8;

	cloud_origin.vx -= CAM_camp[display_loop].vx;
	cloud_origin.vy -= CAM_camp[display_loop].vy;
	cloud_origin.vz -= CAM_camp[display_loop].vz;

	CAM_camm[display_loop].t[0] = 0;
	CAM_camm[display_loop].t[1] = 0;
	CAM_camm[display_loop].t[2] = 0;

	ApplyMatrixSV(&CAM_camm[display_loop], &cloud_origin, &cloud_origin);

//	TransMatrix(&cloud_origin, &trans_mtx);

	ship_rotation.vx = obj_ptr->r.vx >> 4;
	ship_rotation.vy = obj_ptr->r.vy >> 4;
	ship_rotation.vz = obj_ptr->r.vz >> 4;

//    RotMatrix(&ship_rotation, &shp_mtx);

	coord[0].vx = -100;
	coord[0].vy = -100;
	coord[0].vz = 100;

	coord[1].vx = 100;
	coord[1].vy = -100;
	coord[1].vz = 100;

	coord[2].vx = -100;
	coord[2].vy = 100;
	coord[2].vz = 100;

	coord[3].vx = 100;
	coord[3].vy = 100;
	coord[3].vz = 100;

//	for (i = 0; i < 4; i++)
//		ApplyMatrixSV(&shp_mtx, &coord[i], &coord[i]);

	for (i = 0; i < 4; i++)
	{
		coord[i].vx += cloud_origin.vx;
		coord[i].vy += cloud_origin.vy;
		coord[i].vz += cloud_origin.vz;

//		ApplyMatrixSV(&CAM_camm[display_loop], &coord[i], &coord[i]);
	}

	SetRotMatrix(&IDENTITY_MATRIX);
	SetTransMatrix(&IDENTITY_MATRIX);
//	SetRotMatrix(&shp_mtx);
//	SetRotMatrix(&CAM_camm[display_loop]);
//	SetRotMatrix(&CAM_camm[display_loop]);
//	SetTransMatrix(&trans_mtx);

    z_value = RotTransPers4(&coord[0], &coord[1], &coord[2],
                &coord[3], (long *)&face.x0, (long *)&face.x1,
                (long *)&face.x2, (long *)&face.x3, &p, &flag);

	ot_ptr = &Wotwin0[outbuf_idx];
   	GsSortPoly(&face, ot_ptr, z_value >> otzshift);

	return;
}

void PSX_process_muzzle_flash(OBJ_PLAYERSTRUCT *plyr_ptr)
{
	OBJ_STRUCT		*obj_ptr;
	POLY_FT4		face;
	SVECTOR			coord[4], craft_origin;
	SVECTOR			camera_orientation;
	int				display_loop, face_length, rot_angle;
	char			in_range;
	long			x, z_value, p, flag, x_val, y_val, colour_val;
	int				i, j, index, num_guns;
	GsOT			*ot_ptr;
	vec16			offset[2];
	SVECTOR			ship_rot;
	MATRIX			shp_mtx;


	obj_ptr = plyr_ptr->pobj;

	SetPolyFT4(&face);
    SetSemiTrans(&face, 1);

    face.tpage = GetTPage(0, 1, 704, 256);
	face.clut = GetClut(704, 460);

	colour_val = SDRAND(96);

	face.r0 = 96 + colour_val;
	face.g0 = 96 + colour_val;
	face.b0 = 96 + colour_val;

	face_length = SDRAND(10) + power_up_table[plyr_ptr->weapon_level][muzzle_flash_length];
    index = SDRAND(10) + 10;

    face.u0 = EXP_fire_uv_table[index][0];
    face.v0 = EXP_fire_uv_table[index][1];

    face.u1 = EXP_fire_uv_table[index][2];
    face.v1 = EXP_fire_uv_table[index][1];

    face.u2 = EXP_fire_uv_table[index][0];
    face.v2 = EXP_fire_uv_table[index][3];

    face.u3 = EXP_fire_uv_table[index][2];
    face.v3 = EXP_fire_uv_table[index][3];

	offset[0].vx = muzzle_flash_offset_table[plyr_ptr->crafttype][0];
	offset[0].vy = muzzle_flash_offset_table[plyr_ptr->crafttype][1];
	offset[0].vz = muzzle_flash_offset_table[plyr_ptr->crafttype][2];

	if (muzzle_flash_offset_table[plyr_ptr->crafttype][6] == 2)
	{
		offset[1].vx = muzzle_flash_offset_table[plyr_ptr->crafttype][3];
		offset[1].vy = muzzle_flash_offset_table[plyr_ptr->crafttype][4];
		offset[1].vz = muzzle_flash_offset_table[plyr_ptr->crafttype][5];

		num_guns = 2;
	}
	else
		num_guns = 1;

	ot_ptr = &Wotwin0[outbuf_idx];

	for (display_loop = 0; display_loop < STATE.numscreens; display_loop++)
	{
		in_range = clip_point_long(&obj_ptr->p, display_loop);

		if (!in_range)
		{
			ot_ptr = &Wotwin1[outbuf_idx];

			continue;
		}

		camera_orientation.vx = CAM_came[display_loop].vx >> 4;
		camera_orientation.vy = CAM_came[display_loop].vy >> 4;
		camera_orientation.vz = CAM_came[display_loop].vz >> 4;

	    RotMatrix(&camera_orientation, &CAM_camm[display_loop]);

		for (x = 0; x < num_guns; x++)
		{
			craft_origin.vx = (obj_ptr->p.vx >> 8);
			craft_origin.vy = (obj_ptr->p.vy >> 8);
			craft_origin.vz = (obj_ptr->p.vz >> 8);

			craft_origin.vx -= CAM_camp[display_loop].vx;
			craft_origin.vy -= CAM_camp[display_loop].vy;
			craft_origin.vz -= CAM_camp[display_loop].vz;

			rot_angle = SDRAND(256) - 128;

			for (j = 0; j < 4; j++)
			{
				rot_angle += (1024 + (SDRAND(64) - 32));

				x_val = (face_length * rsin(rot_angle)) >> 12;
				y_val = (face_length * rcos(rot_angle)) >> 12;

				coord[0].vx = 0;
				coord[0].vy = 0;
				coord[0].vz = 0;

				coord[1].vx = 0;
				coord[1].vy = 0;
				coord[1].vz = 0;

				coord[2].vx = x_val;
		 		coord[2].vy = y_val;
				coord[2].vz = 0;

				rot_angle += 256 + (SDRAND(64) - 32);

				x_val = (face_length * rsin(rot_angle)) >> 12;
				y_val = (face_length * rcos(rot_angle)) >> 12;

				rot_angle -= 256;

				coord[3].vx = x_val;
				coord[3].vy = y_val;
				coord[3].vz = 0;

				ship_rot.vx = obj_ptr->r.vx >> 4;
				ship_rot.vy = obj_ptr->r.vy >> 4;
				ship_rot.vz = obj_ptr->r.vz >> 4;

			    RotMatrix(&ship_rot, &shp_mtx);

			    for (i = 0; i < 4; i++)
				{
					coord[i].vx += offset[x].vx;
					coord[i].vy += offset[x].vy;
					coord[i].vz += offset[x].vz;

					ApplyMatrixSV(&shp_mtx, &coord[i], &coord[i]);

					coord[i].vx += craft_origin.vx;
					coord[i].vy += craft_origin.vy;
					coord[i].vz += craft_origin.vz;

					ApplyMatrixSV(&CAM_camm[display_loop], &coord[i], &coord[i]);
				}

	  	    	SetRotMatrix(&IDENTITY_MATRIX);
		    	SetTransMatrix(&IDENTITY_MATRIX);

			    z_value = RotTransPers4(&coord[0], &coord[1], &coord[2],
			                &coord[3], (long *)&face.x0, (long *)&face.x1,
			                (long *)&face.x2, (long *)&face.x3, &p, &flag);

			   	GsSortPoly(&face, ot_ptr, z_value >> otzshift);
			}
		}

		ot_ptr = &Wotwin1[outbuf_idx];
	}

	return;
}

void PSX_init_fragment(FLAME_CTRL *ptr, int type)
{
	DROPLET_STRUCT			*drop_ptr;

	drop_ptr = (DROPLET_STRUCT *)init_new_droplet_struct();

	if (drop_ptr == NULL)
		return;

	drop_ptr->position.vx = ptr->position.vx;
	drop_ptr->position.vy = ptr->position.vy;
	drop_ptr->position.vz = ptr->position.vz;

	drop_ptr->velocity.vx = SDRAND(100) - 50;
	drop_ptr->velocity.vy = -(SDRAND(75) + 50);
	drop_ptr->velocity.vz = SDRAND(100) - 50;

	drop_ptr->length = SDRAND(10) + 10;

	SetPolyFT4(&drop_ptr->face);
//	SetSemiTrans(&drop_ptr->face, 1);

	switch (type)
	{
		case water:
		{
			SetSemiTrans(&drop_ptr->face, 1);

			drop_ptr->face.r0 = 128;
			drop_ptr->face.g0 = 128;
			drop_ptr->face.b0 = 128;

			drop_ptr->face.tpage = GetTPage(0, 1, 704, 256);
			drop_ptr->face.clut = GetClut(736, 446);

			drop_ptr->face.u0 = 128;
			drop_ptr->face.v0 = 126;

			drop_ptr->face.u1 = 135;
			drop_ptr->face.v1 = 126;

			drop_ptr->face.u2 = 128;
			drop_ptr->face.v2 = 133;

			drop_ptr->face.u3 = 135;
			drop_ptr->face.v3 = 133;

			break;
		}

		case fragment:
		{
			drop_ptr->face.r0 = 32;
			drop_ptr->face.g0 = 32;
			drop_ptr->face.b0 = 32;

			drop_ptr->face.tpage = GetTPage(0, 0, 704, 256);
			drop_ptr->face.clut = GetClut(704, 471);

			drop_ptr->face.u0 = 128;
			drop_ptr->face.v0 = 134;

			drop_ptr->face.u1 = 135;
			drop_ptr->face.v1 = 134;

			drop_ptr->face.u2 = 128;
			drop_ptr->face.v2 = 149;

			drop_ptr->face.u3 = 135;
			drop_ptr->face.v3 = 149;

			break;
		}

		default:
			break;
	}

	return;
}

void PSX_process_fragment(DROPLET_STRUCT *ptr)
{
	long				display_loop, z_position, p, flag, coll_val, in_range;
	GsOT	 			*ot_ptr;
	SVECTOR	 			camera_orientation, local_position;
	SVECTOR				coord[4];
	vec32				before, after;


	ptr->velocity.vy += 8;

	before.vx = ptr->position.vx;
	before.vy = ptr->position.vy;
	before.vz = ptr->position.vz;

	ptr->position.vx += ptr->velocity.vx;
	ptr->position.vy += ptr->velocity.vy;
	ptr->position.vz += ptr->velocity.vz;

	after.vx = ptr->position.vx;
	after.vy = ptr->position.vy;
	after.vz = ptr->position.vz;

    coll_val = COL_test2(&before, &after); //Determines if the torpedo has hit the

//	if (ptr->position.vy > 0)
	if (coll_val > NULLVAL)
	{
		if (COL_surftype == 17)
			PSX_init_ripple(ptr);
//		PSX_init_ripple(ptr);
//		PSX_init_ripple(ptr);

		ptr->in_use = FALSE;

		return;
	}

	ot_ptr = &Wotwin0[outbuf_idx];

	for (display_loop = 0; display_loop < STATE.numscreens; display_loop++)
	{
		in_range = clip_point_short(&ptr->position, display_loop);

		if (!in_range)
		{
			ot_ptr = &Wotwin1[outbuf_idx];

			continue;
		}

		camera_orientation.vx = CAM_came[display_loop].vx >> 4;
		camera_orientation.vy = CAM_came[display_loop].vy >> 4;
		camera_orientation.vz = CAM_came[display_loop].vz >> 4;

	    RotMatrix(&camera_orientation, &CAM_camm[display_loop]);

		local_position.vx = ptr->position.vx - (CAM_camp[display_loop].vx);
		local_position.vy = ptr->position.vy - (CAM_camp[display_loop].vy);
		local_position.vz = ptr->position.vz - (CAM_camp[display_loop].vz);

	    ApplyMatrixSV(&CAM_camm[display_loop], &local_position, &local_position);

//		if (local_position.vz < 0)
//		{
//			ot_ptr = &Wotwin1[outbuf_idx];
//			continue;
//		}


		SetTransMatrix(&IDENTITY_MATRIX);
		SetRotMatrix(&IDENTITY_MATRIX);

		coord[0].vx = local_position.vx - ptr->length;
		coord[0].vy = local_position.vy - ptr->length;
		coord[0].vz = local_position.vz;

		coord[1].vx = local_position.vx + ptr->length;
		coord[1].vy = local_position.vy - ptr->length;
		coord[1].vz = local_position.vz;

		coord[2].vx = local_position.vx - ptr->length;
		coord[2].vy = local_position.vy + ptr->length;
		coord[2].vz = local_position.vz;

		coord[3].vx = local_position.vx + ptr->length;
		coord[3].vy = local_position.vy + ptr->length;
		coord[3].vz = local_position.vz;

		z_position = RotTransPers4(&coord[0], &coord[1], &coord[2],
        			&coord[3], (long *)&ptr->face.x0, (long *)&ptr->face.x1,
                	(long *)&ptr->face.x2, (long *)&ptr->face.x3, &p, &flag);

      	GsSortPoly(&ptr->face, ot_ptr, z_position >> otzshift);

		ot_ptr = &Wotwin1[outbuf_idx];
	}

	return;
}
/*
s32 PSX_init_new_explosion(OBJ_STRUCT *obj_ptr, int explosion_type, int collision_type)
{
	EXPLOSION_CTRL_STRUCT		*exp_ptr;


//	if (obj_ptr->type == OTbullet)
//	{
//		OBJ_delobj(obj_ptr->index);

//		return;
//	}

//	OBJ_delobj(obj_ptr->index);
//	return FALSE;

	exp_ptr = (EXPLOSION_CTRL_STRUCT *)init_new_explosion_ctrl_struct();

	if (exp_ptr == NULL)
//	{
//	  	OBJ_delobj(obj_ptr->index);

		return FALSE;
//	}
   S2_sfxdynamic( 14, obj_ptr->p.vx>>8, obj_ptr->p.vz>>8, 0, 127, 4300+(rand()&127) );

	exp_ptr->owner = obj_ptr->owner;	//which plyr owns this explosion
	exp_ptr->timer = 0;

	exp_ptr->type = explosion_type;

	exp_ptr->release_total = 0;

	exp_ptr->first_flame = NULLVAL;
	exp_ptr->latest_flame = NULLVAL;

	exp_ptr->total_flames = 25;

	exp_ptr->position.vx = obj_ptr->p.vx;//COL_inter.vx;//obj_ptr->p.vx;
	exp_ptr->position.vy = obj_ptr->p.vy;//COL_inter.vy;//obj_ptr->p.vy;
	exp_ptr->position.vz = obj_ptr->p.vz;//COL_inter.vz;//obj_ptr->p.vz;

	exp_ptr->rotation.vx = obj_ptr->r.vx;

	if (obj_ptr->r.vy < 0)
		exp_ptr->rotation.vy = 4096 + obj_ptr->r.vy;
	else
		exp_ptr->rotation.vy = 4096 - obj_ptr->r.vy;

	exp_ptr->rotation.vz = obj_ptr->r.vz;


	switch (collision_type)
	{
		case SHIP_COLLISION:
			exp_ptr->position.vx = obj_ptr->p.vx >> 8;//COL_inter.vx;//obj_ptr->p.vx;
			exp_ptr->position.vy = obj_ptr->p.vy >> 8;//COL_inter.vy;//obj_ptr->p.vy;
			exp_ptr->position.vz = obj_ptr->p.vz >> 8;//COL_inter.vz;//obj_ptr->p.vz;

			exp_ptr->rotation.vx = obj_ptr->r.vx >> 4;
			exp_ptr->rotation.vy = obj_ptr->r.vy >> 4;
			exp_ptr->rotation.vz = obj_ptr->r.vz >> 4;

			exp_ptr->velocity.vx = obj_ptr->v.vx >> 8;
			exp_ptr->velocity.vy = obj_ptr->v.vy >> 8;
			exp_ptr->velocity.vz = obj_ptr->v.vz >> 8;

			exp_ptr->collision_type = ship_collision_exp;

			break;

		case MESH_COLLISION:
			exp_ptr->position.vx = COL_inter.vx;
			exp_ptr->position.vy = COL_inter.vy;
			exp_ptr->position.vz = COL_inter.vz;

			exp_ptr->rotation.vx = obj_ptr->r.vx >> 4;

			if (obj_ptr->r.vy < 0)
				exp_ptr->rotation.vy = 4096 + (obj_ptr->r.vy >> 4);
			else
				exp_ptr->rotation.vy = 4096 - (obj_ptr->r.vy >> 4);

			exp_ptr->rotation.vz = obj_ptr->r.vz >> 4;

			exp_ptr->velocity.vx = 0;
			exp_ptr->velocity.vy = 0;
			exp_ptr->velocity.vz = 0;

			exp_ptr->collision_type = mesh_collision_exp;

  			break;

		case OTHER_COLLISION:
			exp_ptr->position.vx = obj_ptr->p.vx >> 8;//COL_inter.vx;//obj_ptr->p.vx;
			exp_ptr->position.vy = obj_ptr->p.vy >> 8;//COL_inter.vy;//obj_ptr->p.vy;
			exp_ptr->position.vz = obj_ptr->p.vz >> 8;//COL_inter.vz;//obj_ptr->p.vz;

			exp_ptr->rotation.vx = obj_ptr->r.vx >> 4;
			exp_ptr->rotation.vy = obj_ptr->r.vy >> 4;
			exp_ptr->rotation.vz = obj_ptr->r.vz >> 4;

			exp_ptr->velocity.vx = obj_ptr->v.vx >> 8;
			exp_ptr->velocity.vy = obj_ptr->v.vy >> 8;
			exp_ptr->velocity.vz = obj_ptr->v.vz >> 8;

			exp_ptr->collision_type = other_collision_exp;

			break;

		default:
			break;
	}

	switch (exp_ptr->type)
	{
		case cluster_explosion:
		{
//			exp_ptr->no_fragments = 5;

			exp_ptr->total_flames = 10;
//			exp_ptr->velocity = 0;

			break;
		}

		case fragment_explosion:
		{
			exp_ptr->no_fragments = 3;

			exp_ptr->total_flames = 6;
//			exp_ptr->velocity = 0;

			break;
		}

		case mine_explosion:
		{
			exp_ptr->no_fragments = 10;

			exp_ptr->total_flames = 20;
//			exp_ptr->velocity = 0;

			break;
		}

		case torpedo_explosion:
		{
			exp_ptr->total_flames = 0;
//			exp_ptr->velocity = 0;
			exp_ptr->position.vy = 0;

			break;
		}

		case missile_explosion:
		{
			exp_ptr->no_fragments = 10;

			exp_ptr->total_flames = 15;
//			exp_ptr->velocity = 0;

			break;
		}

		case rocket_explosion:
		{
			exp_ptr->no_fragments = 2;

			exp_ptr->total_flames = 3;
//			exp_ptr->velocity = 0;

			break;
		}

		case ion_explosion:
		{
			exp_ptr->no_fragments = 0;

			exp_ptr->total_flames = 25;
//			exp_ptr->velocity = 0;

			break;
		}

		case bullet_explosion:
		{
			exp_ptr->no_fragments = 0;

			exp_ptr->total_flames = 1;
//			exp_ptr->velocity = 0;

			break;
		}

		case craft_explosion:
		{
			exp_ptr->no_fragments = 15;

			exp_ptr->total_flames = 30;
//			exp_ptr->velocity = 0;

			break;
		}

		case smoke_fx:
		case craft_trail:
		case missile_trail:
		case flare_trail:
		{
			exp_ptr->no_fragments = 0;

			exp_ptr->total_flames = 30;
//			exp_ptr->velocity = 0;
			exp_ptr->obj_index = obj_ptr->index;

			break;
		}

		default :
			exp_ptr->no_fragments = 0;

			break;
	}

	if ((exp_ptr->type != missile_trail) && (exp_ptr->type != smoke_fx) && (exp_ptr->type != cluster_explosion) && (exp_ptr->type != flare_trail))
	{
		if (obj_ptr->index >=  STATE.wankers)
		  	OBJ_delobj(obj_ptr->index);
	}
	else
		obj_ptr->fx_id = exp_ptr->index;

	return TRUE;
}

void PSX_process_explosions(void)
{
	EXPLOSION_CTRL_STRUCT		*exp_ptr;
	EXPLOSION_FLAME				*ptr;
	int							i, num1 = 0, num2 = 0;


	for (i = 0; i < MAX_EXPLOSIONS; i++)
	{
		if (!PSX_expns[i].in_use)
			continue;

		num1++;

		exp_ptr = &PSX_expns[i];

		switch (exp_ptr->collision_type)
		{
			case other_collision_exp:
			case ship_collision_exp:

				exp_ptr->velocity.vx -= exp_ptr->velocity.vx / 20;
 				exp_ptr->velocity.vy -= exp_ptr->velocity.vy / 20;
				exp_ptr->velocity.vz -= exp_ptr->velocity.vz / 20;

				break;

			default:
				break;
		}

		if ((exp_ptr->type == smoke_fx) || (exp_ptr->type == missile_trail) || (exp_ptr->type == flare_trail))
		{
			exp_ptr->velocity.vx = OBJ_objects[exp_ptr->obj_index].v.vx >> 8;
			exp_ptr->velocity.vy = OBJ_objects[exp_ptr->obj_index].v.vy >> 8;
			exp_ptr->velocity.vz = OBJ_objects[exp_ptr->obj_index].v.vz >> 8;

			exp_ptr->position.vx = OBJ_objects[exp_ptr->obj_index].p.vx >> 8;
			exp_ptr->position.vy = OBJ_objects[exp_ptr->obj_index].p.vy >> 8;
			exp_ptr->position.vz = OBJ_objects[exp_ptr->obj_index].p.vz >> 8;

			exp_ptr->rotation.vx = OBJ_objects[exp_ptr->obj_index].r.vx >> 4;
			exp_ptr->rotation.vy = OBJ_objects[exp_ptr->obj_index].r.vy >> 4;
			exp_ptr->rotation.vz = OBJ_objects[exp_ptr->obj_index].r.vz >> 4;
		}

		if (exp_ptr->first_flame != NULLVAL)
		{
			ptr = &PSX_expn_flames[exp_ptr->first_flame];

			process_explosion_flame(ptr);

			while (ptr->next_flame != NULLVAL)
			{
				ptr = &PSX_expn_flames[ptr->next_flame];

				exp_timer_on = 30;

				process_explosion_flame(ptr);

				exp_timer_off = 40;
			}
		}

		exp_ptr->timer ++;

		if (!(exp_ptr->timer % 2))
		{
			if (exp_ptr->no_fragments > 0)
			{
				PSX_init_fragment(exp_ptr, fragment);
				exp_ptr->no_fragments --;
			}
		}

		if (exp_ptr->release_total < exp_ptr->total_flames)
		{
			if (!(exp_ptr->timer % 1))
				release_explosion_flame(exp_ptr);
		}

		if ((exp_ptr->release_total == exp_ptr->total_flames) && (exp_ptr->first_flame == NULLVAL))
			exp_ptr->in_use = FALSE;
	}

	for (i = 0; i < MAX_EXP_FLAMES; i++)
	{
		if (PSX_expn_flames[i].in_use)
			num2 ++;
	}

//	FntPrint("num_exps = %d : num_flames = %d\n", num1, num2);

	return;
}

void release_explosion_flame(EXPLOSION_CTRL_STRUCT *exp_ptr)
{
	EXPLOSION_FLAME		*ptr;
	int					x_val, z_val;


	ptr = (EXPLOSION_FLAME *)init_new_explosion_flame();

	if (ptr == NULL)
		return;

	ptr->owner = exp_ptr->index;
	ptr->timer = 0;

	ptr->next_flame = NULLVAL;
	ptr->prev_flame = exp_ptr->latest_flame;

	if (exp_ptr->latest_flame == NULLVAL)
		exp_ptr->first_flame = ptr->index;
	else
		PSX_expn_flames[exp_ptr->latest_flame].next_flame = ptr->index;

	exp_ptr->latest_flame = ptr->index;

	if ((exp_ptr->type != missile_trail) && (exp_ptr->type != flare_trail) && (exp_ptr->type != smoke_fx))
		exp_ptr->release_total ++;

	ptr->smoke_point = 20;
	ptr->smoke_flag = FALSE;
	ptr->uv_index = 0;
	ptr->delay = 1;
	ptr->increment = 1;//SDRAND(3) + 1;
	ptr->size = 5;
	ptr->end_point = 25;

	ptr->type = exp_ptr->type;

	ptr->position.vx = exp_ptr->position.vx;
	ptr->position.vy = exp_ptr->position.vy;
	ptr->position.vz = exp_ptr->position.vz;

	ptr->rotation.vx = exp_ptr->rotation.vx + (SDRAND(512) - 256);
	ptr->rotation.vy = exp_ptr->rotation.vy + (SDRAND(512) - 256);
	ptr->rotation.vz = exp_ptr->rotation.vz;

	ptr->velocity = SDRAND(75);//exp_ptr->velocity;

	ptr->r = 64;
	ptr->g = 64;
	ptr->b = 64;

	switch (exp_ptr->type)
	{
		case fragment_explosion:
		{
			ptr->velocity = SDRAND(25);//exp_ptr->velocity;

			break;
		}

		case ion_explosion:
		{
//			ptr->end_point = ptr->smoke_point - 1;

			if (exp_ptr->release_total < 0)
			{
				ptr->size = 15;
				ptr->velocity = SDRAND(25);

			}
			else
			{
				ptr->size = 3;
				ptr->velocity = SDRAND(50);

				ptr->position.vx = exp_ptr->position.vx + (SDRAND(100) - 50);
				ptr->position.vy = exp_ptr->position.vy + (SDRAND(100) - 50);
				ptr->position.vz = exp_ptr->position.vz + (SDRAND(100) - 50);

				ptr->rotation.vx = SDRAND(4096);//exp_ptr->rotation.vx + (SDRAND(512) - 256);
				ptr->rotation.vy = SDRAND(4096);//exp_ptr->rotation.vy + (SDRAND(512) - 256);
			}



			break;
		}

		case missile_trail:
		{
			ptr->velocity = (OBJ_objects[PSX_expns[ptr->owner].obj_index].m0 >> 8);// + (SDRAND(20) - 10);//SquareRoot0((x_val + y_val + z_val));

			ptr->rotation.vx = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vx >> 4;
			ptr->rotation.vy = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vy >> 4;
//			ptr->rotation.vz = 0;//obj_ptr->r.vz;

			ptr->r =
			ptr->g =
			ptr->b = 64 + SDRAND(128);

			break;
		}

		case craft_trail:
		{
//			ptr->velocity = (OBJ_objects[PSX_expns[ptr->owner].obj_index].m0 >> 8);// + (SDRAND(20) - 10);//SquareRoot0((x_val + y_val + z_val));

			x_val = (OBJ_objects[PSX_expns[ptr->owner].obj_index].v.vx >> 8) * (OBJ_objects[PSX_expns[ptr->owner].obj_index].v.vx >> 8);// + (SDRAND(20) - 10);//SquareRoot0((x_val + y_val + z_val));
			z_val = (OBJ_objects[PSX_expns[ptr->owner].obj_index].v.vz >> 8) * (OBJ_objects[PSX_expns[ptr->owner].obj_index].v.vz >> 8);// + (SDRAND(20) - 10);//SquareRoot0((x_val + y_val + z_val));

			ptr->velocity = SquareRoot0(x_val + z_val);

			ptr->rotation.vx = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vx >> 4;
			ptr->rotation.vy = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vy >> 4;
//			ptr->rotation.vz = 0;//obj_ptr->r.vz;

			ptr->r =
			ptr->g =
			ptr->b = 64 + SDRAND(128);

			ptr->delay = 1;
			ptr->increment = SDRAND(3) + 1;
			ptr->size = 7;

			break;
		}

		case smoke_fx:
		{
			ptr->r =
			ptr->g =
			ptr->b = 32 + SDRAND(32);//96 + SDRAND(64);

		    ptr->uv_index = 0;

			ptr->delay = 1;

			ptr->increment = 1;

			ptr->velocity = SDRAND(15) + 25;

			ptr->craft_vel.vx = exp_ptr->velocity.vx;
			ptr->craft_vel.vy = exp_ptr->velocity.vy;
			ptr->craft_vel.vz = exp_ptr->velocity.vz;

			ptr->rotation.vx = SDRAND(512);
			ptr->rotation.vy = SDRAND(512);
			ptr->rotation.vz = SDRAND(2048);

			if (exp_ptr->rotation.vy < 0)
				ptr->rotation.vy += (4096 + exp_ptr->rotation.vy);
			else
				ptr->rotation.vy += 4096 - exp_ptr->rotation.vy;

			if (ptr->rotation.vy > 4096)
				ptr->rotation.vy -= 4096;

			if (ptr->rotation.vy < -4096)
				ptr->rotation.vy += 4096;

			break;
		}

		case flare_trail:
		{
			ptr->r =
			ptr->g =
			ptr->b = 32 + SDRAND(32);

		    ptr->uv_index = 0;

			ptr->delay = SDRAND(2) + 1;

			ptr->increment	= SDRAND(3);

			if (!ptr->increment)
				ptr->increment = 2;

			ptr->velocity = (OBJ_objects[PSX_expns[ptr->owner].obj_index].m0 >> 8) + (SDRAND(20) - 10);//SquareRoot0((x_val + y_val + z_val));
			ptr->velocity -= (ptr->velocity / (SDRAND(5) + 3));

			ptr->rotation.vx = exp_ptr->rotation.vx;
			ptr->rotation.vy = exp_ptr->rotation.vy;
			ptr->rotation.vz = exp_ptr->rotation.vz;

			break;
  		}

		case bullet_explosion:
		{
			ptr->increment = 1;//SDRAND(2) + 1;
			ptr->velocity = SDRAND(50);//exp_ptr->velocity;

			break;
		}

		default :
			break;
	}

	return;
}

void process_explosion_flame(EXPLOSION_FLAME *ptr)
{
    SVECTOR   		       	cloud_origin;
    SVECTOR         		cloud_pos;
	GsOT					*ot_ptr;

	EXPLOSION_FLAME			*last_ptr, *next_ptr;
	POLY_FT4				face;
	SVECTOR					coord[4];
	long					z_value, p, flag, two_d_velocity;
	long					display_loop, in_range;


    ptr->timer ++;

	if (!(ptr->timer % ptr->delay))
	{
		ptr->uv_index += ptr->increment;

		if (ptr->uv_index >= ptr->end_point)
		{
			if (ptr->next_flame != NULLVAL)
			{
				if (ptr->prev_flame != NULLVAL)
				{
					last_ptr = &PSX_expn_flames[ptr->prev_flame];
					last_ptr->next_flame = ptr->next_flame;

					next_ptr = &PSX_expn_flames[ptr->next_flame];
					next_ptr->prev_flame = ptr->prev_flame;
				}
				else
				{
					next_ptr = &PSX_expn_flames[ptr->next_flame];
					next_ptr->prev_flame = NULLVAL;

					PSX_expns[ptr->owner].first_flame = ptr->next_flame;
				}
			}
			else
			{
				if (ptr->prev_flame != NULLVAL)
				{
					last_ptr = &PSX_expn_flames[ptr->prev_flame];

					last_ptr->next_flame = NULLVAL;
				}
				else
				{
					PSX_expns[ptr->owner].first_flame = NULLVAL;
					PSX_expns[ptr->owner].latest_flame = NULLVAL;
				}
			}

			ptr->in_use = FALSE;

			return;
		}

		if (ptr->uv_index >= ptr->smoke_point)
		{
			ptr->smoke_flag = TRUE;

		    ptr->r = 192;
			ptr->g = 192;
		    ptr->b = 192;
		}
	}

	SetPolyFT4(&face);
    SetSemiTrans(&face, 1);

	switch (ptr->type)
	{
		case ion_explosion:

			ptr->velocity -= (ptr->velocity / 10);
			ptr->rotation.vx -= 44;

			ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
			two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

			ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
			ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;

			if (ptr->size == 15)
	 			ptr->size = 15;
			else
				ptr->size = EXP_fire_uv_table[ptr->uv_index][6];

			if (ptr->smoke_flag)
			{
			    face.tpage = GetTPage(0, 1, 704, 256);
				face.clut = GetClut(704, 453);
			}
			else
			{
			    face.tpage = GetTPage(0, 1, 704, 256);
				face.clut = GetClut(704, 452);

//				ptr->position.vy -= 5;
			}


			if (ptr->uv_index >= 18)
			{
				ptr->smoke_flag = TRUE;

			    ptr->r = 128;
				ptr->g = 128;
			    ptr->b = 128;
			}

//			ptr->r -= 3;
//			ptr->g -= 3;
//			ptr->b -= 3;

			break;

		case missile_trail:
			if (PSX_expns[ptr->owner].release_total != PSX_expns[ptr->owner].total_flames)
			{
				ptr->velocity -= ptr->velocity / 1000;

				ptr->rotation.vx = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vx >> 4;
				ptr->rotation.vy = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vy >> 4;

//				ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
				ptr->position.vy = OBJ_objects[PSX_expns[ptr->owner].obj_index].p.vy >> 8;
				two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

				ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
				ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;
			}
			else
				ptr->velocity = 0;

		    if (!(ptr->smoke_flag))
			{
			    face.tpage = GetTPage(0, 3, 704, 256);
				face.clut = GetClut(704, 462);

				ptr->size = 4;
			}
			else
			{
			    face.tpage = GetTPage(0, 2, 704, 256);
				face.clut = GetClut(704, 464);

				ptr->size = 6;

//				ptr->size = 6;
			}

			if (ptr->uv_index >= 18)
			{
				ptr->smoke_flag = TRUE;

			    ptr->r = 128;
				ptr->g = 128;
			    ptr->b = 128;
			}

//			ptr->position.vx += (ptr->velocity * rsin(ptr->rotation.vy)) >> 12;
//			ptr->position.vz += (ptr->velocity * rcos(ptr->rotation.vy)) >> 12;

			break;

		case craft_trail:

			if (PSX_expns[ptr->owner].release_total != PSX_expns[ptr->owner].total_flames)
			{
				ptr->position.vx += PSX_expns[ptr->owner].velocity.vx;
				ptr->position.vy += PSX_expns[ptr->owner].velocity.vy;
				ptr->position.vz += PSX_expns[ptr->owner].velocity.vz;

				ptr->velocity -= ptr->velocity / 1000;

				ptr->rotation.vx = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vx >> 4;
				ptr->rotation.vy = OBJ_objects[PSX_expns[ptr->owner].obj_index].r.vy >> 4;

				ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
				two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

				ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
				ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;
			}
			else
				ptr->velocity = 0;

			ptr->position.vy -= 5;

		    if (!(ptr->smoke_flag))
			{
			    face.tpage = GetTPage(0, 3, 704, 256);
				face.clut = GetClut(704, 462);

				ptr->size = 7;
			}
			else
			{
			    face.tpage = GetTPage(0, 2, 704, 256);
				face.clut = GetClut(704, 464);

				ptr->size = 6;

//				ptr->size = 6;
			}

			if (ptr->uv_index >= 18)
			{
				ptr->smoke_flag = TRUE;

			    ptr->r = 128;
				ptr->g = 128;
			    ptr->b = 128;
			}

//			ptr->position.vx += (ptr->velocity * rsin(ptr->rotation.vy)) >> 12;
//			ptr->position.vz += (ptr->velocity * rcos(ptr->rotation.vy)) >> 12;

			break;

		case flare_trail:

		    face.tpage = GetTPage(0, 2, 704, 256);
			face.clut = GetClut(704, 464);

			ptr->size = (EXP_fire_uv_table[ptr->uv_index][6] * 2);
			ptr->velocity -= (ptr->velocity / 20);

			ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
			two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

			ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
			ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;

			ptr->position.vy -= 15;

			break;

		case smoke_fx:

			ptr->craft_vel.vx = PSX_expns[ptr->owner].velocity.vx;
			ptr->craft_vel.vy = PSX_expns[ptr->owner].velocity.vy;
			ptr->craft_vel.vz = PSX_expns[ptr->owner].velocity.vz;

			if (ptr->smoke_flag)
			{
			    face.tpage = GetTPage(0, 2, 704, 256);
				face.clut = GetClut(704, 465);
			}
			else
			{
			    face.tpage = GetTPage(0, 1, 704, 256);
				face.clut = GetClut(704, 462);
			}

//		    face.tpage = GetTPage(0, 1, 704, 256);
//			face.clut = GetClut(704, 462);

			ptr->size = (EXP_fire_uv_table[ptr->uv_index][6] * 3) / 2;

			ptr->position.vx += ptr->craft_vel.vx;
			ptr->position.vy += ptr->craft_vel.vy;
			ptr->position.vz += ptr->craft_vel.vz;

//			ptr->velocity -= (ptr->velocity / 1000);

			ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
			two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

			ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
			ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;

			ptr->position.vy -= 5;//20;

			break;

		case torpedo_explosion:

			ptr->velocity -= (ptr->velocity / 10);
			ptr->rotation.vx -= 44;

			ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
			two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

			ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
			ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;

			if (ptr->size == 15)
	 			ptr->size = 15;
			else
				ptr->size = EXP_fire_uv_table[ptr->uv_index][6];

			if (ptr->smoke_flag)
			{
			    face.tpage = GetTPage(0, 1, 704, 256);
				face.clut = GetClut(704, 453);
			}
			else
			{
			    face.tpage = GetTPage(0, 1, 704, 256);
				face.clut = GetClut(704, 452);

				ptr->position.vy -= 5;
			}


			break;

		case bullet_explosion:
			ptr->velocity -= (ptr->velocity / 2);
			ptr->rotation.vx -= 44;

			ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
			two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

			ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
			ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;

			ptr->size = EXP_fire_uv_table[ptr->uv_index][6];

		    if (!(ptr->smoke_flag))
			{
			    face.tpage = GetTPage(0, 1, 704, 256);
				face.clut = GetClut(704, 462);

//				ptr->size = 2 * EXP_fire_uv_table[ptr->uv_index][6];
			}
			else
			{
			    face.tpage = GetTPage(0, 2, 704, 256);
				face.clut = GetClut(704, 464);

//				ptr->size = 6;
			}


			break;

		default:
			ptr->velocity -= (ptr->velocity / 10);
			ptr->rotation.vx -= 44;

			ptr->position.vy += (ptr->velocity * rsin(ptr->rotation.vx)) >> 12;
			two_d_velocity = (ptr->velocity * rcos(ptr->rotation.vx)) >> 12;

			ptr->position.vx += (two_d_velocity * rsin(ptr->rotation.vy)) >> 12;
			ptr->position.vz += (two_d_velocity * rcos(ptr->rotation.vy)) >> 12;

		    if (!(ptr->smoke_flag))
			{
			    face.tpage = GetTPage(0, 1, 704, 256);
				face.clut = GetClut(704, 462);

				ptr->size = 2 * EXP_fire_uv_table[ptr->uv_index][6];
			}
			else
			{
			    face.tpage = GetTPage(0, 2, 704, 256);
				face.clut = GetClut(704, 464);

				ptr->size = 6;
			}

			break;
	}

	face.r0 = ptr->r;
	face.g0 = ptr->g;
	face.b0 = ptr->b;

    face.u0 = EXP_fire_uv_table[ptr->uv_index][0];
    face.v0 = EXP_fire_uv_table[ptr->uv_index][1];

    face.u1 = EXP_fire_uv_table[ptr->uv_index][2];
    face.v1 = EXP_fire_uv_table[ptr->uv_index][1];

    face.u2 = EXP_fire_uv_table[ptr->uv_index][0];
    face.v2 = EXP_fire_uv_table[ptr->uv_index][3];

    face.u3 = EXP_fire_uv_table[ptr->uv_index][2];
    face.v3 = EXP_fire_uv_table[ptr->uv_index][3];

	ot_ptr = &Wotwin0[outbuf_idx];

	for (display_loop = 0; display_loop < STATE.numscreens; display_loop++)
	{
		in_range = clip_point_short(&ptr->position, display_loop);
 
		if (!in_range)
		{
			ot_ptr = &Wotwin1[outbuf_idx];

			continue;
		}

		cloud_origin.vx = ptr->position.vx - CAM_camp[display_loop].vx;
		cloud_origin.vy = ptr->position.vy - CAM_camp[display_loop].vy;
		cloud_origin.vz = ptr->position.vz - CAM_camp[display_loop].vz;

	    ApplyMatrixSV(&CAM_camm[display_loop], &cloud_origin, &cloud_pos);

//		if (cloud_origin.vz < 0)
//		{
//			ot_ptr = &Wotwin1[outbuf_idx];
//			continue;
//		}

		coord[0].vx = cloud_pos.vx + -(ptr->size * (EXP_fire_uv_table[ptr->uv_index][4] / 2));
		coord[0].vy = cloud_pos.vy + -(ptr->size * (EXP_fire_uv_table[ptr->uv_index][5] / 2));
		coord[0].vz = cloud_pos.vz;

		coord[1].vx = cloud_pos.vx + (ptr->size * (EXP_fire_uv_table[ptr->uv_index][4] / 2));
		coord[1].vy = cloud_pos.vy + -(ptr->size * (EXP_fire_uv_table[ptr->uv_index][5] / 2));
		coord[1].vz = cloud_pos.vz;

		coord[2].vx = cloud_pos.vx + -(ptr->size * (EXP_fire_uv_table[ptr->uv_index][4] / 2));
 		coord[2].vy = cloud_pos.vy + (ptr->size * (EXP_fire_uv_table[ptr->uv_index][5] / 2));
		coord[2].vz = cloud_pos.vz;

		coord[3].vx = cloud_pos.vx + (ptr->size * (EXP_fire_uv_table[ptr->uv_index][4] / 2));
		coord[3].vy = cloud_pos.vy + (ptr->size * (EXP_fire_uv_table[ptr->uv_index][5] / 2));
		coord[3].vz = cloud_pos.vz;

	    SetRotMatrix(&IDENTITY_MATRIX);
	    SetTransMatrix(&IDENTITY_MATRIX);

	    z_value = RotTransPers4(&coord[0], &coord[1], &coord[2],
	                &coord[3], (long *)&face.x0, (long *)&face.x1,
	                (long *)&face.x2, (long *)&face.x3, &p, &flag);

		if ((abs((face.x3 - face.x0)) < 100) && (abs((face.y3 - face.y0)) < 100))
			GsSortPoly(&face, ot_ptr, z_value >> otzshift);

		ot_ptr = &Wotwin1[outbuf_idx];
	}

	return;
}
*/
/*
void PSX_init_ripple(OBJ_STRUCT *obj_ptr)
{
	RIPPLE_CTRL_STRUCT	*ctrl_ptr;


	ctrl_ptr = (RIPPLE_CTRL_STRUCT *)init_ripple_ctrl_struct();

	if (ctrl_ptr == NULL)
	{
		OBJ_delobj(obj_ptr->index);

		return;
	}

	ctrl_ptr->no_ripples = 50;
	ctrl_ptr->current_no = 0;

	ctrl_ptr->owner	= obj_ptr->index;

	obj_ptr->fx_id = ctrl_ptr->index;

	return;
}

void PSX_process_ripple_ctrls(void)
{
	RIPPLE_CTRL_STRUCT		*ctrl_ptr;
	int						i, j;


	for (i = 0; i < MAX_RIPPLE_CTRL_STRUCTS; i++)
	{
		if (!PSX_ripple_ctrls[i].in_use)
			continue;

		ctrl_ptr = &PSX_ripple_ctrls[i];

		ctrl_ptr->position.vx = OBJ_objects[ctrl_ptr->owner].p.vx >> 8;
		ctrl_ptr->position.vy = OBJ_objects[ctrl_ptr->owner].p.vy >> 8;
		ctrl_ptr->position.vz = OBJ_objects[ctrl_ptr->owner].p.vz >> 8;

		if (ctrl_ptr->current_no < ctrl_ptr->no_ripples)
		{
			for (j = 0; j < 3; j++)
				init_new_ripple(ctrl_ptr);
		}
	}

	return;
}

void init_new_ripple(RIPPLE_CTRL_STRUCT *ctrl_ptr)
{
	RIPPLE_SPLASH		*ptr;
	int					x_val, z_val;


	ptr = (RIPPLE_SPLASH *)init_ripple_struct();

	if (ptr == NULL)
		return;

	ptr->timer = 0;
	ptr->radius = 0;
	ptr->end_point = 10;

	ptr->rad_inc = SDRAND(10) + 5;

	ptr->position.vx = ctrl_ptr->position.vx;// + (SDRAND(50) - 25);
	ptr->position.vy = 0;
	ptr->position.vz = ctrl_ptr->position.vz;// + (SDRAND(50) - 25);

	x_val = (OBJ_objects[ctrl_ptr->owner].v.vx >> 8) * (OBJ_objects[ctrl_ptr->owner].v.vx >> 8);
	z_val = (OBJ_objects[ctrl_ptr->owner].v.vz >> 8) * (OBJ_objects[ctrl_ptr->owner].v.vz >> 8);

	ptr->velocity = SquareRoot0((x_val + z_val));
	ptr->velocity -= (ptr->velocity / 10);

	ptr->rotation = OBJ_objects[ctrl_ptr->owner].r.vy >> 4;

	ptr->r = 128;
	ptr->g = 64;
	ptr->b = 64;

	return;
}

void PSX_process_ripples(void)
{
    MATRIX          cloud_matrix;
    VECTOR          cloud_origin;
    VECTOR          cloud_pos;
    SVECTOR         camera_orientation, cloud_orientation;
	RIPPLE_SPLASH	*ptr;
	POLY_FT4		face;
	SVECTOR			coord[4];
	long			i, z_position, p, flag;


    SetPolyFT4(&face);
    SetSemiTrans(&face, 1);

    face.tpage = GetTPage(0, 1, 704, 256);
	face.clut = GetClut(704, 473);

    face.u0 = 96;
    face.v0 = 94;

    face.u1 = 127;
    face.v1 = 94;

    face.u2 = 96;
    face.v2 = 125;

    face.u3 = 127;
    face.v3 = 125;

	for (i = 0; i < MAX_RIPPLE_SPLASHES; i++)
	{
		if (!(PSX_ripples[i].in_use))
			continue;

		ptr = &PSX_ripples[i];

		ptr->timer ++;

		if (ptr->timer == ptr->end_point)
		{
			PSX_ripples[i].in_use = FALSE;

			continue;
		}

		ptr->position.vx += (ptr->velocity * rsin(ptr->rotation)) >> 12;
		ptr->position.vz += (ptr->velocity * rcos(ptr->rotation)) >> 12;

		ptr->velocity -= (ptr->velocity / 25);

		ptr->radius += ptr->rad_inc;

//		ptr->r -= 6;
//		ptr->g -= 6;
//		ptr->b -= 6;

	    coord[0].vx = -ptr->radius;
	    coord[0].vy = 0;
	    coord[0].vz = -ptr->radius;

	    coord[1].vx = ptr->radius;
	    coord[1].vy = 0;
	    coord[1].vz = -ptr->radius;

	    coord[2].vx = -ptr->radius;
	    coord[2].vy = 0;
	    coord[2].vz = ptr->radius;

	    coord[3].vx = ptr->radius;
	    coord[3].vy = 0;
	    coord[3].vz = ptr->radius;

		camera_orientation.vx = CAM_came[CAM_c].vx >> 4;
		camera_orientation.vy = CAM_came[CAM_c].vy >> 4;
		camera_orientation.vz = CAM_came[CAM_c].vz >> 4;

	    RotMatrix(&camera_orientation, &CAM_camm[CAM_c]);

	    cloud_origin.vx = ptr->position.vx;
	    cloud_origin.vy = ptr->position.vy;
	    cloud_origin.vz = ptr->position.vz;

		cloud_origin.vx -= CAM_camp[CAM_c].vx;
		cloud_origin.vy -= CAM_camp[CAM_c].vy;
		cloud_origin.vz -= CAM_camp[CAM_c].vz;

	    ApplyMatrixLV(&CAM_camm[CAM_c], &cloud_origin, &cloud_pos);

		cloud_orientation.vx = 0;
		cloud_orientation.vy = 0;
		cloud_orientation.vz = 0;

	    RotMatrix(&cloud_orientation, &cloud_matrix);

	  	cloud_matrix.m[0][0] = (cloud_matrix.m[0][0] * 25) >> 4;//aspect fix
	    cloud_matrix.m[0][1] = (cloud_matrix.m[0][1] * 25) >> 4;//aspect fix
	    cloud_matrix.m[0][2] = (cloud_matrix.m[0][2] * 25) >> 4;//aspect fix

	    cloud_matrix.t[0] = cloud_pos.vx;
	    cloud_matrix.t[1] = cloud_pos.vy;
	    cloud_matrix.t[2] = cloud_pos.vz;

	    SetTransMatrix(&cloud_matrix);
	    SetRotMatrix(&cloud_matrix);

		face.r0 = ptr->r;
		face.g0 = ptr->g;
		face.b0 = ptr->b;

	    z_position = RotTransPers4(&coord[0], &coord[1], &coord[2],
	                &coord[3], (long *)&face.x0, (long *)&face.x1,
	                (long *)&face.x2, (long *)&face.x3, &p, &flag);

	   	GsSortPoly(&face, ptrwot, z_position >> otzshift);
	}

	return;
}

void PSX_explosions(void)
{
	FLAME_ANIM		*anim_ptr;
    MATRIX          cloud_matrix;
    VECTOR          cloud_origin;
    VECTOR          cloud_pos;
    SVECTOR         camera_orientation, cloud_orientation;

	POLY_FT4		face;
	SVECTOR			coord[4];
	long			i, z_position, p, flag, two_d_velocity, display_loop;
	GsOT			*ot_ptr;


    SetPolyFT4(&face);
    SetSemiTrans(&face, 1);


	for (i = 0; i < MAX_FLAMES; i++)
	{
		if (!PSX_flames[i].in_use)
			continue;

		if ((PSX_flames[i].type != missile_trail) && (PSX_flames[i].type != ion) && (PSX_flames[i].type != flare_smoke))
			continue;

		anim_ptr = &PSX_flames[i];

		switch (anim_ptr->type)
		{
			case ion:
			{
				if (anim_ptr->smoke_flag)
				{
				    face.tpage = GetTPage(0, 1, 704, 256);
					face.clut = GetClut(704, 453);
				}
				else
				{
				    face.tpage = GetTPage(0, 1, 704, 256);
					face.clut = GetClut(704, 452);

					anim_ptr->position.vy -= 5;
				}

				if (anim_ptr->uv_index >= 18)
				{
					anim_ptr->smoke_flag = TRUE;

				    anim_ptr->r = 128;
					anim_ptr->g = 128;
				    anim_ptr->b = 128;
				}

				break;
			}

			case bullet_explosion:
			{
				if (anim_ptr->smoke_flag)
				{
				    face.tpage = GetTPage(0, 2, 704, 256);
					face.clut = GetClut(704, 465);

					anim_ptr->size_scaler = 5;
				}
				else
				{
				    face.tpage = GetTPage(0, 1, 704, 256);
					face.clut = GetClut(704, 462);

					anim_ptr->size_scaler = EXP_fire_uv_table[anim_ptr->uv_index][6];

					anim_ptr->position.vy -= 5;
				}

				if (anim_ptr->uv_index >= 18)
				{
					anim_ptr->smoke_flag = TRUE;

				    anim_ptr->r = 128;
					anim_ptr->g = 128;
				    anim_ptr->b = 128;
				}

				break;
			}

			case missile_trail:
			{
				if (anim_ptr->smoke_flag)
				{
				    face.tpage = GetTPage(0, 2, 704, 256);
					face.clut = GetClut(704, 464);

					anim_ptr->size_scaler = 6;
				}
				else
				{
				    face.tpage = GetTPage(0, 3, 704, 256);
					face.clut = GetClut(704, 462);

					anim_ptr->size_scaler = 4;//EXP_fire_uv_table[anim_ptr->uv_index][6];
				}

				if (anim_ptr->uv_index >= 18)
				{
					anim_ptr->smoke_flag = TRUE;

				    anim_ptr->r = 128;
					anim_ptr->g = 128;
				    anim_ptr->b = 128;
				}

				anim_ptr->craft_vel.vx = OBJ_objects[anim_ptr->obj_index].v.vx;
				anim_ptr->craft_vel.vy = OBJ_objects[anim_ptr->obj_index].v.vy;
				anim_ptr->craft_vel.vz = OBJ_objects[anim_ptr->obj_index].v.vz;

				anim_ptr->velocity -= anim_ptr->velocity / 60;

//				anim_ptr->position.vx = OBJ_objects[anim_ptr->obj_index].p.vx;
//				anim_ptr->position.vy = OBJ_objects[anim_ptr->obj_index].p.vy;
//				anim_ptr->position.vz = OBJ_objects[anim_ptr->obj_index].p.vz;

//				anim_ptr->position.vx += anim_ptr->craft_vel.vx;
//				anim_ptr->position.vy += anim_ptr->craft_vel.vy;
//				anim_ptr->position.vz += anim_ptr->craft_vel.vz;

				anim_ptr->rotation.vx = OBJ_objects[anim_ptr->obj_index].r.vx;
				anim_ptr->rotation.vy = OBJ_objects[anim_ptr->obj_index].r.vy;
				anim_ptr->rotation.vz = OBJ_objects[anim_ptr->obj_index].r.vz;

//				anim_ptr->position.vy += (anim_ptr->velocity * rsin(anim_ptr->rotation.vx)) >> 12;
//				two_d_velocity = (anim_ptr->velocity * rcos(anim_ptr->rotation.vx)) >> 12;

//				two_d_velocity = OBJ_objects[anim_ptr->obj_index].m0;
				two_d_velocity = anim_ptr->velocity;

				anim_ptr->position.vx += (two_d_velocity * rsin(anim_ptr->rotation.vy)) >> 12;
				anim_ptr->position.vz += (two_d_velocity * rcos(anim_ptr->rotation.vy)) >> 12;

				break;
			}

			case flare_smoke:
			{
			    face.tpage = GetTPage(0, 2, 704, 256);
				face.clut = GetClut(704, 464);

				anim_ptr->size_scaler = (EXP_fire_uv_table[anim_ptr->uv_index][6] * 2);
				anim_ptr->velocity -= (anim_ptr->velocity / 20);

				anim_ptr->position.vy += (anim_ptr->velocity * rsin(anim_ptr->rotation.vx)) >> 12;
				two_d_velocity = (anim_ptr->velocity * rcos(anim_ptr->rotation.vx)) >> 12;

				anim_ptr->position.vx += (two_d_velocity * rsin(anim_ptr->rotation.vy)) >> 12;
				anim_ptr->position.vz += (two_d_velocity * rcos(anim_ptr->rotation.vy)) >> 12;

				anim_ptr->position.vy -= 15;

				break;
			}

			case explosion:
			{
				if (anim_ptr->smoke_flag)
				{
				    face.tpage = GetTPage(0, 2, 704, 256);
					face.clut = GetClut(704, 465);

					anim_ptr->size_scaler = 3;
					anim_ptr->position.vy -= 5;
				}
				else
				{
				    face.tpage = GetTPage(0, 1, 704, 256);
					face.clut = GetClut(704, 462);

					anim_ptr->size_scaler = EXP_fire_uv_table[anim_ptr->uv_index][6];
				}

				if (anim_ptr->uv_index >= 18)
				{
					anim_ptr->smoke_flag = TRUE;

				    anim_ptr->r = 64 + SDRAND(64);
					anim_ptr->g = 64 + SDRAND(64);
				    anim_ptr->b = 64 + SDRAND(64);
				}

				anim_ptr->velocity -= (anim_ptr->velocity / 20);

				anim_ptr->position.vy += (anim_ptr->velocity * rsin(anim_ptr->rotation.vx)) >> 12;
				two_d_velocity = (anim_ptr->velocity * rcos(anim_ptr->rotation.vx)) >> 12;

				anim_ptr->position.vx += (two_d_velocity * rsin(anim_ptr->rotation.vy)) >> 12;
				anim_ptr->position.vz += (two_d_velocity * rcos(anim_ptr->rotation.vy)) >> 12;


				break;
	 		}

			default:
			{
				continue;

				break;
			}
		}

		if (!(anim_ptr->timer % anim_ptr->update))
			anim_ptr->uv_index += anim_ptr->increment;

		if (anim_ptr->uv_index >= MAX_EXP_TIMS)
		{
			anim_ptr->in_use = FALSE;
			anim_ptr->type = NULLVAL;

			continue;
		}

		anim_ptr->timer ++;

		face.r0 = anim_ptr->r;
		face.g0 = anim_ptr->g;
		face.b0 = anim_ptr->b;

	    face.u0 = EXP_fire_uv_table[anim_ptr->uv_index][0];
	    face.v0 = EXP_fire_uv_table[anim_ptr->uv_index][1];

	    face.u1 = EXP_fire_uv_table[anim_ptr->uv_index][2];
	    face.v1 = EXP_fire_uv_table[anim_ptr->uv_index][1];

	    face.u2 = EXP_fire_uv_table[anim_ptr->uv_index][0];
	    face.v2 = EXP_fire_uv_table[anim_ptr->uv_index][3];

	    face.u3 = EXP_fire_uv_table[anim_ptr->uv_index][2];
	    face.v3 = EXP_fire_uv_table[anim_ptr->uv_index][3];

	    coord[0].vx = (anim_ptr->size_scaler * (((-EXP_fire_uv_table[anim_ptr->uv_index][4] / 2))));// * 25) >> 4);
	    coord[0].vy = (anim_ptr->size_scaler * (-EXP_fire_uv_table[anim_ptr->uv_index][5] / 2));
	    coord[0].vz = 0;

	    coord[1].vx = (anim_ptr->size_scaler * (((EXP_fire_uv_table[anim_ptr->uv_index][4] / 2))));// * 25) >> 4);
	    coord[1].vy = (anim_ptr->size_scaler * (-EXP_fire_uv_table[anim_ptr->uv_index][5] / 2));
	    coord[1].vz = 0;

	    coord[2].vx = (anim_ptr->size_scaler * (((-EXP_fire_uv_table[anim_ptr->uv_index][4] / 2))));// * 25) >> 4);
	    coord[2].vy = (anim_ptr->size_scaler * (EXP_fire_uv_table[anim_ptr->uv_index][5] / 2));
	    coord[2].vz = 0;

	    coord[3].vx = (anim_ptr->size_scaler * (((EXP_fire_uv_table[anim_ptr->uv_index][4] / 2))));// * 25) >> 4);
	    coord[3].vy = (anim_ptr->size_scaler * (EXP_fire_uv_table[anim_ptr->uv_index][5] / 2));
	    coord[3].vz = 0;

		ot_ptr = &Wotwin0[outbuf_idx];

		for (display_loop = 0; display_loop < STATE.numscreens; display_loop++)
		{
   			camera_orientation.vx = CAM_came[display_loop].vx >> 4;
			camera_orientation.vy = CAM_came[display_loop].vy >> 4;
			camera_orientation.vz = CAM_came[display_loop].vz >> 4;

		    RotMatrix(&camera_orientation, &CAM_camm[display_loop]);

		    cloud_origin.vx = anim_ptr->position.vx;
		    cloud_origin.vy = anim_ptr->position.vy;
		    cloud_origin.vz = anim_ptr->position.vz;

			cloud_origin.vx -= CAM_camp[display_loop].vx;
			cloud_origin.vy -= CAM_camp[display_loop].vy;
			cloud_origin.vz -= CAM_camp[display_loop].vz;

		    ApplyMatrixLV(&CAM_camm[display_loop], &cloud_origin, &cloud_pos);
		    cloud_pos.vx = (cloud_pos.vx * 25) >> 4;

//			if (cloud_origin.vz < 0)
//			{
//				ot_ptr = &Wotwin1[outbuf_idx];
/ /				continue;
//			}

			cloud_orientation.vx = 0;
			cloud_orientation.vy = 2048;
			cloud_orientation.vz = 0;

		    RotMatrix(&cloud_orientation, &cloud_matrix);

		  	cloud_matrix.m[0][0] = (cloud_matrix.m[0][0] * 25) >> 4;//aspect fix
		    cloud_matrix.m[0][1] = (cloud_matrix.m[0][1] * 25) >> 4;//aspect fix
		    cloud_matrix.m[0][2] = (cloud_matrix.m[0][2] * 25) >> 4;//aspect fix

		    cloud_matrix.t[0] = cloud_pos.vx;
		    cloud_matrix.t[1] = cloud_pos.vy;
		    cloud_matrix.t[2] = cloud_pos.vz;

		    SetTransMatrix(&cloud_matrix);
		    SetRotMatrix(&cloud_matrix);

 		    z_position = RotTransPers4(&coord[0], &coord[1], &coord[2],
		                &coord[3], (long *)&face.x0, (long *)&face.x1,
		                (long *)&face.x2, (long *)&face.x3, &p, &flag);

		   	GsSortPoly(&face, ot_ptr, z_position >> otzshift);

			ot_ptr = &Wotwin1[outbuf_idx];
		}

	}

	return;
}
/*		  
RIPPLE_SPLASH *init_ripple_struct(void)
{
	int				i;


	for (i = 0; i < MAX_RIPPLE_SPLASHES; i++)
	{
		if (!PSX_ripples[i].in_use)
		{
			PSX_ripples[i].in_use = TRUE;

			return &PSX_ripples[i];
		}
	}

	return NULL;
}

RIPPLE_CTRL_STRUCT *init_ripple_ctrl_struct(void)
{
	int				i;


	for (i = 0; i < MAX_RIPPLE_CTRL_STRUCTS; i++)
	{
		if (!PSX_ripple_ctrls[i].in_use)
		{
			PSX_ripple_ctrls[i].in_use = TRUE;

			return &PSX_ripple_ctrls[i];
		}
	}

	return NULL;
}

void init_ripple_structs(void)
{
	int				i;

	for (i = 0; i < MAX_RIPPLE_SPLASHES; i++)
	{
		PSX_ripples[i].in_use = FALSE;
		PSX_ripples[i].index = i;
	}

	return;
}

void init_ripple_ctrl_structs(void)
{
	int				i;

	for (i = 0; i < MAX_RIPPLE_CTRL_STRUCTS; i++)
	{
		PSX_ripple_ctrls[i].in_use = FALSE;
		PSX_ripple_ctrls[i].index = i;
	}

	return;
}


void init_explosion_ctrl_structs(void)
{
	int				i;

	for (i = 0; i < MAX_EXPLOSIONS; i++)
	{
		PSX_expns[i].in_use = FALSE;
		PSX_expns[i].index = i;
	}

	return;
}

EXPLOSION_CTRL_STRUCT	*init_new_explosion_ctrl_struct(void)
{
	int				i;


	for (i = 0; i < MAX_EXPLOSIONS; i++)
	{
		if (PSX_expns[i].in_use)
			continue;

		PSX_expns[i].in_use = TRUE;

		return &PSX_expns[i];
	}

	return NULL;
}

void init_explosion_flames(void)
{
	int				i;

	for (i = 0; i < MAX_EXP_FLAMES; i++)
	{
		PSX_expn_flames[i].in_use = FALSE;
		PSX_expn_flames[i].index = i;
	}

	return;
}

EXPLOSION_FLAME	*init_new_explosion_flame(void)
{
	int				i;


	for (i = 0; i < MAX_EXP_FLAMES; i++)
	{
		if (PSX_expn_flames[i].in_use)
			continue;

		PSX_expn_flames[i].in_use = TRUE;

		return &PSX_expn_flames[i];
	}

	return NULL;
}
*/
void set_identity_matrix(void)
{
	IDENTITY_MATRIX.m[0][0] = 6400;	  //aspect fix
	IDENTITY_MATRIX.m[0][1] = 0;
	IDENTITY_MATRIX.m[0][2] = 0;

	IDENTITY_MATRIX.m[1][0] = 0;
	IDENTITY_MATRIX.m[1][1] = 4096;
	IDENTITY_MATRIX.m[1][2] = 0;

	IDENTITY_MATRIX.m[2][0] = 0;
	IDENTITY_MATRIX.m[2][1] = 0;
	IDENTITY_MATRIX.m[2][2] = 4096;

	IDENTITY_MATRIX.t[0] = 0;
	IDENTITY_MATRIX.t[1] = 0;
	IDENTITY_MATRIX.t[2] = 0;

	return;
}

void set_camera_matrix(void)
{
	SVECTOR			camera_orientation;


	switch (STATE.numscreens)
	{
		case 1:
		{
			CAM_c = 0;

			camera_orientation.vx = CAM_came[CAM_c].vx >> 4;
			camera_orientation.vy = CAM_came[CAM_c].vy >> 4;
			camera_orientation.vz = CAM_came[CAM_c].vz >> 4;

		    RotMatrix(&camera_orientation, &CAM_camm[CAM_c]);

			break;
		}

		case 2:
		{
			CAM_c = 0;

			camera_orientation.vx = CAM_came[CAM_c].vx >> 4;
			camera_orientation.vy = CAM_came[CAM_c].vy >> 4;
			camera_orientation.vz = CAM_came[CAM_c].vz >> 4;

		    RotMatrix(&camera_orientation, &CAM_camm[CAM_c]);

			CAM_c = 1;

			camera_orientation.vx = CAM_came[CAM_c].vx >> 4;
			camera_orientation.vy = CAM_came[CAM_c].vy >> 4;
			camera_orientation.vz = CAM_came[CAM_c].vz >> 4;

		    RotMatrix(&camera_orientation, &CAM_camm[CAM_c]);

			break;
		}
	}

    SetTransMatrix(&IDENTITY_MATRIX);
	SetRotMatrix(&IDENTITY_MATRIX);
}

void PSX_init_start_finish_markers(void)
{
	int					index, i;
	OBJ_STRUCT			*obj_ptr;


	for (i = 0; i < 2; i++)
	{
	    index = OBJ_newobj(OTstart_pt);

	    if (index == NULLVAL)
	        return;                             //No more object structures.

	    obj_ptr = &OBJ_objects[index];

	    obj_ptr->p.vx = start_finish_pt[STATE.currlevel][i].vx << 8;
	    obj_ptr->p.vy = start_finish_pt[STATE.currlevel][i].vy << 8;
	    obj_ptr->p.vz = start_finish_pt[STATE.currlevel][i].vz << 8;

	    obj_ptr->r.vx = 0;
	    obj_ptr->r.vy = 0;
	    obj_ptr->r.vz = 0;

	    obj_ptr->v.vx = 0;
	    obj_ptr->v.vy = 0;
	    obj_ptr->v.vz = 0;

	    obj_ptr->owner = NULLVAL;       //Sets owner of torpedo

		obj_ptr->pkup_data.type = flare_smoke;
		obj_ptr->pkup_data.timer = 0;

		obj_ptr->m0 = 0;
		
//		OPSX_init_new_explosion(obj_ptr, flare_trail, OTHER_COLLISION);
	}
		
	return;
}

//START OF NEW EXPLOSION CODE

void PSX_init_flame_ctrl(OBJ_STRUCT *obj_ptr, char type, int collision_type)
{
	FLAME_CTRL				*ptr;
	int						temp_vel;	
	short					rot_val;


	ptr = create_flame_ctrl(type);

	if (ptr == NULL)
		return;
	else
		flame_ctrl_count++;

	if (collision_type == MESH_COLLISION)
	{
		ptr->position.vx = COL_inter.vx;
		ptr->position.vy = COL_inter.vy;
		ptr->position.vz = COL_inter.vz;
	}
	else
	{
		ptr->position.vx = obj_ptr->p.vx >> 8;
		ptr->position.vy = obj_ptr->p.vy >> 8;
		ptr->position.vz = obj_ptr->p.vz >> 8;
	}

		
	if (flame_ctrl_start == NULL)
		flame_ctrl_start = ptr;

	set_flame_ctrl_attributes(ptr, obj_ptr);

	return;
}

void PSX_process_flame_ctrls(void)
{
	FLAME_CTRL				*ptr, *next;
	short					i;
		

	if (flame_ctrl_start != NULL)
		ptr = flame_ctrl_start;
	else
		return;

	while (ptr != NULL)
	{
		ptr->timer ++;

		update_flame_ctrl(ptr);

		if ((!(ptr->timer % ptr->delay)) && (ptr->used != ptr->total))
		{
			for (i = 0; i < ptr->release_qty; i++)
			{
				init_flame(ptr);

				if (!ptr->dont_increment)
					ptr->used ++;
			}
		}

		if (ptr->used >= ptr->total)
		{
			next = (FLAME_CTRL *)ptr->next;	
			delete_flame_ctrl(ptr);

			if (ptr->type == pck_gas)
			{
				if (ptr->in_use == FALSE)
					OBJ_delobj(ptr->owner);
			}


			ptr = next;
		}
		else
//		{
//			if (ptr->type == craft_flames)
//			{
//				next = (FLAME_CTRL *)ptr->next;	
//	
//				if (OBJ_players[ptr->owner].armour_level > 350)		
//					delete_flame_ctrl(ptr);
//		
//				ptr = next;
//			}
//			else
				ptr = (FLAME_CTRL *)ptr->next;
//		}
	}

	return;
}

void update_flame_ctrl(FLAME_CTRL *ptr)
{
	OBJ_STRUCT				*obj_ptr;
	int						power_level;

	
	if (ptr->owner != NULLVAL)
	{
		obj_ptr = &OBJ_objects[ptr->owner];

		if (ptr->type == craft_flames)
		{
			ptr->position.vx = (obj_ptr->p.vx - obj_ptr->v.vx) >> 8;
			ptr->position.vy = (obj_ptr->p.vy - obj_ptr->v.vy) >> 8;
			ptr->position.vz = (obj_ptr->p.vz - obj_ptr->v.vz) >> 8;
		}
		else
		{
			ptr->position.vx = obj_ptr->p.vx >> 8;
			ptr->position.vy = obj_ptr->p.vy >> 8;
			ptr->position.vz = obj_ptr->p.vz >> 8;
		}

		ptr->rotation.vx = obj_ptr->r.vx >> 4;
		ptr->rotation.vy = obj_ptr->r.vy >> 4;
		ptr->rotation.vz = obj_ptr->r.vz >> 4;
	}
	else
	{
		ptr->position.vx += ptr->velocity.vx;
		ptr->position.vy += ptr->velocity.vy;
		ptr->position.vz += ptr->velocity.vz;
		
		if (ptr->decceleration != 0)
		{
			ptr->velocity.vx -= (ptr->velocity.vx / ptr->decceleration);
			ptr->velocity.vy -= (ptr->velocity.vy / ptr->decceleration);
			ptr->velocity.vz -= (ptr->velocity.vz / ptr->decceleration);
		}
	}
	
	if ((ptr->type == craft_flames) && (ptr->used != ptr->total))
	{
		power_level = OBJ_players[ptr->owner].armour_level / 50;

		ptr->release_qty = PSX_craft_flames[power_level][release_val];
		ptr->delay = PSX_craft_flames[power_level][delay_val];
	}
				
	return;
}

void init_flame(FLAME_CTRL *ptr)
{
	FLAME_ANIMATION				*flame;
	int							temp_vel;

	
	flame = create_flame_animation();

	if (flame == NULL)
		return;

	flame->length = 1;
	flame->type = ptr->type;

	flame->position.vx = ptr->position.vx;
	flame->position.vy = ptr->position.vy;
	flame->position.vz = ptr->position.vz;

	set_flame_attributes(flame, ptr, NULL, OTHER_COLLISION);

	return;
}

void PSX_init_flame_animation(OBJ_STRUCT *obj_ptr, int type, int collision_type)
{
	FLAME_ANIMATION				*flame;
	int							temp_vel;

	
	flame = create_flame_animation();

	if (flame == NULL)
		return;

	if (collision_type == MESH_COLLISION)
	{
		flame->position.vx = COL_inter.vx;
		flame->position.vy = COL_inter.vy;
		flame->position.vz = COL_inter.vz;
	}
	else
	{
		flame->position.vx = obj_ptr->p.vx >> 8;
		flame->position.vy = obj_ptr->p.vy >> 8;
		flame->position.vz = obj_ptr->p.vz >> 8;
	}
	   
	flame->length = 1;
	flame->type = type;

	set_flame_attributes(flame, NULL, obj_ptr, collision_type);

	return;
}

void PSX_process_flame_animations(void)
{
	GsOT						*ot_ptr;
	FLAME_ANIMATION				*ptr, *next, *flame_ptr[3];
	int							no_flames, display_loop;


	if (flame_animation_start != NULL)	
		ptr = flame_animation_start;	
	else
		return;

	while (ptr != NULL)
	{
		update_flame(ptr);

		if (ptr->uv_index >= MAX_EXP_TIMS)
		{
			next = (FLAME_ANIMATION *)ptr->next;
			delete_flame_animation(ptr);
			ptr = next;
		}
		else
			ptr = (FLAME_ANIMATION *)ptr->next;
	}			 

	ot_ptr = &Wotwin0[outbuf_idx];

	SetPolyFT4(&ft4);
	SetSemiTrans(&ft4, 1);

	for (display_loop = 0; display_loop < STATE.numscreens; display_loop ++)
	{
		if (flame_animation_start != NULL)	
			ptr = flame_animation_start;
		else
			continue;

		calculate_camera_matrix(display_loop);

		while (ptr != NULL)
		{
			no_flames = 1;

			flame_ptr[0] = ptr;
			flame_ptr[1] = NULL;
			flame_ptr[2] = NULL;

			ptr = (FLAME_ANIMATION *)ptr->next;
			
			if (ptr != NULL)
			{
				no_flames ++;

				flame_ptr[1] = ptr;
				ptr = (FLAME_ANIMATION *)ptr->next;

				if (ptr != NULL)
				{
					flame_ptr[2] = ptr;
					ptr = (FLAME_ANIMATION *)ptr->next;
				
					no_flames ++;
				}
			}
			
			display_flame_animations(flame_ptr, no_flames, display_loop, ot_ptr);
		}

		ot_ptr = &Wotwin1[outbuf_idx];
	}

	return;
}

void update_flame(FLAME_ANIMATION *ptr)
{
	OBJ_STRUCT					*obj_ptr;


	if (ptr->owner > NULLVAL)
	{
		obj_ptr = &OBJ_objects[ptr->owner];
	
		if ((ptr->type == missile_trail))// || (ptr->type == craft_flames))
		{
			ptr->position.vx += ((obj_ptr->v.vx >> 8) / ptr->decceleration);
			ptr->position.vy += ((obj_ptr->v.vy >> 8) / ptr->decceleration);
			ptr->position.vz += ((obj_ptr->v.vz >> 8) / ptr->decceleration);
		}
		else
		{
			ptr->position.vx += obj_ptr->v.vx >> 8;
			ptr->position.vy += obj_ptr->v.vy >> 8;
			ptr->position.vz += obj_ptr->v.vz >> 8;
		}
	}

	ptr->position.vx += ptr->velocity.vx;
	ptr->position.vy += ptr->velocity.vy;
	ptr->position.vz += ptr->velocity.vz;

	if (ptr->decceleration != 0)
	{
		if ((ptr->type != missile_trail))// && (ptr->type != craft_flames))
		{
			ptr->velocity.vx -= (ptr->velocity.vx / ptr->decceleration);
			ptr->velocity.vy -= (ptr->velocity.vy / ptr->decceleration);
			ptr->velocity.vz -= (ptr->velocity.vz / ptr->decceleration);
		}
	}

	ptr->timer ++;
	
	if (!(ptr->timer % ptr->delay))
	{
		ptr->uv_index += ptr->increment;
		
		if (ptr->timer >= ptr->smoke_point)
		{
			ptr->tp = GetTPage(0, 2, 704, 256);
			ptr->cl = GetClut(704, 464);

		    ptr->r = 64;
			ptr->g = 64;
		    ptr->b = 64;
		}
		else
		{
		    ptr->r -= ptr->colour_decrement;
			ptr->g -= ptr->colour_decrement;
		    ptr->b -= ptr->colour_decrement;
		}
	}

	return;
}

void display_flame_animations(FLAME_ANIMATION *ptr[3], int no_flames, int display_loop, GsOT *ot_ptr)
{
	#define				MAX_SCREEN_POLY_SIZE		80

	long				otz, flag, screen_pos[3], i;
	char				in_range[3];
	short				screen_x, screen_y;
	SVECTOR				origin[3];
	int					width, height, z_inv;
	long				z_addr[3];


	for (i = 0; i < no_flames; i++)
	{
		in_range[i] = clip_point_short(&ptr[i]->position, display_loop);
	
		if (in_range[i])
		{
			origin[i].vx = ptr[i]->position.vx - CAM_camp[display_loop].vx;
			origin[i].vy = ptr[i]->position.vy - CAM_camp[display_loop].vy;
			origin[i].vz = ptr[i]->position.vz - CAM_camp[display_loop].vz;
		}
	}

	if (no_flames == 1)
   		gte_ldv3(&origin[0], &origin[0], &origin[0]);		   //svectors loaded in (world pos)
	else
	{
		if (no_flames == 2)
   			gte_ldv3(&origin[0], &origin[1], &origin[0]);		   //svectors loaded in (world pos)
		else
	   		gte_ldv3(&origin[0], &origin[1], &origin[2]);		   //svectors loaded in (world pos)
	}

	gte_rtpt(); //rot trans pers
	gte_stsxy3(&screen_pos[0], &screen_pos[1], &screen_pos[2]);	   //shorts stored out (screen pos)
	gte_stflg(&flag);
		
//	if((flag&0x80000000))		   // clip if off screen
//		return;

	gte_stsz3c(z_addr);	  //z value
//	printf("z1 = %d : z2 = %d : z3 = %d\n", z_addr[0], z_addr[1], z_addr[2]);
	
	gte_stszotz(&otz);	  //z value

	for (i = 0; i < no_flames; i++)
	{
		if (in_range[i] == FALSE)
			continue;

		z_addr[i] >>= 2;

		if (z_addr[i] <= 0)
			continue;

		screen_x = screen_pos[i];
		screen_y = screen_pos[i] >> 16;

//		if (ptr[i]->type == pck_flame)
//		{
//			screen_x = screen_x >> 1;
//			screen_y = screen_y >> 1;
//		}

		z_inv = (65536 / z_addr[i]) * 20;

		width = (((EXP_fire_uv_table[ptr[i]->uv_index][4] * 200) / 10) * z_inv) >> 16;
		height = (((EXP_fire_uv_table[ptr[i]->uv_index][5] * 200) / 10) * z_inv) >> 16;

		width = (width * 25) >> 4;
		width *= ptr[i]->length;
		height *= ptr[i]->length;
		
		if (ptr[i]->type == craft_flames)
		{
			width /= 4;
			height /= 4;
		}
		else
		{
			width /= 2;
			height /= 2;
		}

		if (width > MAX_SCREEN_POLY_SIZE)
			continue;
		
		ft4.r0 = ptr[i]->r;	
		ft4.g0 = ptr[i]->g; 	
		ft4.b0 = ptr[i]->b; 	

	    ft4.u0 = EXP_fire_uv_table[ptr[i]->uv_index][0];
	    ft4.v0 = EXP_fire_uv_table[ptr[i]->uv_index][1];

	    ft4.u1 = EXP_fire_uv_table[ptr[i]->uv_index][2];
	    ft4.v1 = EXP_fire_uv_table[ptr[i]->uv_index][1];

	    ft4.u2 = EXP_fire_uv_table[ptr[i]->uv_index][0];
	    ft4.v2 = EXP_fire_uv_table[ptr[i]->uv_index][3];

	    ft4.u3 = EXP_fire_uv_table[ptr[i]->uv_index][2];
	    ft4.v3 = EXP_fire_uv_table[ptr[i]->uv_index][3];

		ft4.tpage = ptr[i]->tp;
		ft4.clut = ptr[i]->cl;
		
		ft4.x0 = screen_x - width;		
		ft4.y0 = screen_y - height;		

		ft4.x1 = screen_x + width;		
		ft4.y1 = screen_y - height;		

		ft4.x2 = screen_x - width;		
		ft4.y2 = screen_y + height;		

		ft4.x3 = screen_x + width;		
		ft4.y3 = screen_y + height; 		

		if ((abs((ft4.x3 - ft4.x0)) < 100) && (abs((ft4.y3 - ft4.y0)) < 100))
			GsSortPoly(&ft4, ot_ptr, z_addr[i] >> otzshift); // time to 'sort' it out!
	}
		 
	return;
}

FLAME_CTRL *create_flame_ctrl(int type)
{
	int					i, oldest_timer;
	FLAME_CTRL			*ptr, *delete_node;

	
	oldest_timer = 1000;

	for (i = 0, ptr = &PSX_flame_ctrls[0]; i < MAX_EXPLOSIONS; i++, ptr++)
	{
		if (ptr->in_use)
		{
			if (ptr->reserved == FALSE)
			{
				if (ptr->timer < oldest_timer)
				{
					oldest_timer = ptr->timer;
					delete_node = ptr;
				}
			}

			continue;
		}

		ptr->in_use = TRUE;
		ptr->next = NULL;
		ptr->prev = (unsigned long *)flame_ctrl_current;

		ptr->type = type;
		ptr->timer = 0;
		ptr->used = 0;

		flame_ctrl_current->next = (unsigned long *)ptr;
		flame_ctrl_current = ptr;

		return ptr;
	}				  

//	return delete_node;
	return NULL;
}

void delete_flame_ctrl(FLAME_CTRL *ptr)
{
  	FLAME_CTRL				*prev, *next;


	if ((ptr->prev != NULL) && (ptr->next != NULL))
	{
		prev = (FLAME_CTRL *)ptr->prev;
		next = (FLAME_CTRL *)ptr->next;

		prev->next = (unsigned long *)next;
		next->prev = (unsigned long *)prev;

		if (flame_ctrl_current == ptr)
  			flame_ctrl_current = (FLAME_CTRL *)ptr->prev;
	}

	if ((ptr->prev != NULL) && (ptr->next == NULL))
	{
		prev = (FLAME_CTRL *)ptr->prev;
		prev->next = NULL;

		if (flame_ctrl_current == ptr)
			flame_ctrl_current = (FLAME_CTRL *)ptr->prev;
	}

	if ((ptr->prev == NULL) && (ptr->next != NULL))
	{
		next = (FLAME_CTRL *)ptr->next;
		next->prev = NULL;

		flame_ctrl_start = next;

		if (flame_ctrl_current == ptr)
			flame_ctrl_current = NULL;
	}

	if ((ptr->prev == NULL) && (ptr->next == NULL))
	{
		flame_ctrl_start = NULL;

		if (flame_ctrl_current == ptr)
			flame_ctrl_current = NULL;
	}

	ptr->in_use = FALSE;
 	flame_ctrl_count --;
	
	return;
}

FLAME_ANIMATION *create_flame_animation(void)
{
	int					i;
	FLAME_ANIMATION		*ptr;

	
	for (i = 0, ptr = &PSX_flames[0]; i < MAX_EXP_FLAMES; i++, ptr++)
	{
		if (ptr->in_use)
			continue;

		ptr->in_use = TRUE;
		ptr->owner = NULLVAL;
		
		ptr->next = NULL;
		ptr->prev = (unsigned long *)flame_animation_current;

		ptr->timer = 0;
		ptr->uv_index = 0;

		if (flame_animation_start == NULL)
			flame_animation_start = ptr;

		flame_animation_current->next = (unsigned long *)ptr;
		flame_animation_current = ptr;

		return ptr;
	}				  

	return NULL;
}

void delete_flame_animation(FLAME_ANIMATION *ptr)
{
	FLAME_ANIMATION				*prev, *next;


	if ((ptr->prev != NULL) && (ptr->next != NULL))
	{
		prev = (FLAME_ANIMATION *)ptr->prev;
		next = (FLAME_ANIMATION *)ptr->next;

		prev->next = (unsigned long *)next;
		next->prev = (unsigned long *)prev;

		if (flame_animation_current == ptr)
			flame_animation_current = prev;
	}

	if ((ptr->prev != NULL) && (ptr->next == NULL))
	{
		prev = (FLAME_ANIMATION *)ptr->prev;
		prev->next = NULL;

		if (flame_animation_current == ptr)
			flame_animation_current = prev;
	}

	if ((ptr->prev == NULL) && (ptr->next != NULL))
	{
		next = (FLAME_ANIMATION *)ptr->next;
		next->prev = NULL;

		flame_animation_start = next;

		if (flame_animation_current == ptr)
			flame_animation_current = NULL;
	}

	if ((ptr->prev == NULL) && (ptr->next == NULL))
	{
		flame_animation_start = NULL;

		if (flame_animation_current == ptr)
			flame_animation_current = NULL;
	}

	ptr->in_use = FALSE;
	
	return;
}

void PSX_init_all_flame_ctrls(void)
{
	int					i;
	
	
	flame_ctrl_start = NULL;
	flame_ctrl_current = NULL;

	flame_animation_start = NULL;
	flame_animation_current = NULL;

	flame_ctrl_count = 0;

	for (i = 0; i < MAX_EXPLOSIONS; i++)
	{
		PSX_flame_ctrls[i].in_use = FALSE;
		PSX_flame_ctrls[i].index = i;
		PSX_flame_ctrls[i].owner = NULLVAL;

		PSX_flame_ctrls[i].next = NULL;
		PSX_flame_ctrls[i].prev = NULL;
	}

	for (i = 0; i < MAX_EXP_FLAMES; i++)
	{
		PSX_flames[i].in_use = FALSE;
		PSX_flames[i].index = i;

		PSX_flames[i].next = NULL;
   		PSX_flames[i].prev = NULL;
	}

	return;
}	

void calculate_camera_matrix(int display_loop)
{
	SVECTOR				camera_orientation;


	camera_orientation.vx = CAM_came[display_loop].vx >> 4;
	camera_orientation.vy = CAM_came[display_loop].vy >> 4;
	camera_orientation.vz = CAM_came[display_loop].vz >> 4;

    RotMatrix(&camera_orientation, &camera_matrix);

	camera_position.vx = CAM_camp[display_loop].vx;
	camera_position.vy = CAM_camp[display_loop].vy;
	camera_position.vz = CAM_camp[display_loop].vz;

	ApplyMatrixSV(&camera_matrix, &camera_position, &camera_position);	

	camera_matrix.t[0] = 0;//-(camera_position.vx);// * 25) >> 4;
	camera_matrix.t[1] = 0;//-camera_position.vy;
	camera_matrix.t[2] = 0;//-camera_position.vz;

	camera_matrix.m[0][0] = (camera_matrix.m[0][0] * 25) >> 4;	
	camera_matrix.m[0][1] = (camera_matrix.m[0][1] * 25) >> 4;	
	camera_matrix.m[0][2] = (camera_matrix.m[0][2] * 25) >> 4;	

	SetRotMatrix(&camera_matrix);
//	SetTransMatrix(&camera_matrix);
	SetTransMatrix(&IDENTITY_MATRIX);
		
	return;
}

void PSX_make_scenic_effects_for_kam_and_co(void)
{
	OBJ_STRUCT					*obj_ptr;
	int							index;
//	LEVEL_EFFECT				*ptr;


	index = OBJ_newobj(OTlevel_fx);

	if (index == NULLVAL)
		printf("shit no level fx\n");
	
	obj_ptr = &OBJ_objects[index];

	obj_ptr->v.vx = 0;
	obj_ptr->v.vy = 0;
	obj_ptr->v.vz = 0;

	obj_ptr->r.vx = 0;
	obj_ptr->r.vy = 0;
	obj_ptr->r.vz = 0;

	obj_ptr->p.vx = 21979 << 8;
	obj_ptr->p.vy = 0;
	obj_ptr->p.vz = 27078 << 8;
		
//	PSX_init_flame_ctrl(obj_ptr, lava_steam, OTHER_COLLISION);

/*	
	for (i = 0; i < ptr->total; i++)
	{
		obj_ptr->p.vx = ptr->origin.vx + ptr->direction.vx;
		obj_ptr->p.vy = ptr->origin.vy + ptr->direction.vy * scaler;
		obj_ptr->p.vz = ptr->origin.vz + ptr->direction.vz;
	
		PSX_init_flame_ctrl(obj_ptr, lava_steam, OTHER_COLLISION);
*/
	OBJ_delobj(index);

	return;
}

void set_flame_ctrl_attributes(FLAME_CTRL *ptr, OBJ_STRUCT *obj_ptr)
{
	short						rot_val;
	int							temp_vel;


	switch (ptr->type)	
	{
		case pck_rocket:
		case rocket_explosion:
		
			ptr->reserved = FALSE;
			ptr->dont_increment = FALSE;

			ptr->owner = NULLVAL;

			ptr->delay = 2;
			ptr->total = 5;
			ptr->release_qty = 1;

			ptr->rotation.vx = SDRAND(1024) - 1536;
			
			rot_val = obj_ptr->r.vy >> 4;

			if (rot_val < 0)
				ptr->rotation.vy = 2048 - rot_val;
			
			else
				ptr->rotation.vy = -2048 + rot_val;
										  			
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = (5 * rsin(ptr->rotation.vx)) >> 12;
			temp_vel = (5 * rcos(ptr->rotation.vx)) >> 12;

			ptr->velocity.vx = (temp_vel * rsin(ptr->rotation.vy)) >> 12;
			ptr->velocity.vz = (temp_vel * rcos(ptr->rotation.vy)) >> 12;

			ptr->decceleration = 5;

			break;

		case lava_steam:
		
			ptr->reserved = TRUE;
			ptr->dont_increment = TRUE;

			ptr->owner = NULLVAL;

			ptr->delay = 3;
			ptr->total = 5;
			ptr->release_qty = 1;

			ptr->rotation.vx = SDRAND(1024) - 1536;
			
			rot_val = obj_ptr->r.vy >> 4;

			if (rot_val < 0)
				ptr->rotation.vy = 2048 - rot_val;
			
			else
				ptr->rotation.vy = -2048 + rot_val;
										  			
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = 0;//(5 * rsin(ptr->rotation.vx)) >> 12;
			temp_vel = 0;//(5 * rcos(ptr->rotation.vx)) >> 12;

			ptr->velocity.vx = 0;//(temp_vel * rsin(ptr->rotation.vy)) >> 12;
			ptr->velocity.vz = 0;//(temp_vel * rcos(ptr->rotation.vy)) >> 12;

			ptr->decceleration = 0;//5;

			break;

		case pck_missile:
		case pck_mine:
		case pck_cluster:
		case torpedo_explosion:
		
			ptr->reserved = FALSE;
			ptr->dont_increment = FALSE;

			ptr->owner = NULLVAL;

			ptr->delay = 2;
			ptr->total = 10;
			ptr->release_qty = 1;

			ptr->rotation.vx = SDRAND(1024) - 1536;
			
			rot_val = obj_ptr->r.vy >> 4;

			if (rot_val < 0)
				ptr->rotation.vy = 2048 - rot_val;
			
			else
				ptr->rotation.vy = -2048 + rot_val;
										  			
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = (5 * rsin(ptr->rotation.vx)) >> 12;
			temp_vel = (5 * rcos(ptr->rotation.vx)) >> 12;

			ptr->velocity.vx = (temp_vel * rsin(ptr->rotation.vy)) >> 12;
			ptr->velocity.vz = (temp_vel * rcos(ptr->rotation.vy)) >> 12;

			ptr->decceleration = 5;

			break;

		case missile_trail:
		
			obj_ptr->fx_id = ptr->index;
	
			ptr->reserved = TRUE;
			ptr->dont_increment = TRUE;

			ptr->owner = obj_ptr->index;

			ptr->delay = 1;
			ptr->total = 1;
			ptr->release_qty = 1;

			ptr->rotation.vx = obj_ptr->r.vx >> 4;
			ptr->rotation.vy = obj_ptr->r.vy >> 4;
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = 0;
			ptr->velocity.vx = 0;
			ptr->velocity.vz = 0;

			ptr->decceleration = 0;

			break;

		case pck_gas:
		
			obj_ptr->fx_id = ptr->index;
	
			ptr->reserved = TRUE;
			ptr->dont_increment = FALSE;

			ptr->owner = NULLVAL;

			ptr->delay = 2;
			ptr->total = 250;
			ptr->release_qty = 2;

			ptr->rotation.vx = obj_ptr->r.vx >> 4;
			ptr->rotation.vy = obj_ptr->r.vy >> 4;
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = 0;
			ptr->velocity.vx = 0;
			ptr->velocity.vz = 0;

			ptr->decceleration = 0;

			break;

		case craft_flames:
		
			obj_ptr->fx_id = ptr->index;
	
			ptr->reserved = TRUE;
			ptr->dont_increment = TRUE;

			ptr->owner = obj_ptr->index;

			ptr->delay = 3;
			ptr->total = 1;
			ptr->release_qty = 1;

			ptr->rotation.vx = obj_ptr->r.vx >> 4;
			ptr->rotation.vy = obj_ptr->r.vy >> 4;
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = 0;
			ptr->velocity.vx = 0;
			ptr->velocity.vz = 0;

			ptr->decceleration = 0;

			break;

		case craft_explosion:
		
			ptr->reserved = TRUE;
			ptr->dont_increment = FALSE;

			ptr->owner = obj_ptr->owner;

			ptr->delay = 1;
			ptr->total = 25;
			ptr->release_qty = 1;

			ptr->rotation.vx = SDRAND(1024) - 1536;
			
			rot_val = obj_ptr->r.vy >> 4;

			if (rot_val < 0)
				ptr->rotation.vy = 2048 - rot_val;
			
			else
				ptr->rotation.vy = -2048 + rot_val;
										  			
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = (-50 * rsin(ptr->rotation.vx)) >> 12;
			temp_vel = (-50 * rcos(ptr->rotation.vx)) >> 12;

			ptr->velocity.vx = (temp_vel * rsin(ptr->rotation.vy)) >> 12;
			ptr->velocity.vz = (temp_vel * rcos(ptr->rotation.vy)) >> 12;

			ptr->decceleration = 2;

			break;

		case pck_ion:
		
			ptr->reserved = TRUE;
			ptr->dont_increment = FALSE;

			ptr->owner = NULLVAL;

			ptr->delay = 1;
			ptr->total = 25;
			ptr->release_qty = 1;

			ptr->rotation.vx = SDRAND(1024) - 1536;
			
			rot_val = obj_ptr->r.vy >> 4;

			if (rot_val < 0)
				ptr->rotation.vy = 2048 - rot_val;
			
			else
				ptr->rotation.vy = -2048 + rot_val;
										  			
			ptr->rotation.vz = obj_ptr->r.vz >> 4;

			ptr->velocity.vy = (0 * rsin(ptr->rotation.vx)) >> 12;
			temp_vel = (0 * rcos(ptr->rotation.vx)) >> 12;

			ptr->velocity.vx = (temp_vel * rsin(ptr->rotation.vy)) >> 12;
			ptr->velocity.vz = (temp_vel * rcos(ptr->rotation.vy)) >> 12;

			ptr->decceleration = 0;

			break;
	}

	return;
}

void set_flame_attributes(FLAME_ANIMATION *flame, FLAME_CTRL *ptr, OBJ_STRUCT *obj_ptr, int collision_type)
{
	short						rot_val;
	int							temp_vel, owner, vel, power_level;
	vec16						rotation;


	if (obj_ptr == NULL)
	{
		rotation.vx = ptr->rotation.vx;
		rotation.vy = ptr->rotation.vy;
		rotation.vz = ptr->rotation.vz;
		
		owner = ptr->owner;
	}
	else
	{
		rotation.vx = obj_ptr->r.vx >> 4;
		rotation.vy = obj_ptr->r.vy >> 4;
		rotation.vz = obj_ptr->r.vz >> 4;

		owner = obj_ptr->index;
	}

	if (collision_type == MESH_COLLISION)
		owner = NULLVAL;

	switch (flame->type)
	{
		case pck_rocket:
		case rocket_explosion:
				
			flame->owner = NULLVAL;

			flame->delay = 1;
			flame->increment = 1;
			flame->smoke_point = 18;
			
			flame->r = 48;
			flame->g = 48;
			flame->b = 48;

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 1, 704, 256);
			flame->cl = GetClut(704, 462);
			
			flame->rotation.vx = SDRAND(1024) - 1536;
			flame->rotation.vy = rotation.vy + (SDRAND(2048) - 1024);
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (20 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (20 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 25;		

			break;

		case pck_missile:
		case pck_cluster:
		case pck_mine:
		case torpedo_explosion:
				
			flame->owner = NULLVAL;

			flame->delay = 1;
			flame->increment = 1;
			flame->smoke_point = 18;
			
			flame->r = 48;
			flame->g = 48;
			flame->b = 48;

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 1, 704, 256);
			flame->cl = GetClut(704, 462);
			
			flame->rotation.vx = SDRAND(1024) - 1536;
			flame->rotation.vy = rotation.vy + (SDRAND(2048) - 1024);
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (20 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (20 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 25;		

			break;

		case lava_steam:
				
			flame->owner = NULLVAL;

			flame->delay = 1;
			flame->increment = 1;
			flame->smoke_point = 100;
			
			flame->r = 148;
			flame->g = 148;
			flame->b = 148;

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 1, 704, 256);
			flame->cl = GetClut(704, 453);
			
			flame->rotation.vx = SDRAND(1024) - 1536;
			flame->rotation.vy = rotation.vy + (SDRAND(2048) - 1024);
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (20 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (20 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 25;		

			break;

		case missile_trail:
				
			flame->owner = owner;

			flame->delay = 1;
			flame->increment = SDRAND(2) + 1;
			flame->smoke_point = 18;
			
			flame->r = 
			flame->g = 
			flame->b = SDRAND(128);

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 3, 704, 256);
			flame->cl = GetClut(704, 462);

			flame->rotation.vx = rotation.vx;
			flame->rotation.vy = rotation.vy;
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (100 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (100 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 2;		

			break;

		case pck_flame:
				
			flame->owner = owner;

			flame->delay = 1;
			flame->increment = SDRAND(3) + 1;
			flame->smoke_point = 18;
			
			flame->r = 
			flame->g = 
			flame->b = SDRAND(128);

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 3, 704, 256);
			flame->cl = GetClut(704, 462);

			flame->rotation.vx = rotation.vx;
			flame->rotation.vy = rotation.vy;
			flame->rotation.vz = rotation.vz;

			vel = SDRAND(100) + 50;

			flame->velocity.vy = (vel * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (vel * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 10;		

			break;

		case pck_gas:
				
			flame->owner = NULLVAL;

			flame->delay = 1;
			flame->increment = SDRAND(2) + 1;
			flame->smoke_point = 27;
			
			flame->r = 
			flame->g = 
			flame->b = 8 + SDRAND(24);

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 1, 704, 256);
			flame->cl = GetClut(704, 469);

			flame->rotation.vx = 0;
			flame->rotation.vy = SDRAND(4096);
			flame->rotation.vz = rotation.vz;

			vel = SDRAND(45) + 20;

			flame->velocity.vy = (vel * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (vel * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 4;		

			break;

		case bullet_explosion:
				
			flame->owner = NULLVAL;

			flame->delay = 1;
			flame->increment = 1;
			flame->smoke_point = 18;
			
			flame->r = 
			flame->g = 
			flame->b = SDRAND(64) + 128;

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 3, 704, 256);
			flame->cl = GetClut(704, 462);
			
			flame->rotation.vx = rotation.vx;
			flame->rotation.vy = rotation.vy;
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (0 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (0 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 0;		

			break;

		case craft_flames:
				
			power_level = OBJ_players[owner].armour_level / 50;
			
			flame->owner = OBJ_players[owner].plyrnum;

			flame->delay = 1;
			flame->increment = PSX_craft_flames[power_level][inc_val];
			flame->smoke_point = 18;
			
			flame->r = 
			flame->g = 
			flame->b = PSX_craft_flames[power_level][rgb_val];

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = PSX_craft_flames[power_level][tp_val];
			flame->cl = PSX_craft_flames[power_level][cl_val];

			flame->rotation.vx = rotation.vx + SDRAND(256);
			flame->rotation.vy = rotation.vy + (SDRAND(512) - 256);
			flame->rotation.vz = rotation.vz;

			vel = OBJ_players[owner].speed >> 3;
			
			if (vel < 25)
				vel = 25;

			flame->velocity.vy = (-vel * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (-vel * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 10;		

			break;

		case craft_explosion:
				
			flame->owner = NULLVAL;

			flame->delay = 1;
			flame->increment = 1;
			flame->smoke_point = 18;
			
			flame->r = 48;
			flame->g = 48;
			flame->b = 48;

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 1, 704, 256);
			flame->cl = GetClut(704, 462);
			
			flame->rotation.vx = SDRAND(1024) - 1536;
			flame->rotation.vy = rotation.vy + (SDRAND(2048) - 1024);
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (20 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (20 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 25;		

			break;

		case pck_ion:
				
			flame->owner = owner;

			flame->delay = 1;
			flame->increment = 1;
			flame->smoke_point = 27;
			
			flame->r = 48;
			flame->g = 48;
			flame->b = 48;

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 1, 704, 256);
			flame->cl = GetClut(704, 453);
			
			flame->rotation.vx = SDRAND(1024) - 1536;
			flame->rotation.vy = rotation.vy + (SDRAND(2048) - 1024);
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (20 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (20 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 25;		

			break;

		case flare_trail:
				
			flame->owner = owner;

			flame->delay = 1;
			flame->increment = 1;
			flame->smoke_point = 18;
			
			flame->r = 
			flame->g = 
			flame->b = SDRAND(64);

			flame->colour_decrement = (flame->r / (26 / flame->increment));

			flame->tp = GetTPage(0, 2, 704, 256);
			flame->cl = GetClut(704, 464);

			flame->rotation.vx = rotation.vx;
			flame->rotation.vy = rotation.vy + (SDRAND(512) - 256);
			flame->rotation.vz = rotation.vz;

			flame->velocity.vy = (100 * rsin(flame->rotation.vx)) >> 12;
			temp_vel = (100 * rcos(flame->rotation.vx)) >> 12;

			flame->velocity.vx = (temp_vel * rsin(flame->rotation.vy)) >> 12;
			flame->velocity.vz = (temp_vel * rcos(flame->rotation.vy)) >> 12;
			
			flame->decceleration = 2;		

			break;

	}

	return;
}

