/*******************************************************************
 *
 *    DESCRIPTION: Smart Dog PSX lib number types
 *
 *    AUTHOR: D.Saunders
 *
 *    HISTORY: started Mon Jan 5 1998    
 *
 *******************************************************************/

#ifndef _SDTYPES_
#define _SDTYPES_

typedef long	FIXED;  		// 0x00018000 is 1.5  
typedef	short	ANGLE;			// 0x10000 is 360 deg 

typedef unsigned long		Uint32;
typedef unsigned short		Uint16;
typedef unsigned char		Uint8;	
typedef signed char	Sint8;	
typedef short		Sint16;
typedef long		Sint32;

extern	Sint32	MulDS(Sint32 a,Sint32 b);	// assembler routines >PS specific
extern	Sint32	GetHiDs();




typedef struct {				// position,rotation,scale for an object
	FIXED 	xpos;
	FIXED 	ypos;
	FIXED 	zpos;
	ANGLE	xrot;
	ANGLE	yrot;
	ANGLE	zrot;
	FIXED	xscl;
	FIXED	yscl;
	FIXED	zscl;
}OBJ_PRS;


#if NEWPAD

#define u8	unsigned char
#define s8	signed char
#define u16	unsigned short

typedef struct
	{
	u8 Id;		// id of peripheral type pluged in 
	u16 pull;	// when let go 
	u16 push;	// when push 
	u16 push2;	// when push 
	u16 data;	// raw 
	u16 old;	// last frame raw 
	u8 In;		// is a pad pluged into this port 
//	u8 InCount;	// count to 5 bad reads before flag pad pulled out !
	s8 An1X;	// analog 1 x  -128 > 0	< 127
	s8 An1Y;	
	s8 An2X;	
	s8 An2Y;
	
    u8 actAligned;
	u16 rumble[2];
	}APAD;

#undef u8
#undef u16
#undef s8

extern APAD SD_PadData[8];

#else

typedef struct {	
Uint8	Id;		// id of peripheral type pluged in 
Uint16	pull;	// when let go 
Uint16	push;	// when push 
Uint16	push2;	// when push 
Uint16	data;	// raw 
Uint16	old;	// last frame raw 
Uint8	In;		// is a pad pluged into this port 
Uint8	InCount;	// count to 5 bad reads before flag pad pulled out !
Sint8 	An1X;	// analog 1 x  -128 > 0	< 127
Sint8 	An1Y;	
Sint8 	An2X;	
Sint8 	An2Y;	
} APAD;

#endif

// macros

// 1 deg is 182 so 360 is 0x10000
#define ONEDEG 182

// toFIXED(0.5) will turn 0.5 into 0x00008000 1:15:16 fixed point
#define toFIXED(x)		((x)*65536)

// F2L takes a fixed (0x00108000) and turns in into a long (16)  the .5	is lost
#define	F2L(x)			(Sint32)((short)(x/65536))

// F2L takes a fixed (0x00108000) and turns in into a short (16) the .5	is lost	
#define	F2S(x)			(short)(x/65536)		

// SD_Sin takes a ANGLE type ie (0x10000) is 360 deg
// and returns a fixed (0x00008000) is 0.5 type
#define SD_Sin(x)		(    (rsin( (Sint32)(x>>4)) )   <<4)

// SD_Sin takes a ANGLE type ie (0x10000) is 360 deg
// and returns a fixed (0x00008000) is 0.5 type
#define SD_Cos(x)		(    (rcos( (Sint32)(x>>4)) )   <<4)

// SD_Mul takes two fixed numbers and multiplies them returning a fixed 
#define SD_Mul(x,y)		(MulDS((x),(y)))

// SD_Div takes X(fixed) and divides it by Y(fixed) returning a fixed
#define SD_Div(x,y)	(Sint32) (65536*( (float) ((float)(x))/((float)(y)) )   )


// SD_SquareRoot takes X(fixed) finds square root and returns a fixed 
//#define	SD_SquareRoot(x)	(SquareRoot0(x)<<8)
#define	SD_SquareRoot(x)	(csqrt(x)<<2)			// new more accurate

// SD_SquareRootI takes X(Integer 32bits !) finds square root and returns a fixed 
#define	SD_SquareRootI(x)	(SquareRoot0(x)<<16)

// SD_SquareRootNF takes X(Integer 32bits !) finds square root and returns a Integer 32 bits
#define	SD_SquareRootNF(x)	(SquareRoot0(x))

// SD_Atan2 take x and y size of right angled triangle (number format for x,y must be same)
// and returns angle in fixed (0x00010000) is 360 deg
//
// 			 /|					  |
//			/ |				  -180|	  -
//		   /  |	y			  ----+----	 0
//		  /A  |				   180|	  +
//		  -----					  |
//			x
#define SD_Atan2(y,x)	(ratan2(y,x)<<4)

// return non negative! 
#define SD_Abs(x)		(abs(x))


// old sega macros div is back to front ie y div by x !
//#define slDivFX(x,y)	(Sint32) (65536*( (float) ((float)(y))/((float)(x)) )   )
//#define slMulFX(x,y)	(MulDS((x),(y)))   // replaced with SD_Mul



#endif // _SDTYPES_
