#include "header.h"
#include "render.h"
#include "camera.h"
#include "objects.h"
#include "collide.h"
#include <sys/types.h>
#include <libetc.h>
#include <libgte.h>
#include <libgpu.h>
#include <libgs.h>
#include <libsn.h>
#include <gtemac.h>
#include <inline_c.h>
#include <libsn.h>
#include <stdio.h>
#include "sdlib\sdtypes.h"
#include "sdlib\sdfunc.h"
#include "ctrller.h"
#include "gameloop.h"
#include "state.h"
#include "wake.h"
#include "psx\mathpsx.h"

#define	IGNOREFLAG		0
#define __DEBUG_WAKE	0

int globalWakeDisable = 0;

// 20 - 100

// Biggest splash when lower or at this value
#define	WAKE_SHIPHEIGHTMIN		20

// No wake, ship to far from water
#define	WAKE_SHIPHEIGHTMAX		52

// Range of min - max height
#define	WAKE_SHIPHEIGHTRANGE	(WAKE_SHIPHEIGHTMAX-WAKE_SHIPHEIGHTMIN)

PLWAKE wake[6];

const vec16 wakeOffset[8] [3] =
    {
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        },
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        },
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        },
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        },
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        },
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        },
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        },
        {
            { 80<<8,  0<<8,-30<<8},
            {  0<<8,  0<<8,-30<<8},
            {-80<<8,  0<<8,-30<<8}
        }
    };

const vec32 sprayOffset[8] =
	{
/*
		{-120<<8,   0<<8, -70<<8},
		{ -50<<8,   0<<8,   0<<8},
		{  50<<8,   0<<8,   0<<8},
		{ 120<<8,   0<<8, -70<<8},
		{-155<<8, -50<<8, -35<<8},
		{ -85<<8, -50<<8,  35<<8},
		{  85<<8, -50<<8,  35<<8},
		{ 155<<8, -50<<8, -35<<8}
*/
		{-120<<8,   0<<8, -70<<8},
		{ -50<<8,   0<<8,   0<<8},
		{  50<<8,   0<<8,   0<<8},
		{ 120<<8,   0<<8, -70<<8},
		{-155<<8,   0<<8, -35<<8},
		{ -85<<8,   0<<8,  35<<8},
		{  85<<8,   0<<8,  35<<8},
		{ 155<<8,   0<<8, -35<<8}

    };

vec32 sprayRot[8];
vec32 sprayRotS[8];
vec32 sprayRotSC[8];

const int sprayUV[4] [2] =
	{
    {64+64,150},
    {79+64,150},
    {64+64,181},
    {79+64,181}
    };

#if __DEBUG_WAKE

enum{
	ERR_noerror = 0,
	ERR_invalidinitparam = 1
	};

int regerr = ERR_noerror;

#endif

extern GsOT Wotwin0[2];
extern GsOT Wotwin1[2];
extern GsOT *ptrwot;
extern int outbuf_idx;
extern mat16 objm;
extern vec32 pos;
extern MATRIX IDENTITY_MATRIX;

int offsy = 0;

/*

    v0----v1v2----v3
     | \ p0 |p1  / |
     |   \  |  /   |
     |p2   \|/  p3 |
    v4----v5v6----v7

*/

static s8 wakeUV[8] [2] =
    {
//  {16,94},        {31,94},        {32,94},        {47,94},
//  {16,125},       {31,125},       {32,125},       {47,125},

    {16,94},        {16,109},       {16,110},       {16,125},
    {47,94},        {47,109},       {47,110},       {47,125}
    };

POLY_GT3 gt3[4];
POLY_GT4 gt4;

void initWake(int p)
	{
	register PLWAKE *objw;
	register int i;
	
#if __DEBUG_WAKE
	if((p<0)||(p>7))
		{
		if(regerr==ERR_noerror)
			regerr = ERR_invalidinitparam;
		return;
		}
#endif

	objw = &wake[p];
	objw->inited = 1;

	for(i=0;i<10;i++)
		objw->trail[i].valid[0] =
		objw->trail[i].valid[1] =
		objw->trail[i].valid[2] = 0;

#if __DEBUG_WAKE
	printf("WakeInit %d (pl==%p, st==%p, wa==%p)\n",p,&OBJ_players[p],OBJ_players[p].pobj,objw);
#endif

	for(i=0;i<4;i++)
		{	
		SetPolyGT3(&gt3[i]);
		setSemiTrans(&gt3[i],1);
	
		gt3[i].tpage = GetTPage(1,1,704,256);
		gt3[i].clut = GetClut(768,375);
		}

	gt3[0].u0 = wakeUV[0] [0];	gt3[0].v0 = wakeUV[0] [1];
	gt3[0].u1 = wakeUV[1] [0];	gt3[0].v1 = wakeUV[1] [1];
	gt3[0].u2 = wakeUV[5] [0];	gt3[0].v2 = wakeUV[5] [1];

	gt3[1].u0 = wakeUV[2] [0];	gt3[1].v0 = wakeUV[2] [1];
	gt3[1].u1 = wakeUV[3] [0];	gt3[1].v1 = wakeUV[3] [1];
	gt3[1].u2 = wakeUV[6] [0];	gt3[1].v2 = wakeUV[6] [1];

	gt3[2].u0 = wakeUV[0] [0];	gt3[2].v0 = wakeUV[0] [1];
	gt3[2].u1 = wakeUV[5] [0];	gt3[2].v1 = wakeUV[5] [1];
	gt3[2].u2 = wakeUV[4] [0];	gt3[2].v2 = wakeUV[4] [1];

	gt3[3].u0 = wakeUV[3] [0];	gt3[3].v0 = wakeUV[3] [1];
	gt3[3].u1 = wakeUV[7] [0];	gt3[3].v1 = wakeUV[7] [1];
	gt3[3].u2 = wakeUV[6] [0];	gt3[3].v2 = wakeUV[6] [1];
    
/*
    gt3[0].r0 = gt3[0].g0 = gt3[0].b0 = 
    gt3[0].r1 = gt3[0].g1 = gt3[0].b1 = 
    gt3[0].r2 = gt3[0].g2 = gt3[0].b2 = 
    gt3[1].r0 = gt3[1].g0 = gt3[1].b0 = 
    gt3[1].r1 = gt3[1].g1 = gt3[1].b1 = 
    gt3[1].r2 = gt3[1].g2 = gt3[1].b2 = 
    gt3[2].r0 = gt3[2].g0 = gt3[2].b0 = 
    gt3[2].r1 = gt3[2].g1 = gt3[2].b1 = 
    gt3[2].r2 = gt3[2].g2 = gt3[2].b2 = 
    gt3[3].r0 = gt3[3].g0 = gt3[3].b0 = 
    gt3[3].r1 = gt3[3].g1 = gt3[3].b1 = 
	gt3[3].r2 = gt3[3].g2 = gt3[3].b2 = 127;
*/

	SetPolyGT4(&gt4);
	setSemiTrans(&gt4,1);
	
	gt4.tpage = GetTPage(0,1,704,256);
	gt4.clut = GetClut(720,446);
	
    gt4.u0 = sprayUV[0] [0];	gt4.v0 = sprayUV[0] [1];
    gt4.u1 = sprayUV[1] [0];	gt4.v1 = sprayUV[1] [1];
    gt4.u2 = sprayUV[2] [0];	gt4.v2 = sprayUV[2] [1];
    gt4.u3 = sprayUV[3] [0];	gt4.v3 = sprayUV[3] [1];
	}

void doWake(register int p)
	{
	register int collret;
	register OBJ_STRUCT *objs;
	register OBJ_PLAYERSTRUCT *objp;
	register PLWAKE *objw;
	register GsOT *ot;
	register int i;
	register int j;
	register int otshift;
	register int shipType;

    static vec32 shipUpVect = {0,4096,0};
    static vec32 shipFwdVect;
    static vec32 shipRightVect;
    
//	vec32 op[3];

    s32 otz,opz,flag;
    SVECTOR coord[4];

	MATRIX cam;
	SVECTOR camEuler;

    s32 vLen;
	
	int br;

	if(!wake[p].inited)
		return;

    ot = ptrwot;
	otshift = 14 - ptrwot->length;
	objp = &OBJ_players[p];
	objs = objp->pobj;
	objw = &wake[p];
    shipType = objp->crafttype;

	// Todo : cyclic trail buffer. (speed ?)
	for(i=9;i>=1;i--)
		{
		objw->trail[i] = objw->trail[i-1];
		
		objw->trail[i].br -= 16;
		if(objw->trail[i].br<0)
			objw->trail[i].br = 0;
		}

	// -- Calculate unit velocity vector
	
	// objs->v 1.23.8
    // calculate length of vector, nasty C style code.
	{
	register s32 x = objs->v.vx;
	register s32 y = objs->v.vy;
	register s32 z = objs->v.vz;
    register s32 vLen2;
	x >>= 4;	y >>= 4;	z >>= 4;
	x *= x;		y *= y;		z *= z;
    vLen2 = x + y + z;
    vLen = SquareRoot0(vLen2);
	}
	// unit velocity
	{
	vec16 fwd;
	register s32 x = objs->v.vx;	// 1.23.8
	register s32 y = objs->v.vy;	// 1.23.8
	register s32 z = objs->v.vz;	// 1.23.8
	// vLen	// 1.23.8
	// ((x>>8) / (vLen>>8))<<8
	if(vLen<50)		// Adjust ?? 
		{
		static SVECTOR fwdSrc = {0,0,4096};
	    ApplyMatrix((MATRIX *)&objp->SIM_objmat,(SVECTOR *)&fwdSrc,(VECTOR *)&fwd);
//		fwd.vx = fwd.vy = fwd.vz = 0;
		}
	else{
        fwd.vx = (x<<8)/vLen;           // fwd is 1.3.12
    	fwd.vy = (y<<8)/vLen;
	    fwd.vz = (z<<8)/vLen;
		}
    shipFwdVect.vx = fwd.vx;
    shipFwdVect.vy = fwd.vy;
    shipFwdVect.vz = fwd.vz;
	}
	// Cross product the son of a bitch's
	// fwd: 1.19.12
	// up: 1.19.12
    OuterProduct12((VECTOR *)&shipFwdVect,(VECTOR *)&shipUpVect,(VECTOR *)&shipRightVect);
/*
	// right: 1.7.24
    shipRightVect.vx >>= 12;
    shipRightVect.vy >>= 12;
    shipRightVect.vz >>= 12;
*/
    // right: 1.19.12

	for(i=0;i<3;i++)
		{
		// Calculate points about objects rot matrix
//	    ApplyMatrix((MATRIX *)&objp->SIM_objmat,(SVECTOR *)&wakeOffset[shipType] [i],(VECTOR *)&objw->rotPoint[i]);
		
		// Calculate points using cross product of velocity + up vector
        objw->rotPoint[i].vx = ((wakeOffset[shipType] [i].vx * shipRightVect.vx)>>12)
                             + ((wakeOffset[shipType] [i].vy * shipUpVect.vx)>>12)
                             + ((wakeOffset[shipType] [i].vz * shipFwdVect.vx)>>12);
        objw->rotPoint[i].vy = ((wakeOffset[shipType] [i].vx * shipRightVect.vy)>>12)
                             + ((wakeOffset[shipType] [i].vy * shipUpVect.vy)>>12)
                             + ((wakeOffset[shipType] [i].vz * shipFwdVect.vy)>>12);
        objw->rotPoint[i].vz = ((wakeOffset[shipType] [i].vx * shipRightVect.vz)>>12)
                             + ((wakeOffset[shipType] [i].vy * shipUpVect.vz)>>12)
                             + ((wakeOffset[shipType] [i].vz * shipFwdVect.vz)>>12);

        objw->shipPos[i].vx = (objs->p.vx + objw->rotPoint[i].vx) >> 8;
        objw->shipPos[i].vy = ((objs->p.vy + objw->rotPoint[i].vy) >> 8) - 100;
        objw->shipPos[i].vz = (objs->p.vz + objw->rotPoint[i].vz) >> 8;

        objw->shipCol[i].vx = objw->shipPos[i].vx;
        objw->shipCol[i].vy = objw->shipPos[i].vy + 600;
        objw->shipCol[i].vz = objw->shipPos[i].vz;

        collret = COL_test2(&objw->shipPos[i],&objw->shipCol[i]);
        if(collret!=-1)
            objw->floorType[i] = COL_surftype;    // Cannot use COL_surftype currently
        else{
			objw->floorType[i] = 0;
			continue;
			}
		
        objw->shipPos[i].vy += 100;		// adjust ship pos after check

        objw->shipRideHeight[i] = COL_inter.vy - objw->shipPos[i].vy;

		objw->wakePos[i].vx = objw->shipPos[i].vx;
        objw->wakePos[i].vy = COL_inter.vy;
		objw->wakePos[i].vz = objw->shipPos[i].vz;

		// Cap height 
		if(objw->shipRideHeight[i]<0)				// <-20 ???
			objw->shipRideHeight[i] = WAKE_SHIPHEIGHTMAX;
		else
		if(objw->shipRideHeight[i]>WAKE_SHIPHEIGHTMAX)
			objw->shipRideHeight[i] = WAKE_SHIPHEIGHTMAX;
        }

	// Process wake, currently *very* basic
	for(i=0;i<3;i++)
		{
		objw->trail[0].valid[i] = objw->floorType[i]==17?1:0;
		
		if(!objw->trail[0].valid[i])
			continue;
		
		objw->trail[0].coord[i] = objw->wakePos[i];
        }
	
    if((!objw->trail[0].valid[0])||(!objw->trail[0].valid[1])||(!objw->trail[0].valid[2]))
		objw->trail[0].br = 0;
	else{
        register s32 h0 = objw->shipRideHeight[0];
        register s32 h1 = objw->shipRideHeight[1];
        register s32 h2 = objw->shipRideHeight[2];
		register s32 h;
    
        // temp: hard coded height min/max 20-52
	
        h0 = 52 - h0;       // 0 - 32
        h1 = 52 - h1;       // 0 - 32
        h2 = 52 - h2;       // 0 - 32
    
		h = (h0 + h1 + h2) / 3;		// 0 - 24
	
        objw->trail[0].br = h + 38;		// 0 - 78
		
//		printf("%08X %08X %08X\n",objs->v.vx,objs->v.vy,objs->v.vz);
//		objs->v;
//		printf("%3d ",objw->trail[0].br);
		}

	// Draw wake *basic*
	camEuler.vx = CAM_came[CAM_c].vx >> 4;
	camEuler.vy = CAM_came[CAM_c].vy >> 4;
	camEuler.vz = CAM_came[CAM_c].vz >> 4;

    RotMatrix(&camEuler,&cam);

	cam.m[0][0] = (cam.m[0][0]*25) >> 4;	// aspect fix
	cam.m[0][1] = (cam.m[0][1]*25) >> 4;	// aspect fix
	cam.m[0][2] = (cam.m[0][2]*25) >> 4;	// aspect fix

    SetRotMatrix(&cam);
	SetTransMatrix((MATRIX *)&IDENTITY_MATRIX);

	// Transform wake coords to screen
	// Temp: Draw basic version

    br = objw->trail[0].br;
//  objw->trail[0].br = 64;
   
    for(j=0;j<9;j++)
		{
		// Temp: Draw 2 tris for each trail link.
		
		// Temp: Abort if any coords are not on water.
        if(!objw->trail[j].valid[0])
            continue;
        if(!objw->trail[j].valid[1])
            continue;
        if(!objw->trail[j].valid[2])
			continue;
        if(!objw->trail[j+1].valid[0])
            continue;
        if(!objw->trail[j+1].valid[1])
            continue;
        if(!objw->trail[j+1].valid[2])
			continue;

		/*
		
		     x------x------x      [j]
		     | \  1 | 2  / |
		     |   \  |  /   |
		     | 3   \|/  4  |
		     x------x------x      [j+1]
			  
		*/

//		nextpoly1:
		
		// Adjust for camera location
        coord[0].vx = (objw->trail[j].coord[0].vx) - (CAM_camp[CAM_c].vx);
        coord[0].vy = (objw->trail[j].coord[0].vy) - (CAM_camp[CAM_c].vy);
        coord[0].vz = (objw->trail[j].coord[0].vz) - (CAM_camp[CAM_c].vz);
        coord[1].vx = (objw->trail[j].coord[1].vx) - (CAM_camp[CAM_c].vx);
        coord[1].vy = (objw->trail[j].coord[1].vy) - (CAM_camp[CAM_c].vy);
        coord[1].vz = (objw->trail[j].coord[1].vz) - (CAM_camp[CAM_c].vz);
        coord[2].vx = (objw->trail[j+1].coord[1].vx) - (CAM_camp[CAM_c].vx);
        coord[2].vy = (objw->trail[j+1].coord[1].vy) - (CAM_camp[CAM_c].vy);
        coord[2].vz = (objw->trail[j+1].coord[1].vz) - (CAM_camp[CAM_c].vz);

		// Rot trans pers this tri.
	    gte_ldv3(&coord[0],&coord[1],&coord[2]);
	    gte_rtpt();
	    gte_stsxy3(&gt3[0].x0,&gt3[0].x1,&gt3[0].x2);
		gte_stflg(&flag);
	
#if !IGNOREFLAG
		if((flag&0x80000000))		// Valid screen coords ?
			goto nextpoly2;
#endif

		gte_nclip();
	    gte_stopz(&opz);
		gte_stszotz(&otz);

#if !IGNOREFLAG
	    if(otz<0)					// Behind camera ?
			goto nextpoly2;
#endif

		// Temp : brightness of vertex = dist down table
        gt3[0].r0 = gt3[0].g0 = gt3[0].b0 = objw->trail[j].br;
   	    gt3[0].r1 = gt3[0].g1 = gt3[0].b1 = objw->trail[j].br;
       	gt3[0].r2 = gt3[0].g2 = gt3[0].b2 = objw->trail[j+1].br;

		// Todo: Check otz < ot range,  should be ok though.
        GsSortPoly(&gt3[0],ptrwot,otz>>otshift);
        
		nextpoly2:
		
		// Adjust for camera location
        coord[0].vx = (objw->trail[j].coord[1].vx) - (CAM_camp[CAM_c].vx);
        coord[0].vy = (objw->trail[j].coord[1].vy) - (CAM_camp[CAM_c].vy);
        coord[0].vz = (objw->trail[j].coord[1].vz) - (CAM_camp[CAM_c].vz);
        coord[1].vx = (objw->trail[j].coord[2].vx) - (CAM_camp[CAM_c].vx);
        coord[1].vy = (objw->trail[j].coord[2].vy) - (CAM_camp[CAM_c].vy);
        coord[1].vz = (objw->trail[j].coord[2].vz) - (CAM_camp[CAM_c].vz);
        coord[2].vx = (objw->trail[j+1].coord[1].vx) - (CAM_camp[CAM_c].vx);
        coord[2].vy = (objw->trail[j+1].coord[1].vy) - (CAM_camp[CAM_c].vy);
        coord[2].vz = (objw->trail[j+1].coord[1].vz) - (CAM_camp[CAM_c].vz);

		// Rot trans pers this tri.
	    gte_ldv3(&coord[0],&coord[1],&coord[2]);
	    gte_rtpt();
	    gte_stsxy3(&gt3[1].x0,&gt3[1].x1,&gt3[1].x2);
		gte_stflg(&flag);
	
#if !IGNOREFLAG
		if((flag&0x80000000))		// Valid screen coords ?
			goto nextpoly3;
#endif
		
		gte_nclip();
	    gte_stopz(&opz);
		gte_stszotz(&otz);

#if !IGNOREFLAG
	    if(otz<0)					// Behind camera ?
			goto nextpoly3;
#endif

		// Temp : brightness of vertex = dist down table
        gt3[1].r0 = gt3[1].g0 = gt3[1].b0 = objw->trail[j].br;
        gt3[1].r1 = gt3[1].g1 = gt3[1].b1 = objw->trail[j].br;
        gt3[1].r2 = gt3[1].g2 = gt3[1].b2 = objw->trail[j+1].br;

		// Todo: Check otz < ot range,  should be ok though.
        GsSortPoly(&gt3[1],ptrwot,otz>>otshift);
        
		nextpoly3:
		
		// Adjust for camera location
        coord[0].vx = (objw->trail[j].coord[0].vx) - (CAM_camp[CAM_c].vx);
        coord[0].vy = (objw->trail[j].coord[0].vy) - (CAM_camp[CAM_c].vy);
        coord[0].vz = (objw->trail[j].coord[0].vz) - (CAM_camp[CAM_c].vz);
        coord[1].vx = (objw->trail[j+1].coord[1].vx) - (CAM_camp[CAM_c].vx);
        coord[1].vy = (objw->trail[j+1].coord[1].vy) - (CAM_camp[CAM_c].vy);
        coord[1].vz = (objw->trail[j+1].coord[1].vz) - (CAM_camp[CAM_c].vz);
        coord[2].vx = (objw->trail[j+1].coord[0].vx) - (CAM_camp[CAM_c].vx);
        coord[2].vy = (objw->trail[j+1].coord[0].vy) - (CAM_camp[CAM_c].vy);
        coord[2].vz = (objw->trail[j+1].coord[0].vz) - (CAM_camp[CAM_c].vz);

		// Rot trans pers this tri.
	    gte_ldv3(&coord[0],&coord[1],&coord[2]);
	    gte_rtpt();
	    gte_stsxy3(&gt3[2].x0,&gt3[2].x1,&gt3[2].x2);
		gte_stflg(&flag);
	
#if !IGNOREFLAG
		if((flag&0x80000000))		// Valid screen coords ?
			goto nextpoly4;
#endif
		
		gte_nclip();
	    gte_stopz(&opz);
		gte_stszotz(&otz);

#if !IGNOREFLAG
	    if(otz<0)					// Behind camera ?
			goto nextpoly4;
#endif
		// Temp : brightness of vertex = dist down table
        gt3[2].r0 = gt3[2].g0 = gt3[2].b0 = objw->trail[j].br;
        gt3[2].r1 = gt3[2].g1 = gt3[2].b1 = objw->trail[j+1].br;
        gt3[2].r2 = gt3[2].g2 = gt3[2].b2 = objw->trail[j+1].br;

		// Todo: Check otz < ot range,  should be ok though.
        GsSortPoly(&gt3[2],ptrwot,otz>>otshift);
        
		nextpoly4:
		
		// Adjust for camera location
        coord[0].vx = (objw->trail[j].coord[2].vx) - (CAM_camp[CAM_c].vx);
        coord[0].vy = (objw->trail[j].coord[2].vy) - (CAM_camp[CAM_c].vy);
        coord[0].vz = (objw->trail[j].coord[2].vz) - (CAM_camp[CAM_c].vz);
        coord[1].vx = (objw->trail[j+1].coord[2].vx) - (CAM_camp[CAM_c].vx);
        coord[1].vy = (objw->trail[j+1].coord[2].vy) - (CAM_camp[CAM_c].vy);
        coord[1].vz = (objw->trail[j+1].coord[2].vz) - (CAM_camp[CAM_c].vz);
        coord[2].vx = (objw->trail[j+1].coord[1].vx) - (CAM_camp[CAM_c].vx);
        coord[2].vy = (objw->trail[j+1].coord[1].vy) - (CAM_camp[CAM_c].vy);
        coord[2].vz = (objw->trail[j+1].coord[1].vz) - (CAM_camp[CAM_c].vz);

		// Rot trans pers this tri.
	    gte_ldv3(&coord[0],&coord[1],&coord[2]);
	    gte_rtpt();
	    gte_stsxy3(&gt3[3].x0,&gt3[3].x1,&gt3[3].x2);
		gte_stflg(&flag);
	
#if !IGNOREFLAG
		if((flag&0x80000000))		// Valid screen coords ?
			goto nextpoly5;
#endif
		
		gte_nclip();
	    gte_stopz(&opz);
		gte_stszotz(&otz);

#if !IGNOREFLAG
	    if(otz<0)					// Behind camera ?
			goto nextpoly5;
#endif
		// Temp : brightness of vertex = dist down table
        gt3[3].r0 = gt3[3].g0 = gt3[3].b0 = objw->trail[j].br;
        gt3[3].r1 = gt3[3].g1 = gt3[3].b1 = objw->trail[j+1].br;
        gt3[3].r2 = gt3[3].g2 = gt3[3].b2 = objw->trail[j+1].br;

		// Todo: Check otz < ot range,  should be ok though.
        GsSortPoly(&gt3[3],ptrwot,otz>>otshift);
        
		nextpoly5:
		continue;
        }

	objw->trail[0].br = br;

	br = vLen;
	
    br >>= 6; // 8192 => 128
	
//	printf("%d\n",br);
	
	if(br<0)
		br = 0;
	else
	if(br>128)
		br = 128;

    gt4.r0 = gt4.g0 = gt4.b0 =
    gt4.r1 = gt4.g1 = gt4.b1 =
    gt4.r2 = gt4.g2 = gt4.b2 =
    gt4.r3 = gt4.g3 = gt4.b3 = br;

	// Do front / side bits
	for(i=0;i<8;i++)
		{
//      ApplyMatrix((MATRIX *)&objp->SIM_objmat,(SVECTOR *)&sprayOffset[i],(VECTOR *)&sprayRot[i]);

        sprayRot[i].vx = ((sprayOffset[i].vx * shipRightVect.vx)>>12)
                       + ((sprayOffset[i].vy * shipUpVect.vx)>>12)
                       + ((sprayOffset[i].vz * shipFwdVect.vx)>>12);
        sprayRot[i].vy = ((sprayOffset[i].vx * shipRightVect.vy)>>12)
                       + ((sprayOffset[i].vy * shipUpVect.vy)>>12)
                       + ((sprayOffset[i].vz * shipFwdVect.vy)>>12);
        sprayRot[i].vz = ((sprayOffset[i].vx * shipRightVect.vz)>>12)
                       + ((sprayOffset[i].vy * shipUpVect.vz)>>12)
                       + ((sprayOffset[i].vz * shipFwdVect.vz)>>12);

        sprayRotS[i].vx = (objs->p.vx + sprayRot[i].vx) >> 8;
        sprayRotS[i].vy = ((objs->p.vy + sprayRot[i].vy) >> 8) - 100;
        sprayRotS[i].vz = (objs->p.vz + sprayRot[i].vz) >> 8;

        sprayRotSC[i].vx = sprayRotS[i].vx;
        sprayRotSC[i].vy = sprayRotS[i].vy + 600;
        sprayRotSC[i].vz = sprayRotS[i].vz;

        collret = COL_test2(&sprayRotS[i],&sprayRotSC[i]);
/*
        if(collret!=-1)
            objw->floorType[i] = COL_surftype;    // Cannot use COL_surftype currently
        else{
			objw->floorType[i] = 0;
			continue;
			}
*/
        sprayRotS[i].vy = COL_inter.vy;
		
		if(i>=4)
			sprayRotS[i].vy -= 50;
        }

    SetRotMatrix(&cam);
	SetTransMatrix((MATRIX *)&IDENTITY_MATRIX);

    coord[0].vx = (sprayRotS[4].vx) - (CAM_camp[CAM_c].vx);
    coord[0].vy = (sprayRotS[4].vy) - (CAM_camp[CAM_c].vy);
    coord[0].vz = (sprayRotS[4].vz) - (CAM_camp[CAM_c].vz);
    coord[1].vx = (sprayRotS[5].vx) - (CAM_camp[CAM_c].vx);
    coord[1].vy = (sprayRotS[5].vy) - (CAM_camp[CAM_c].vy);
    coord[1].vz = (sprayRotS[5].vz) - (CAM_camp[CAM_c].vz);
    coord[2].vx = (sprayRotS[0].vx) - (CAM_camp[CAM_c].vx);
    coord[2].vy = (sprayRotS[0].vy) - (CAM_camp[CAM_c].vy);
    coord[2].vz = (sprayRotS[0].vz) - (CAM_camp[CAM_c].vz);
    coord[3].vx = (sprayRotS[1].vx) - (CAM_camp[CAM_c].vx);
    coord[3].vy = (sprayRotS[1].vy) - (CAM_camp[CAM_c].vy);
    coord[3].vz = (sprayRotS[1].vz) - (CAM_camp[CAM_c].vz);

	// Rot trans pers 1st part of this quad.
    gte_ldv3(&coord[0],&coord[1],&coord[2]);
    gte_rtpt();
    gte_stsxy3(&gt4.x0,&gt4.x1,&gt4.x2);
	gte_stflg(&flag);
#if !IGNOREFLAG
	if((flag&0x80000000))		// Valid screen coords ?
		goto nextpoly6;
#endif
	gte_nclip();
    gte_stopz(&opz);
	gte_stszotz(&otz);
#if !IGNOREFLAG
    if(otz<0)					// Behind camera ?
		goto nextpoly6;
#endif
	gte_ldv0(&coord[3]);
	gte_rtps();
    gte_stsxy(&gt4.x3);
	
    // Temp : arsed up
//	gt4.r0 = 255;	gt4.g0 = 0;	gt4.b0 = 0;

	// Todo: Check otz < ot range,  should be ok though.
	if(!globalWakeDisable)
	    GsSortPoly(&gt4,ptrwot,otz>>otshift);

	nextpoly6:

    coord[0].vx = (sprayRotS[6].vx) - (CAM_camp[CAM_c].vx);
    coord[0].vy = (sprayRotS[6].vy) - (CAM_camp[CAM_c].vy);
    coord[0].vz = (sprayRotS[6].vz) - (CAM_camp[CAM_c].vz);
    coord[1].vx = (sprayRotS[7].vx) - (CAM_camp[CAM_c].vx);
    coord[1].vy = (sprayRotS[7].vy) - (CAM_camp[CAM_c].vy);
    coord[1].vz = (sprayRotS[7].vz) - (CAM_camp[CAM_c].vz);
    coord[2].vx = (sprayRotS[2].vx) - (CAM_camp[CAM_c].vx);
    coord[2].vy = (sprayRotS[2].vy) - (CAM_camp[CAM_c].vy);
    coord[2].vz = (sprayRotS[2].vz) - (CAM_camp[CAM_c].vz);
    coord[3].vx = (sprayRotS[3].vx) - (CAM_camp[CAM_c].vx);
    coord[3].vy = (sprayRotS[3].vy) - (CAM_camp[CAM_c].vy);
    coord[3].vz = (sprayRotS[3].vz) - (CAM_camp[CAM_c].vz);

	// Rot trans pers 1st part of this quad.
    gte_ldv3(&coord[0],&coord[1],&coord[2]);
    gte_rtpt();
    gte_stsxy3(&gt4.x0,&gt4.x1,&gt4.x2);
	gte_stflg(&flag);
#if !IGNOREFLAG
	if((flag&0x80000000))		// Valid screen coords ?
		goto nextpoly7;
#endif
	gte_nclip();
    gte_stopz(&opz);
	gte_stszotz(&otz);
#if !IGNOREFLAG
    if(otz<0)					// Behind camera ?
		goto nextpoly7;
#endif
	gte_ldv0(&coord[3]);
	gte_rtps();
    gte_stsxy(&gt4.x3);
	
    // Temp : arsed up
//	gt4.r0 = 255;	gt4.g0 = 0;	gt4.b0 = 0;

	// Todo: Check otz < ot range,  should be ok though.
	if(!globalWakeDisable)
	    GsSortPoly(&gt4,ptrwot,otz>>otshift);

	nextpoly7:
	
	return;

/*
	register long *tag;
	register GsOT_TAG *org;

    tag = (u_long *)(org+(z>>(14-OT_LENGTH)));
	sktptr->tag = (u_long)((*tag&0x00ffffff)|0x09000000);	// 9 = FT4
	*tag = ((u_long)(sktptr++)&0x00ffffff);
*/
    }
