#include "header.h"
#include "state.h"
#include "objects.h"
#include "collide.h"
#include "weapons.h"
#include "sdlib\sdfunc.h"
#include "psx\soundpsx.h"
#include <rand.h>
#if PSX
	#include "PSX\PSX_pcks.h"
    #include "PSX\PSX_expn.h"
#endif
#include "psx\mathpsx.h"
#include "psx\textpsx.h"


s32	power_up_table[10][5] = {

/*Order of elements are:

0:	Damage Value.
1:	Replenish Rate.
2:	Tracer Size.
3:	Explosion Rating.
4:	Muzzle Flash Length

*/

{15, 5, 5, exp_level_0, 35},		//Level 0
{10, 12, 2, exp_level_1, 15},		//Level 1
{15, 14, 3, exp_level_2, 20},		//Level 2
{20, 16, 4, exp_level_3, 25},		//Level 3
{25, 18, 5, exp_level_4, 30},		//Level 4
{30, 20, 6, exp_level_5, 35},		//Level 5
{35, 22, 7, exp_level_6, 40},		//Level 6
{40, 24, 8, exp_level_7, 45},		//Level 7
{45, 26, 9, exp_level_8, 50},		//Level 8
{50, 30, 10, exp_level_9, 60} 	//Level 8

};

int blt_timer_on;
int blt_timer_off;

//Local function prototypes
void init_bullet(OBJ_PLAYERSTRUCT *plyr_ptr);
void init_tracer(OBJ_PLAYERSTRUCT *plyr_ptr);
//void init_bullet_exp(OBJ_STRUCT *obj_ptr);


void WPN_process_weapon(OBJ_PLAYERSTRUCT *plyr_ptr)
//This function is the main controlling function for object interaction. It checks
//pad data for a particular player and checks to see if a weapon/pickup has either
//been used or changed. It either is true it perform a few simple tests and updates
//structures accordingly

// *plyr_ptr is a pointer to the current player structure bieng processed.
{
    s32             pad_data;
	int				tracer;


	if (plyr_ptr->timer < 250)
		return;

//	blt_timer_on = 230;
    pad_data = plyr_ptr->SIM_input_data;                         //gets pad data for player
                                                            //for current player
//if (plyr_ptr->plyrnum != 1)
//{
    if (plyr_ptr->ammo < MAX_AMMO_LEVEL)
    	plyr_ptr->ammo += (power_up_table[plyr_ptr->weapon_level][replenish_rate] * 2);
//}
    if (plyr_ptr->ammo > MAX_AMMO_LEVEL)
    	plyr_ptr->ammo = MAX_AMMO_LEVEL;

    if (pad_data & PAD_RR)         //if fire button pressed
    {
	    if (plyr_ptr->ammo > 150)
       	{

         S2_sfxdynamic( 4, plyr_ptr->pobj->p.vx>>8, plyr_ptr->pobj->p.vz>>8, 0, 127, 6000 + (rand()&31) );
			tracer = SDRAND(50);

			if (tracer < 10)
				init_tracer(plyr_ptr);
			else
    	    {
            	plyr_ptr->ammo -= 150;

				if (plyr_ptr->ammo < 0)
					plyr_ptr->ammo = 0;

        	    init_bullet(plyr_ptr);

				tracer = SDRAND(6);

				if (tracer < 4)
					PSX_process_muzzle_flash(plyr_ptr);
			}
        }
    }

//	blt_timer_off = 240;
    return;
}


void init_tracer(OBJ_PLAYERSTRUCT *plyr_ptr)
{
    s32             index;
    OBJ_STRUCT      *obj_ptr, *plyr_obj_ptr;
    long            x_vel, z_vel, y_vel, temp_vel, vel, side;


    index = OBJ_newobj(OTtracer);

    if (index == NULLVAL)
        return;                             //No more object structures.

    obj_ptr = &OBJ_objects[index];
    plyr_obj_ptr = plyr_ptr->pobj;

	if (muzzle_flash_offset_table[plyr_ptr->crafttype][6] == 2)
	{
		side = SDRAND(2);

		if (side)
		{
			obj_ptr->p.vx = muzzle_flash_offset_table[plyr_ptr->crafttype][0] << 8;
			obj_ptr->p.vy = muzzle_flash_offset_table[plyr_ptr->crafttype][1] << 8;
			obj_ptr->p.vz = muzzle_flash_offset_table[plyr_ptr->crafttype][2] << 8;
		}
		else
		{
			obj_ptr->p.vx = muzzle_flash_offset_table[plyr_ptr->crafttype][3] << 8;
			obj_ptr->p.vy = muzzle_flash_offset_table[plyr_ptr->crafttype][4] << 8;
			obj_ptr->p.vz = muzzle_flash_offset_table[plyr_ptr->crafttype][5] << 8;
		}
	}
	else
	{
		obj_ptr->p.vx = muzzle_flash_offset_table[plyr_ptr->crafttype][0] << 8;
		obj_ptr->p.vy = muzzle_flash_offset_table[plyr_ptr->crafttype][1] << 8;
		obj_ptr->p.vz = muzzle_flash_offset_table[plyr_ptr->crafttype][2] << 8;
	}

    obj_ptr->p.vx += (plyr_obj_ptr->p.vx);         //Sets torpedo position and euler orientation to
    obj_ptr->p.vy += (plyr_obj_ptr->p.vy);         //that of the owner ship
    obj_ptr->p.vz += (plyr_obj_ptr->p.vz);

    obj_ptr->r.vx = (plyr_obj_ptr->r.vx);//0;                     //Not yet set by Jon
    obj_ptr->r.vy = (plyr_obj_ptr->r.vy);    //Range is 0-65536
    obj_ptr->r.vz = (plyr_obj_ptr->r.vz);                     //Not yet set by Jon

	vel = 2000;

    y_vel = (vel * rsin(plyr_obj_ptr->r.vx >> 4)) >> 12;   //Obtain velocities in both x and z planes, using TVEL,
    temp_vel = (vel * rcos(plyr_obj_ptr->r.vx >> 4)) >> 12;   //the scaler velocity of the torpedo and the crafts

    x_vel = (temp_vel * rsin(obj_ptr->r.vy >> 4)) >> 12;   //the scaler velocity of the torpedo and the crafts
    z_vel = (temp_vel * rcos(obj_ptr->r.vy >> 4)) >> 12;   //Obtain velocities in both x and z planes, using TVEL,
                                        //euler orientation.

//    obj_ptr->v.vx = plyr_obj_ptr->v.vx + x_vel;                 //Set velocity of torpedo in all 3 planes respectively.
//    obj_ptr->v.vy = plyr_obj_ptr->v.vy + y_vel;
//    obj_ptr->v.vz = plyr_obj_ptr->v.vz + z_vel;

    obj_ptr->v.vx = x_vel << 8;                 //Set velocity of torpedo in all 3 planes respectively.
    obj_ptr->v.vy = y_vel << 8;
    obj_ptr->v.vz = z_vel << 8;

    obj_ptr->owner = plyr_ptr->plyrnum;       //Sets owner of torpedo
    //jofobj_ptr->sect = plyr_obj_ptr->sect;         //Sets current section of course that torpedo is in

	obj_ptr->pkup_data.type = NULLVAL;
	obj_ptr->pkup_data.power = 0;
	obj_ptr->m0 = 0;

    return;
}

void init_bullet(OBJ_PLAYERSTRUCT *plyr_ptr)
{
    s32             index, success;
	OBJ_PLAYERSTRUCT	*target_ptr;
    OBJ_STRUCT      *obj_ptr, *plyr_obj_ptr;
    vec32           first_arg, second_arg, impact_vector;
    s32             coll_val, vel, temp_vel, x_vel, y_vel, z_vel;
	s32				owner, type, power;
	u8				text[25];


//	sprintf(text, "MATTY is VERY gay!!!!!");

//	TEXT_x = 0;
//	TEXT_y = 0;
//	TEXT_prstr_BIG(text, 255, 64, 32);

//    return;


    plyr_obj_ptr = plyr_ptr->pobj;

    owner = plyr_ptr->plyrnum;       //Sets owner of torpedo
    //jofobj_ptr->sect = plyr_obj_ptr->sect;         //Sets current section of course that torpedo is in

	type = bullet_explosion;
	power = power_up_table[plyr_ptr->weapon_level][damage_value];

    first_arg.vx = plyr_obj_ptr->p.vx >> 8;
    first_arg.vy = plyr_obj_ptr->p.vy >> 8;
    first_arg.vz = plyr_obj_ptr->p.vz >> 8;

	vel = 10000;

    y_vel = -(vel * rsin(plyr_obj_ptr->r.vx >> 4)) >> 12;   //Obtain velocities in both x and z planes, using TVEL,
    temp_vel = (vel * rcos(plyr_obj_ptr->r.vx >> 4)) >> 12;   //the scaler velocity of the torpedo and the crafts

    x_vel = (temp_vel * rsin(plyr_obj_ptr->r.vy >> 4)) >> 12;   //the scaler velocity of the torpedo and the crafts
    z_vel = (temp_vel * rcos(plyr_obj_ptr->r.vy >> 4)) >> 12;   //Obtain velocities in both x and z planes, using TVEL,
                          					        //euler orientation
    second_arg.vx = (plyr_obj_ptr->p.vx >> 8) + x_vel;
    second_arg.vy = (plyr_obj_ptr->p.vy >> 8) + y_vel;
    second_arg.vz = (plyr_obj_ptr->p.vz >> 8) + z_vel;

    // mongotron squirts here
{
    vec32   v0, v1, cr, dif;
       s32     d, lensq;
    s16     i;
    s32     cp, dp;


    index = OBJ_newobj(OTbullet);

    if (index == NULLVAL)
        return;                             //No more object structures.

//	if (index < 20)
//		mprintf("not now matey!\n");

//   S2_sfxdynamic( 4, plyr_obj_ptr->p.vx>>8, plyr_obj_ptr->p.vz>>8, 0, 127, 6000 + (rand()&31) );

	obj_ptr = &OBJ_objects[index];

	obj_ptr->owner = plyr_ptr->plyrnum;

	obj_ptr->r.vx = plyr_obj_ptr->r.vx;
	obj_ptr->r.vy = plyr_obj_ptr->r.vy;
	obj_ptr->r.vz = plyr_obj_ptr->r.vz;

	obj_ptr->pkup_data.type = bullet_explosion;
	obj_ptr->pkup_data.power = power_up_table[plyr_ptr->weapon_level][damage_value];

	for (i = 0; i <  STATE.wankers; i ++)
    {
    	if (i == obj_ptr->owner)
    		continue;

      	if ((OBJ_objects[i].stat == ON) && (OBJ_objects[i].type == OTcraft))
      	{
        	mov( dif, second_arg );
        	sub( dif, first_arg );

         	mov( cr, OBJ_objects[i].p );
         	shr( cr, 8 );

	        mov( v0, cr );
	        sub( v0, first_arg );

	        lensq    =  SquareRoot0(dif.vx*dif.vx + dif.vz*dif.vz);
	        cp       =  (dif.vx*v0.vz - dif.vz*v0.vx)/lensq;
	        dp       =  dif.vx*v0.vx + dif.vz*v0.vz;

	        if( (cp<200) && (cp>-200) && (dp>=0) )
	        {
	     //       FntPrint("smeeeee bullet %d\n", cp);

				obj_ptr->p.vx = OBJ_objects[i].p.vx;
				obj_ptr->p.vy = OBJ_objects[i].p.vy;
				obj_ptr->p.vz = OBJ_objects[i].p.vz;

	            target_ptr = &OBJ_players[i];

				OBJ_damage_player(target_ptr, obj_ptr, FALSE);

				if (obj_ptr->stat)
					OBJ_delobj(obj_ptr->index);

	           	return;
	        }
      	}
	}
}

   //	OBJ_delobj(obj_ptr->index);

    coll_val = COL_test2(&first_arg, &second_arg); //Determines if the torpedo has hit the

    if ((coll_val > -1) /*|| (obj_ptr->m0 > 750)*/)
    {
		obj_ptr->p.vx = plyr_obj_ptr->p.vx;
		obj_ptr->p.vy = plyr_obj_ptr->p.vy;
		obj_ptr->p.vz = plyr_obj_ptr->p.vz;

	    if(COL_surftype==17)//water
        {
    	    if( (rand()&3)==0 )
            //SOUND_sfx_p(19, 34+(rand()&3) );
            S2_sfxdynamic( 19, obj_ptr->p.vx>>8, obj_ptr->p.vz>>8, 0, 120, 4000 );
		OBJ_delobj(obj_ptr->index);
        }
    	else
    	{
    		if( (rand()&3)==0 )
            S2_sfxdynamic( 6+rand()%6, obj_ptr->p.vx>>8, obj_ptr->p.vz>>8, 0, 120, 4000 );

//    		OPSX_init_new_explosion(obj_ptr, bullet_explosion, MESH_COLLISION);
			PSX_init_flame_animation(obj_ptr, bullet_explosion, MESH_COLLISION);

			if (obj_ptr->stat)
				OBJ_delobj(obj_ptr->index);
    	}


    }
	else
		OBJ_delobj(obj_ptr->index);

	if (obj_ptr->stat)
		OBJ_delobj(obj_ptr->index);

    return;
}

void WPN_process_tracer(OBJ_STRUCT *obj_ptr)
//This function process's a torpedo for the current player this frame. It controls the movement
//of the object and handles collision with scenary.

//*obj_ptr is a pointer to the current object to be processed
//*plyr_ptr is a pointer to the curent player structure owning the torpedo.
{
	obj_ptr->m0 ++;

	if (obj_ptr->m0 == 25)
		OBJ_delobj(obj_ptr->index);

    obj_ptr->p.vx += obj_ptr->v.vx;
    obj_ptr->p.vy += obj_ptr->v.vy;
    obj_ptr->p.vz += obj_ptr->v.vz;

	PSX_process_tracer(obj_ptr);

    return;
}